#!/usr/bin/env perl

=head1 NAME

MLoader -- a Perl program for loading historical data

=head1 DESCRIPTION

This program is designed for data migration of molecular genetics labs.

=cut

BEGIN {
   use Env qw( APIIS_HOME );
   die "\n\tAPIIS_HOME is not set!\n\n" unless $APIIS_HOME;
   push @INC, "$APIIS_HOME/lib";
}

use strict;
use warnings;
use Apiis;
use Data::Dumper;
use Config::IniFiles;
use Spreadsheet::Read;
use List::Compare;
use String::Compare;
use File::MimeInfo;
use Archive::Tar;
use Bio::SeqIO;
# command line parameters
use vars qw( $opt_h $opt_n $opt_u $opt_p $opt_d $opt_m);
use Getopt::Std;
require apiis_alib;
Apiis->initialize;
use open ':utf8';

use constant TAB 			=> "    ";
use constant COMMENT 		=> "Batch loading";
use constant DUMMY_PROTOCOL => "dummy_protocol.txt";
use constant OUTPUT_DIR 	=> "/tmp";
use constant LEN 			=> 50;
use constant DUMMY			=> "dummy";
use constant DEFAULT_FORMAT => "EXCEL";

# constants stands for error values, please don't change
use constant NO_ERROR			=> 0;
use constant EMPTY_PROJECT		=> 1;
use constant DUPLICATED_PROJECT => 2;
use constant EMPTY_RECORD_NO	=> 3;
use constant INVALID_RECORD_NO	=> 4;
use constant EMPTY_MARKER_NO	=> 5;
use constant INVALID_MARKER_NO	=> 6;
use constant EMPTY_MARKER		=> 7;
use constant DUPLICATED_MARKER	=> 8;
use constant NOT_EQUAL_MARKER	=> 9;
use constant EMPTY_PRIMER		=> 10;
use constant NOT_EQUAL_PRIMER	=> 11;
use constant EMPTY_SPECIES  	=> 12;
use constant EMPTY_SEX		  	=> 13;
use constant FILE_NOT_FOUND	  	=> 14;
use constant EMPTY_MATERIAL	  	=> 15;
use constant EMPTY_VESSEL	  	=> 16;
use constant EMPTY_DNA		  	=> 17;
use constant EMPTY_SQ		  	=> 18;
use constant UNKNOWN_MODE	  	=> 19;
use constant DATA_NOT_FOUND		=> 20;

# showwing messages when getting runtime error
my @em;
$em[0]  = "There is no error.                                       ";
$em[1]  = "Project name can not be empty, please check the loadfile ";
$em[2]  = "A dulicated value was found when inserting a new project ";
$em[3]  = "Number of records can not be empty, check the loadfile   ";
$em[4]  = "Loaded records must be greater than 0, check the loadfile";
$em[5]  = "Number of markers can not be empty, check the loadfile   ";
$em[6]  = "Number of markers must be greater than 0, check loadfile ";
$em[7]  = "Marker name can not be empty, please check the loadfile  ";
$em[8]  = "Different markers keep the same name, check the loadfile ";
$em[9]  = "Number of markers are not equal to the number required   ";
$em[10] = "Primer name can not be empty, please check the loadfile  ";
$em[11] = "Number of primers are not equal to the number required   ";
$em[12] = "Do not accept an empty value when loading species        ";
$em[13] = "Do not accept an empty value when loading sexes          ";
$em[14] = "File was not found, please check & correct the file name ";
$em[15] = "Do not accept an empty value when loading material types ";
$em[16] = "Do not accept an empty value when loading vessel types   ";
$em[17] = "Do not accept an empty value when loading DNA ID         ";
$em[18] = "Do not accept an empty value when loading final sequences";
$em[19] = "Unknown action mode, please use   -m check or -m load    ";
$em[20] = "Cannot find expected data in the datafile, please check  ";

# debuging mode: 1=debug, 0=none
my $debug = 0;

# declare global variables
my ($db_name,$user,$pass,$path,$mode,$flag,$datafile,$msg);
my ($sp_file_ref,$sq_file_ref,$ms_file_ref,$sq_rawfile_ref,$ms_rawfile_ref);
my ($project_name,$project_id,$keyword,$description,$record_no);
my ($sp_datafile_format,$sq_datafile_format,$ms_datafile_format);
my ($sp_loading,$sample_storage,$defined_project);
my ($sq_rawdata_load,$sq_finaldata_load,$ms_rawdata_load,$ms_finaldata_load);
my (%sq_markers,%sq_marker_names,%sq_primer_names,%sq_primer1,%sq_primer2);
my (%ms_markers,%ms_marker_names,%ms_primer_names,%ms_primer1,%ms_primer2);
my (%users,%protocols,%pictures,%sample_storages,%dna_storages,%samples);
my (%raw_files,%sq_elecs,%ms_elecs,%marker_ids,%dna_ids,%dummy_dna_ids);
my ($sq_marker_no,$sq_rawfile_format,$sq_max);
my ($ms_marker_no,$ms_rawfile_format,$ms_max,$ms_datafile,$ms_data_file_type);

run();	# the script starts from here

################################################################################

sub run{
	init();
	my $error_value;
	if (uc($mode) eq 'LOAD'){
		$error_value = load_data();
	} elsif (uc($mode) eq 'CHECK'){
		$error_value = check_data();
	} else {
		$error_value = UNKNOWN_MODE;
	}
	if (($error_value eq NO_ERROR) and (uc($mode) eq 'LOAD')){
	  $apiis->DataBase->dbh->commit;
	  print "\n------------------------ Data Commit ------------------------";
      print "\n-    Historical data was successfully loaded to Database    -";
      print "\n-------------------------------------------------------------\n";
	}
	elsif (($error_value eq NO_ERROR) and (uc($mode) eq 'CHECK')){
	  print "\n------------------------ Report Done ------------------------";
	  print "\n- Please check the report.txt which was saved in the folder -";
      print "\n- $path",         " "x(58-length($path)),                  "-";
      print "\n-------------------------------------------------------------\n";
	}
	else {
	  print "\n--------------------------- Error ---------------------------";
	  print "\n- " .                $em[$error_value]                  . " -";
	  print "\n-------------------------------------------------------------\n";
	}
}

# getting parameters and getting database connection
sub init{
	# read command line parameters
	getopts('hn:u:p:m:d:');	
	usage() if $opt_h;		
	usage() if (!$opt_n or !$opt_u or !$opt_p or !$opt_d or !$opt_m);
	($db_name,$user,$pass,$mode,$path)=($opt_n,$opt_u,$opt_p,$opt_m,$opt_d);
	# prepare a file for rollback data
	open UNLOAD_FILE, ">", "$path/unload.sql"; 
    $flag = NO_ERROR;
    $defined_project = 'no';
	# connect to MolabIS database
	require Apiis::DataBase::User;
	my $dummy = Apiis::DataBase::User->new(id=>$user,password=>$pass,);
	$apiis->join_model($db_name,userobj=>$dummy);
	$apiis->check_status(die=>'ERR');
}


################################################################################
sub check_data{
	my $error = read_loadfile();
	my @key_types = (
      "SPECIES",
	  "BREED",
	  "SEX",
	  "MATERIAL_TYPE",
	  "SAMPLE_VESSEL",
	  "DNA_VESSEL",
	  "ORGANIZATION",
	);
	print "\n-------------------------------------------------------------\n";
	foreach my $type (@key_types){
		my @old_list = int_dataset($type);
		my @new_list = ext_dataset($type);
		compare_keys(\@new_list,\@old_list,$type);
	}
	return NO_ERROR;
}

sub compare_keys{
	my ($new_list_ref,$old_list_ref,$type) = @_;
	my $lc = List::Compare->new($new_list_ref,$old_list_ref);
	my @duplicated_list = $lc->get_intersection;
	my @unique_list = $lc->get_Lonly;
	print "$type:\n";
	print "Available keys stored in database:\n";
	print "\t".join(", ",@$old_list_ref),"\n";
	print "Keys collected from datafile:\n";
	print "\t".join(", ",@$new_list_ref),"\n";
	print "Duplicated keys:\n";
	print "\t".join(", ",@duplicated_list),"\n";
	print "New keys:\n";
	print "\t".join(", ",@unique_list),"\n";
	print "Similar keys:\n";
	foreach my $new_str (@unique_list) {
		foreach my $old_str (@$old_list_ref) {
        	my $denta = compare ($new_str,$old_str);
			if ($denta > 0.7){
				print "\t".$new_str." >> ".$old_str."\n";
			}
		}
	}
    print "\n-------------------------------------------------------------\n";
}

sub int_dataset{
	my $type_of_key = shift;
    my %sql_hash = (
		'SPECIES' =>"SELECT short_name FROM codes WHERE class='SPECIES'",
        'BREED' =>"SELECT short_name FROM codes WHERE class='BREED'",
        'SEX' =>"SELECT short_name FROM codes WHERE class='SEX'",
        'MATERIAL_TYPE' =>"SELECT short_name FROM codes WHERE class='MATERIAL_TYPE'",
		'SAMPLE_VESSEL' =>"SELECT short_name FROM codes WHERE class='VESSEL_TYPE'",
		'DNA_VESSEL' =>"SELECT short_name FROM codes WHERE class='VESSEL_TYPE'",
		'ORGANIZATION' =>"SELECT ext_id FROM unit WHERE ext_unit='ORGANIZATION'",
    );
    my $sql = $sql_hash{$type_of_key};
	my $sql_ref = $apiis->DataBase->user_sql($sql);
	my @data_array; 
	my @record_set;
	while ( @data_array = $sql_ref->handle->fetchrow_array ) {
    	push(@record_set,$data_array[0]);
    }
	return (@record_set);
}

sub ext_dataset{
	my $type_of_key = shift;
    my %col_hash = (
		'SPECIES'		=> 2,
		'BREED'	 		=> 3,
		'SEX'	 		=> 6,
		'ORGANIZATION'	=> 9,
		'MATERIAL_TYPE' => 11,
		'SAMPLE_VESSEL'	=> 15,
		'DNA_VESSEL'	=> 29,
      );
    my ($row,$col) = (3,$col_hash{$type_of_key});
    my %data_hash;
	my @record_set;
	$sp_file_ref = ReadData ("$path/datafile_sp.xls", dtfmt =>"dd-mm-yyy",sep => ",");
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$data_hash{$key}++ if ($key ne "");
	}
	for my $key ( keys %data_hash ) {
		push(@record_set,$key);
	}
	return (@record_set);
}

sub load_data{
	my $error = read_loadfile();
	if ($defined_project eq 'no'){
		return $flag if (load_project() ne NO_ERROR);
		$defined_project = 'yes';
	}
	if ($sp_loading eq 'yes'){
		$error = get_samples() if ($error eq NO_ERROR);
	} 
	elsif ($sp_loading eq 'no'){
		return $flag if (load_primers()	ne NO_ERROR);
		return $flag if (load_markers()	ne NO_ERROR);
		$error = load_dummy() if ($error eq NO_ERROR);
	}
  	if (uc($sq_rawdata_load) eq 'YES'){
		$error = load_sq_rawdata() if ($error eq NO_ERROR);
	}
  	if (uc($ms_rawdata_load) eq 'YES'){
		$error = load_ms_rawdata() if ($error eq NO_ERROR);
	}
	if (uc($sq_finaldata_load) eq 'YES'){
		$error = load_sequences() if ($error eq NO_ERROR);
	}
	if (uc($ms_finaldata_load) eq 'YES'){
		$error = load_microsatellites() if ($error eq NO_ERROR);
	}
	return $error;
}

sub read_loadfile{
	my %ini;
	my $loadfile = "$path/loadfile.conf";
	tie %ini, 'Config::IniFiles', ( -file => $loadfile );
	my %global_section = %{$ini{'GLOBAL'}};
	my %sp_section = %{$ini{'SAMPLES'}};
	my %sq_section = %{$ini{'SEQUENCES'}};
	my %ms_section = %{$ini{'MICROSATELLITES'}};
	print "Reading from the loadfile ...\n";
	for my $key ( keys %global_section ) {
		my $value = $global_section{$key};
		print  "\t$key"," "x(25-length($key)),"= ".$value."\n";
	}
	# read parameters
	$project_name 	= trimstr($global_section{'project_name'});
	if (defined  $global_section{'keyword'}){
    	$keyword = trimstr($global_section{'keyword'});
	}
	else{
    	$keyword = trimstr($global_section{'project_name'});
	}
	if (defined  $global_section{'description'}){
    	$description = trimstr($global_section{'description'});
	}
	else{
    	$description = trimstr($global_section{'project_name'});
	}
	if (defined  $sp_section{'datafile_format'}){
	    $sp_datafile_format = trimstr($sp_section{'datafile_format'});
	} else {
		$sp_datafile_format = DEFAULT_FORMAT;
	}
	if (defined $sq_section{'datafile_format'}){
		$sq_datafile_format = trimstr($sq_section{'datafile_format'});
	} else {
		$sq_datafile_format = DEFAULT_FORMAT;
	}
	if (defined $ms_section{'datafile_format'}){
		$ms_datafile_format = trimstr($ms_section{'datafile_format'});
	} else {
		$ms_datafile_format = DEFAULT_FORMAT;
	}
	if (defined $sq_section{'rawfile_format'}){
		$sq_rawfile_format = trimstr($sq_section{'rawfile_format'});
	} else {
		$sq_rawfile_format = DEFAULT_FORMAT;
	}
	if (defined $ms_section{'rawfile_format'}){
		$ms_rawfile_format = trimstr($ms_section{'rawfile_format'});
	} else {
		$ms_rawfile_format = DEFAULT_FORMAT;
	}
	# point to data files
	my $filename;
	$filename = "datafile_sp.". get_file_extension($sp_datafile_format);
	print "\t* DEBUG: Assigning the file: $path/$filename ...\n" if $debug;
	$sp_file_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");

    $filename = "rawfile_sq.". get_file_extension($sq_rawfile_format);
	$sq_rawfile_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");
	$filename = "rawfile_ms.". get_file_extension($ms_rawfile_format);
	$ms_rawfile_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");
	my $row = 3;
	$record_no = 0;
	while (1){
		last if (read_data($sp_file_ref,$row,1,'yes') eq '');
		$record_no++;
		$row++;
	}
	$sp_loading		   = trimstr($sp_section{'loading'});
	$sq_rawdata_load   = trimstr($sq_section{'rawdata_loading'});
	$sq_finaldata_load = trimstr($sq_section{'finaldata_loading'});
	$sq_marker_no 	   = trimstr($sq_section{'sq_marker_no'});
	$ms_rawdata_load   = trimstr($ms_section{'rawdata_loading'});
	$ms_finaldata_load = trimstr($ms_section{'finaldata_loading'});
	$ms_marker_no 	   = trimstr($ms_section{'ms_marker_no'});
	# getting the file name and data file type
	if (defined $ms_section{'data_file_name'}){
		$ms_datafile = trimstr($ms_section{'ms_datafile'});
	} else {
		$ms_datafile	= "microsatellites.xls";
	}
	if (defined $ms_section{'data_file_type'}){
		$ms_data_file_type = trimstr($ms_section{'data_file_type'});
	} else {
		$ms_data_file_type = 2;
	}

	return EMPTY_PROJECT 		if($project_name eq '');
 	return EMPTY_RECORD_NO 	 	if($record_no eq '');
 	return INVALID_RECORD_NO 	if($record_no <= 0);
	return EMPTY_MARKER_NO 	 	if ($sq_marker_no eq '');
	return INVALID_MARKER_NO 	if ($sq_marker_no <= 0);
	return EMPTY_MARKER_NO 	 	if ($ms_marker_no eq '');
	return INVALID_MARKER_NO 	if ($ms_marker_no <= 0);
	# read all markers and related primers
	$msg = "Sequencing markers";
	print  "\t$msg"," "x(25-length($msg)),"= ".$sq_marker_no."\n";
	for (my $i=1; $i<= $sq_marker_no; $i++){
		if (defined $sq_section{'sq_marker_'.$i}){
			$sq_markers{'sq_marker_'.$i} = $sq_section{'sq_marker_'.$i};
		}
		else{
			return NOT_EQUAL_MARKER;
		}
		return EMPTY_MARKER if ($sq_markers{'sq_marker_'.$i} eq '');
		$sq_marker_names{$sq_section{'sq_marker_'.$i}}++;
		print "\tsq_marker_".$i ." = ".$sq_section{'sq_marker_'.$i}."\n";
		if (defined $sq_section{'sq_marker_'.$i.'_primer1'}){
			$sq_primer1{'sq_marker_'.$i} =
				$sq_section{'sq_marker_'.$i.'_primer1'};
		}
		else{
			return NOT_EQUAL_PRIMER;
		}
		if (defined $sq_section{'sq_marker_'.$i.'_primer2'}){
			$sq_primer2{'sq_marker_'.$i} =
				$sq_section{'sq_marker_'.$i.'_primer2'};
		}
		else{
			return NOT_EQUAL_PRIMER;
		}
		$msg = TAB."sq_marker_".$i."_primer1";
		print "\t$msg"," "x(25-length($msg));
		print "= ".$sq_primer1{'sq_marker_'.$i}."\n";
		$msg = TAB."sq_marker_".$i."_primer2";
		print "\t$msg"," "x(25-length($msg));
		print "= ".$sq_primer2{'sq_marker_'.$i}."\n";
		return EMPTY_PRIMER if ($sq_primer1{'sq_marker_'.$i} eq '');
		return EMPTY_PRIMER if ($sq_primer2{'sq_marker_'.$i} eq '');
		$sq_primer_names{$sq_section{'sq_marker_'.$i.'_primer1'}}++;
		$sq_primer_names{$sq_section{'sq_marker_'.$i.'_primer2'}}++;
	}
	my $marker_size = scalar keys %sq_marker_names;
	my $primer_size = scalar keys %sq_primer_names;
	print TAB. "$primer_size primer(s), $marker_size marker(s)\n";
	return NOT_EQUAL_MARKER if ($marker_size ne $sq_marker_no );
	# read all markers and related primers
	$msg = "Microsatellite markers";
	print  "\t$msg"," "x(25-length($msg)),"= ".$ms_marker_no."\n";
	for (my $i=1; $i<= $ms_marker_no; $i++){
		if (defined $ms_section{'ms_marker_'.$i}){
			$ms_markers{'ms_marker_'.$i} = $ms_section{'ms_marker_'.$i};
		}
		else{
			return NOT_EQUAL_MARKER;
		}
		return EMPTY_MARKER if ($ms_markers{'ms_marker_'.$i} eq '');
		$ms_marker_names{$ms_section{'ms_marker_'.$i}}++;
		print "\tms_marker_".$i ." = ".$ms_section{'ms_marker_'.$i}."\n";
		if (defined $ms_section{'ms_marker_'.$i.'_primer1'}){
			$ms_primer1{'ms_marker_'.$i} =
				$ms_section{'ms_marker_'.$i.'_primer1'};
		}
		else{
			return NOT_EQUAL_PRIMER;
		}
		if (defined $ms_section{'ms_marker_'.$i.'_primer2'}){
			$ms_primer2{'ms_marker_'.$i} =
				$ms_section{'ms_marker_'.$i.'_primer2'};
		}
		else{
			return NOT_EQUAL_PRIMER;
		}
		$msg = TAB. "ms_marker_".$i."_primer1";
		print "\t$msg"," "x(25-length($msg));
		print "= ".$ms_primer1{'ms_marker_'.$i}."\n";
		$msg = TAB. "ms_marker_".$i."_primer2";
		print "\t$msg"," "x(25-length($msg));
		print "= ".$ms_primer2{'ms_marker_'.$i}."\n";
		return EMPTY_PRIMER if ($ms_primer1{'ms_marker_'.$i} eq '');
		return EMPTY_PRIMER if ($ms_primer2{'ms_marker_'.$i} eq '');
		$ms_primer_names{$ms_section{'ms_marker_'.$i.'_primer1'}}++;
		$ms_primer_names{$ms_section{'ms_marker_'.$i.'_primer2'}}++;
	}
	$marker_size = scalar keys %ms_marker_names;
	$primer_size = scalar keys %ms_primer_names;
	print TAB. "$primer_size primer(s), $marker_size marker(s)\n";
	return NOT_EQUAL_MARKER if ($marker_size ne $ms_marker_no );
	# maximum number of sequences and microsatellites per sample
# 	$sq_max      = $global_section{'max_sq'};
# 	$ms_max      = $global_section{'max_ms'};
	print "\n";
	return NO_ERROR;
}

sub load_dummy{	
	print TAB . "DUMMIES loading\n";
	my $ins_cd = Apiis::DataBase::Record->new(tablename => 'codes',);
	my $ins_ct = Apiis::DataBase::Record->new(tablename => 'contacts',);
	my $ins_un = Apiis::DataBase::Record->new(tablename => 'unit',);
	my $ins_pt = Apiis::DataBase::Record->new(tablename => 'protocols',);
	my $ins_sp = Apiis::DataBase::Record->new(tablename => 'samples',);
	my $ins_de = Apiis::DataBase::Record->new(tablename => 'dna_extraction',);
	my $ins_st = Apiis::DataBase::Record->new(tablename => 'storage',);
	my $ins_ss = Apiis::DataBase::Record->new(tablename => 'storage_samples',);
	# vessel type
	$msg = "Inserting a dummy vessel type";
	print "\t$msg","."x(LEN-length($msg));
	$ins_cd->column('ext_code')->extdata(DUMMY);
	$ins_cd->column('class')->extdata("VESSEL_TYPE");
	$ins_cd->column('short_name')->extdata(DUMMY);
	$ins_cd->column('long_name')->extdata(DUMMY);
	$ins_cd->column('description')->extdata(DUMMY);
	$ins_cd->column('opening_dt')->extdata($apiis->now);
	$ins_cd->insert;	
	$ins_cd->check_status (die => 'ERR');
	print "yes\n";
	my $unload_sql  = "";
	my $cd_guid = $ins_cd->column('guid')->intdata;
	$unload_sql .= "DELETE FROM codes ";
	$unload_sql .= "WHERE guid = '$cd_guid';\n";
	# material type
	$msg = "Inserting a dummy material type";
	print "\t$msg","."x(LEN-length($msg));
	$ins_cd->column('ext_code')->extdata(DUMMY);
	$ins_cd->column('class')->extdata("MATERIAL_TYPE");
	$ins_cd->column('short_name')->extdata(DUMMY);
	$ins_cd->column('long_name')->extdata(DUMMY);
	$ins_cd->column('description')->extdata(DUMMY);
	$ins_cd->column('opening_dt')->extdata($apiis->now);
	$ins_cd->insert;
	$ins_cd->check_status (die => 'ERR');
	print "yes\n";
	$cd_guid = $ins_cd->column('guid')->intdata;
	$unload_sql .= "DELETE FROM codes ";
	$unload_sql .= "WHERE guid = '$cd_guid';\n";
	# contacts
	$msg = "Inserting a dummy contact";
	print "\t$msg","."x(LEN-length($msg));
	my $db_contact = next_id('contacts','db_contact');
	$ins_ct->column('db_contact')->extdata($db_contact);
	$ins_ct->column('second_name')->extdata(DUMMY);
	$ins_ct->column('opening_dt')->extdata($apiis->now);
	$ins_ct->insert;	
	$ins_ct->check_status (die => 'ERR');
	my $ct_guid = $ins_ct->column('guid')->intdata;
	print "yes\n";
	# insert into unit
	$msg = "Inserting a dummy user unit";
	print "\t$msg","."x(LEN-length($msg));
	my $db_unit = next_id('unit','db_unit');
	$ins_un->column('db_unit')->extdata($db_unit);
	$ins_un->column('db_contact')->extdata($db_contact);
	$ins_un->column('ext_unit')->extdata('USERS');
	$ins_un->column('ext_id')->extdata(DUMMY);
	$ins_un->column('opening_dt')->extdata($apiis->now);
	$ins_un->insert;
	$ins_un->check_status (die => 'ERR');
	my $un_guid = $ins_un->column('guid')->intdata;
	print "yes\n";
	$unload_sql .= "DELETE FROM contacts ";
	$unload_sql .= "WHERE guid = '$ct_guid';\n";
	$unload_sql .= "DELETE FROM unit ";
	$unload_sql .= "WHERE guid = '$un_guid';\n";
	# protocols
	$msg = "Inserting a dummy file";
	print "\t$msg","."x(LEN-length($msg));
	my $filename  = DUMMY_PROTOCOL;
	my $blobid = load_blob($path."/protocols/".$filename);
	print "yes\n"  if (defined $blobid);
	print "no \n"  if (not defined $blobid);
	$msg = "Inserting a dummy protocol";
	print "\t$msg","."x(LEN-length($msg));
	if (defined $blobid){
		$ins_pt->column('protocol_id')->extdata($blobid);
		$ins_pt->column('protocol_name')->extdata(DUMMY);
		$ins_pt->column('db_protocol_type')->extdata('DNA extraction');
		$ins_pt->column('comment')->extdata(DUMMY);
		$ins_pt->insert;
		$ins_pt->check_status (die => 'ERR');
		print "yes\n";
		$unload_sql .= "DELETE FROM protocols ";
		$unload_sql .= "WHERE protocol_id = '$blobid';\n";
		$unload_sql .= "DELETE FROM blobs ";
		$unload_sql .= "WHERE blob_id = '$blobid';\n";
	}	else{	print "No\n"; }
	
	if (uc($sq_rawdata_load) eq 'YES'){
		$msg = "Inserting dummy DNA samples for loading sequences";
		print "\t$msg","\n";
		my $filename = "datafile_sq.". get_file_extension($sq_datafile_format);
		$sq_file_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");
		for (my $i=2;$i<(2+$record_no);$i++){
			my $ext_dna = read_data($sq_file_ref,$i,1,'yes');
			if ($ext_dna eq ""){
            	next;
			}
			if ($ext_dna eq ""){
				$msg =  TAB. "Inserting $ext_dna(row $i)";
				print "\t$msg","."x(LEN-length($msg));
				print "error\n";				
				return DATA_NOT_FOUND;	
			}
			else {
				$msg =  TAB. "Inserting $ext_dna";
				print "\t$msg","."x(LEN-length($msg));
			}
			my $dna_id	= next_id('samples','sample_id');
		   	$ins_sp->column('sample_id')->intdata($dna_id);
		    $ins_sp->column('sample_id')->encoded(1);
		    $ins_sp->column('ext_sample')->extdata($ext_dna);
		    $ins_sp->column('db_vessel_type')->extdata(DUMMY);
		    $ins_sp->column('db_material_type')->extdata(DUMMY);
		    $ins_sp->column('db_user')->extdata($db_contact);
		    $ins_sp->column('protocol_id')->intdata($blobid);
		    $ins_sp->column('protocol_id')->encoded(1);
		    $ins_sp->column('comment')->extdata(DUMMY);
		    $ins_sp->insert;
		    $ins_sp->check_status (die => 'ERR');
		    $dummy_dna_ids{$ext_dna} = $dna_id;
		    print "ok\n";
			my $sp_guid = $ins_sp->column('guid')->intdata;
			$unload_sql .= "DELETE FROM samples ";
		    $unload_sql .= "WHERE guid = '$sp_guid';\n";
		    # dna extraction
			$ins_de->column('dna_id')->intdata($dna_id);
		    $ins_de->column('dna_id')->encoded(1);		    
			$ins_de->column('sample_id')->intdata($dna_id);
		    $ins_de->column('sample_id')->encoded(1);
		    $ins_de->column('project_id')->intdata($project_id);
		    $ins_de->column('project_id')->encoded(1);
		    $ins_de->column('reused')->extdata('N');
		    $ins_de->insert;
		    $ins_de->check_status (die => 'ERR');
			my $de_guid = $ins_de->column('guid')->intdata;
			$unload_sql .= "DELETE FROM dna_extraction ";
		    $unload_sql .= "WHERE guid = '$de_guid';\n";
		}
	}
	elsif (uc($ms_rawdata_load) eq 'YES'){
		$msg = "Inserting dummy DNA samples for loading microsatellites";
		print "\t$msg","\n";
		my $filename = "final_ms/$ms_datafile";
		$ms_file_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");
		for (my $i=2;$i<(2+$record_no);$i++){
			my $ext_dna = read_data($ms_file_ref,$i,1,'yes');
			if ($ext_dna eq ""){
				$msg =  TAB. "Inserting $ext_dna (row $i)";
				print "\t$msg","."x(LEN-length($msg));
				print "error\n";				
				return DATA_NOT_FOUND;	
			} else {
				$msg =  TAB. "Inserting $ext_dna";
				print "\t$msg","."x(LEN-length($msg));
			}
			$msg =  TAB. "Inserting $ext_dna";
			print "\t$msg","."x(LEN-length($msg));
			my $dna_id	= next_id('samples','sample_id');
		   	$ins_sp->column('sample_id')->intdata($dna_id);
		    $ins_sp->column('sample_id')->encoded(1);
		    $ins_sp->column('ext_sample')->extdata($ext_dna);
		    $ins_sp->column('db_vessel_type')->extdata(DUMMY);
		    $ins_sp->column('db_material_type')->extdata(DUMMY);
		    $ins_sp->column('db_user')->extdata($db_contact);
		    $ins_sp->column('protocol_id')->intdata($blobid);
		    $ins_sp->column('protocol_id')->encoded(1);
		    $ins_sp->column('comment')->extdata(DUMMY);
		    $ins_sp->insert;
		    $ins_sp->check_status (die => 'ERR');
		    $dummy_dna_ids{$ext_dna} = $dna_id;
		    print "ok\n";
			my $sp_guid = $ins_sp->column('guid')->intdata;
			$unload_sql .= "DELETE FROM samples ";
		    $unload_sql .= "WHERE guid = '$sp_guid';\n";
		}
	}
	print UNLOAD_FILE $unload_sql;
	return NO_ERROR;
}

sub get_samples {
	return $flag if (load_primers()	   						 ne NO_ERROR);
	return $flag if (load_markers()	    					 ne NO_ERROR);
	return $flag if (load_codes(3,1,'SPECIES','no')			 ne NO_ERROR);
	return $flag if (load_codes(3,2,'BREED','unknown_breed') ne NO_ERROR);
	return $flag if (load_codes(3,5,'SEX','u')     ne NO_ERROR);
	return $flag if (load_codes(3,9,'MATERIAL_TYPE','no')	 ne NO_ERROR);
	return $flag if (load_vessel_types()					 ne NO_ERROR);
	return $flag if (load_users()         					 ne NO_ERROR);
	return $flag if (load_protocols(3,11,"Sampling")		 ne NO_ERROR);
	return $flag if (load_protocols(3,23,"DNA extraction")	 ne NO_ERROR);
	return $flag if (load_pictures(3,6)	    				 ne NO_ERROR);
	return $flag if (load_individuals(3,3)    				 ne NO_ERROR);
	return $flag if (load_storage(3,15,'SAMPLE')			 ne NO_ERROR);
	return $flag if (load_storage(3,29,'DNA')				 ne NO_ERROR);
	return $flag if (load_samples(3,8)	    				 ne NO_ERROR);
	return $flag if (load_dna(3,21)							 ne NO_ERROR);
	return $flag;
}

sub load_sq_rawdata {
	print TAB . "Loading raw files for DNA sequencing ...\n";
	my (%dirs);
	my $row = 2;
	while (1){
		my $dir = read_data($sq_rawfile_ref,$row,1,'yes');
		last if ($dir eq '');
		# make a tar for each dir
		my $msg = "Reading from directory .../raw_sq/$dir/";
		print "\t$msg", "\n";
		opendir MYDIR, "$path/raw_sq/$dir/";
		my $tar = Archive::Tar->new;
		chdir "$path/raw_sq";
		while ($_ = readdir (MYDIR)){
			if (($_ ne ".") and ($_ ne "..")){
				print "\t".TAB. $_, "."x(LEN-length($_)-length(TAB)), "ok\n";
				$tar->add_files("$dir/$_");
			}
		}
		chdir OUTPUT_DIR;
		$tar->write("$dir.tar");
		$msg = "Creating $dir.tar and sending to the database";
		print "\t$msg", "."x(LEN-length($msg));
		my $blobid = load_blob(OUTPUT_DIR."/$dir.tar");
		print "yes\n" if (defined $blobid);
		print "no \n" if (not defined $blobid);
		my $unload_sql  = "";
		if (defined $blobid){
			$raw_files{$dir} = $blobid;
			$unload_sql .= "DELETE FROM blobs ";
			$unload_sql .= "WHERE blob_id = '$blobid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			$flag = FILE_NOT_FOUND;
			return $flag;
		}
		# read date and marker
		my $date   = read_data($sq_rawfile_ref,$row,2,"yes");
		my $marker = read_data($sq_rawfile_ref,$row,3,"yes");
		my $col = 4;
		# insert one record to the table 'Electrophoresis'
		my $ins_el=Apiis::DataBase::Record->new(tablename =>'electrophoresises',);
		my %el_ids; # ext_electrophoresis -> electrophoresis_id
		my $purpose = 'S';
		my $method	= 'CE';
		my $elecs_ref =  \%sq_elecs;
		my $ext_electrophoresis = $dir;
		my $product_id	= $raw_files{$dir};
		my $el_id = next_id('electrophoresises','electrophoresis_id');
		$el_ids{$ext_electrophoresis} = $el_id;
		$msg = "Inserting $marker/$ext_electrophoresis";
		print TAB.TAB."$msg";
		$ins_el->column('electrophoresis_id')->intdata($el_id);
		$ins_el->column('electrophoresis_id')->encoded(1);
		$ins_el->column('ext_electrophoresis')->extdata($ext_electrophoresis);
		if ($date ne ""){
			$ins_el->column('electrophoresis_dt')->extdata($date);
		}
		$ins_el->column('project_id')->extdata($project_id);
		$ins_el->column('purpose')->extdata($purpose);
		$ins_el->column('db_method')->extdata($method);
		$ins_el->column('product_id')->intdata($product_id);
		$ins_el->column('product_id')->encoded(1);
		$ins_el->column('comment')->extdata(COMMENT);
		$ins_el->insert;
		$ins_el->check_status (die => 'ERR');
		print " which was linked to ";
		my $el_guid = $ins_el->column('guid')->intdata;
		$unload_sql  = "";
		$unload_sql .= "DELETE FROM electrophoresises ";
		$unload_sql .= "WHERE guid = '$el_guid';\n";
		print UNLOAD_FILE $unload_sql;
		while (1){
			my $dna = read_data($sq_rawfile_ref,$row,$col,"yes");
			last if ($dna eq '');
			# insert one record for each DNA ID found
			my $ins_an=Apiis::DataBase::Record->new(tablename =>'analyzed_samples', );
			$el_id = $el_ids{$ext_electrophoresis};
		    my $marker_id  = $marker_ids{$marker};
			my $dna_id	   = $dna_ids{$dna};
			$ins_an->column('electrophoresis_id')->intdata($el_id);
			$ins_an->column('electrophoresis_id')->encoded(1);
			$ins_an->column('marker_id')->intdata($marker_id);
			$ins_an->column('marker_id')->encoded(1);
			$ins_an->column('dna_id')->extdata($dna_id);
			$ins_an->insert;
			$ins_an->check_status (die => 'ERR');
			print $dna. "  ";
			my $an_guid = $ins_an->column('guid')->intdata;
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM analyzed_samples ";
			$unload_sql .= "WHERE guid = '$an_guid';\n";
			print UNLOAD_FILE $unload_sql;
			$col++;
		}
		print "\n";
		$row++;
	}
	return NO_ERROR;
}

sub load_ms_rawdata {
	print TAB . "Loading raw files for microsatellites ...\n";
	my (%dirs);
	my %el_ids; # ext_electrophoresis -> electrophoresis_id
	my $row = 2;
    # read all folders which consist of raw files
	while (1){
		my $dna = read_data($ms_rawfile_ref,$row,1,'yes');
		last if ($dna eq '');
		my $col = 2;
		my $dir;
		my $date;
		for (my $i=0;$i<$ms_marker_no;$i++){
			$dir = read_data($ms_rawfile_ref,$row,$col,'yes');
			$date = read_data($ms_rawfile_ref,$row,$col+1,'yes');
			$dirs{$dir} = $date;
 			$col = $col + 2;
		}
		$row++;
	}
	# make a tar for each folder and send to database
	for my $key ( keys %dirs ) {
		my $msg = "Reading from directory .../raw_ms/raw_data/$key/";
		print "\t$msg", "\n";
		opendir MYDIR, "$path/raw_ms/raw_data/$key/";
		my $tar = Archive::Tar->new;
		chdir "$path/raw_ms/raw_data/";
		while ($_ = readdir (MYDIR)){
			if (($_ ne ".") and ($_ ne "..")){
				print "\t".TAB. $_, "."x(LEN-length($_)-length(TAB)), "ok\n";
				$tar->add_files("$key/$_");
			}
		}
		chdir OUTPUT_DIR;
		$key =~ s/\//-/;
		$tar->write("$key.tar");
		$msg = "Creating $key.tar and sending to DB";
		print "\t$msg", "."x(LEN-length($msg));
		my $blobid = load_blob(OUTPUT_DIR."/$key.tar");
		print "yes\n" if (defined $blobid);
		print "no \n" if (not defined $blobid);
		my $unload_sql  = "";
		if (defined $blobid){
			$raw_files{$key} = $blobid;
			$unload_sql .= "DELETE FROM blobs ";
			$unload_sql .= "WHERE blob_id = '$blobid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			$flag = FILE_NOT_FOUND;
			return $flag;
		}
	}
	# insert a record to 'electrophoresises' for each dir
	for my $key ( keys %dirs ) {
		my $ins_el=Apiis::DataBase::Record->new(tablename =>'electrophoresises',);
		my $purpose = 'G';
		my $method	 = 'CE';
		my $elec_date = $dirs{$key};
		my $elecs_ref =  \%sq_elecs;
		$key =~ s/\//-/;
		my $ext_electrophoresis = $key;
		my $product_id	= $raw_files{$key};
		my $el_id = next_id('electrophoresises','electrophoresis_id');
		$el_ids{$ext_electrophoresis} = $el_id;
		$msg = "Inserting $ext_electrophoresis";
		print "\t$msg","."x(LEN-length($msg));
		$ins_el->column('electrophoresis_id')->intdata($el_id);
		$ins_el->column('electrophoresis_id')->encoded(1);
		$ins_el->column('ext_electrophoresis')->extdata($ext_electrophoresis);
		if ($elec_date ne ""){
			$ins_el->column('electrophoresis_dt')->extdata($elec_date);
		}
		$ins_el->column('project_id')->extdata($project_id);
		$ins_el->column('purpose')->extdata($purpose);
		$ins_el->column('db_method')->extdata($method);
		$ins_el->column('product_id')->intdata($product_id);
		$ins_el->column('product_id')->encoded(1);
		$ins_el->column('comment')->extdata(COMMENT);
		$ins_el->insert;
		$ins_el->check_status (die => 'ERR');
		print "yes\n";
		my $el_guid = $ins_el->column('guid')->intdata;
		my $unload_sql  = "";
		$unload_sql .= "DELETE FROM electrophoresises ";
		$unload_sql .= "WHERE guid = '$el_guid';\n";
		print UNLOAD_FILE $unload_sql;
	}
	# insert one record for each DNA ID found
	$row = 2;
	while (1){
		my $dna = read_data($ms_rawfile_ref,$row,1,'yes');
		last if ($dna eq '');
		my $col = 2;
		my $marker;
		my $date;
		my $dir;
		for (my $i=0;$i<$ms_marker_no;$i++){
			$dir    = read_data($ms_rawfile_ref,$row,$col,'yes');
			$marker = read_data($ms_rawfile_ref,1,$col,'yes');
			$date   = read_data($ms_rawfile_ref,$row,$col+1,'yes');
			$dir =~ s/\//-/;
			my $ext_electrophoresis = $dir;
			my $ins_an=Apiis::DataBase::Record->new(tablename =>'analyzed_samples', );
			my $el_id = $el_ids{$ext_electrophoresis};
		    my $marker_id  = $marker_ids{$marker};
			my $dna_id	   = $dna_ids{$dna};
			$ins_an->column('electrophoresis_id')->intdata($el_id);
			$ins_an->column('electrophoresis_id')->encoded(1);
			$ins_an->column('marker_id')->intdata($marker_id);
			$ins_an->column('marker_id')->encoded(1);
			$ins_an->column('dna_id')->extdata($dna_id);
			$ins_an->insert;
			$ins_an->check_status (die => 'ERR');
			$msg = $dna. " --> " . "$dir.tar";
            print "\t$msg","."x(LEN-length($msg)) . "yes\n";
			my $an_guid = $ins_an->column('guid')->intdata;
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM analyzed_samples ";
			$unload_sql .= "WHERE guid = '$an_guid';\n";
			print UNLOAD_FILE $unload_sql;
			$col = $col + 2;
		}
		$row++;
	}
	return NO_ERROR;
}

sub load_project{
	my $ins_pj = Apiis::DataBase::Record->new(tablename =>'projects',);
	print TAB . "PROJECT loading\n";
	if (check_existing('projects','project_name',$project_name,'') eq 'false'){
		$msg = "Inserting '$project_name'";
		print "\t$msg","."x(LEN-length($msg));
		$ins_pj->column('project_name')->extdata($project_name);
		$ins_pj->column('keyword')->extdata($keyword);
		$ins_pj->column('description')->extdata($description);
		$ins_pj->insert;
		$ins_pj->check_status (die => 'ERR');
		print "yes\n";
		$project_id =  $ins_pj->column('project_id')->intdata;
		print UNLOAD_FILE "-- Rolls back PROJECTS \n";
		my $pj_guid = $ins_pj->column('guid')->intdata;
		my $unload_sql  = "";
		$unload_sql .= "DELETE FROM projects ";
		$unload_sql .= "WHERE guid = '$pj_guid';\n";
		print UNLOAD_FILE $unload_sql;
	}
	else{
		$flag = DUPLICATED_PROJECT;
	}
	return $flag;
}

sub load_codes{
	my ($row,$col,$class_name,$convert) = @_;
	print TAB . "$class_name loading\n";
	my %data_hash;
 	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		if($convert ne 'no'){
			if (defined $key){
				$key = $convert if ($key eq "");
			}
			else{
				$key = $convert;
			}
		}
		if ($key eq ""){
			$flag = EMPTY_SPECIES	if ($class_name eq 'SPECIES');
			$flag = EMPTY_MATERIAL	if ($class_name eq 'MATERIAL_TYPE');
			return $flag;
		}
		$data_hash{$key}++;
	}
	print UNLOAD_FILE "-- Rolls back $class_name \n";
	my $ins_codes=Apiis::DataBase::Record->new( tablename => 'codes', );
	for my $key ( keys %data_hash ) {
		if(check_existing('codes','ext_code',$key,$class_name) eq 'false'){
			$msg = "Inserting '$key'";
 			$msg = Encode::encode_utf8($msg);
			print "\t$msg","."x(LEN-length($msg));
			$ins_codes->column('ext_code')->extdata($key);
			$ins_codes->column('class')->extdata($class_name);
			$ins_codes->column('short_name')->extdata($key);
			$ins_codes->column('long_name')->extdata($key);
			$ins_codes->column('description')->extdata(COMMENT);
			$ins_codes->column('opening_dt')->extdata($apiis->now);
			$ins_codes->insert;	
			$ins_codes->check_status (die => 'ERR');
			my $cd_guid = $ins_codes->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM codes ";
			$unload_sql .= "WHERE guid = '$cd_guid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			$msg = "$class_name '$key' exists, so skip inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			print "\n";
		}
    }
    return NO_ERROR;
}

sub load_vessel_types{
	my ($row,$col) = (3,13);
	my $class_name = "VESSEL_TYPE";
	print TAB . "$class_name loading\n";
	my %data_hash;
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$key = convert_dummy('VESSEL',$key);
		$data_hash{$key}++;
	}
	$col = 27;
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$key = convert_dummy('VESSEL',$key);
# 		}
		$data_hash{$key}++;
	}
	print UNLOAD_FILE "-- Rolls back $class_name \n";
	my $ins_codes=Apiis::DataBase::Record->new( tablename => 'codes', );
	for my $key ( keys %data_hash ) {
		if(check_existing('codes','ext_code',$key,$class_name) eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_codes->column('ext_code')->extdata($key);
			$ins_codes->column('class')->extdata($class_name);
			$ins_codes->column('short_name')->extdata($key);
			$ins_codes->column('long_name')->extdata($key);
			$ins_codes->column('description')->extdata(COMMENT);
			$ins_codes->column('opening_dt')->extdata($apiis->now);
			$ins_codes->insert;	
			$ins_codes->check_status (die => 'ERR');
			my $cd_guid = $ins_codes->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM codes ";
			$unload_sql .= "WHERE guid = '$cd_guid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			$msg = "$class_name '$key' exists, so skip inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			print "\n";
		}
    }
    return NO_ERROR;
}

sub load_users{
	my ($row,$col) = (3,20);
	print TAB . "USERS loading\n";
	# read all sample collectors in column 20
	my %collectors;
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$key = trimstr($key) if defined $key;
		$key = convert_dummy('CONTACT',$key);
		$collectors{$key}++;
	}
	$col = 34;
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$key = trimstr($key) if defined $key;
		$key = convert_dummy('CONTACT',$key);
		$collectors{$key}++;
	}
    # load new collectors to Contacts where ext_unit = USER
	print UNLOAD_FILE "-- Rolls back USERS \n";
	my $ins_contacts=Apiis::DataBase::Record->new( tablename => 'contacts', );
	my $ins_unit    =Apiis::DataBase::Record->new( tablename => 'unit', );
	for my $key ( keys %collectors ) {
		if (check_existing('unit','ext_id',$key,'') eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			my $db_contact = next_id('contacts','db_contact');
			# insert into cotacts
			$ins_contacts->column('db_contact')->extdata($db_contact);
			$ins_contacts->column('second_name')->extdata($key);
			$ins_contacts->column('first_name')->extdata($key);
			$ins_contacts->column('opening_dt')->extdata($apiis->now);
			$ins_contacts->insert;	
			$ins_contacts->check_status (die => 'ERR');
			my $ct_guid = $ins_contacts->column('guid')->intdata;
			# insert into unit
			my $db_unit = next_id('unit','db_unit');
			$ins_unit->column('db_unit')->extdata($db_unit);
			$ins_unit->column('db_contact')->extdata($db_contact);
			$ins_unit->column('ext_unit')->extdata('USERS');
			$ins_unit->column('ext_id')->extdata($key);
			$ins_unit->column('opening_dt')->extdata($apiis->now);
			$ins_unit->insert;	
			$ins_unit->check_status (die => 'ERR');
			my $un_guid = $ins_unit->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM contacts ";
			$unload_sql .= "WHERE guid = '$ct_guid';\n";
			$unload_sql .= "DELETE FROM unit ";
			$unload_sql .= "WHERE guid = '$un_guid';\n";
			print UNLOAD_FILE $unload_sql;
			$users{$key} = $db_contact;
		}
		else{
			my $db_contact = get_uniqueID('unit','db_contact','ext_id',$key);
			$users{$key} = $db_contact;
			print "\tUser '$key' exists, so skip inserting '$key'\n";
		}
    }
    return NO_ERROR;
}

sub load_protocols{
	my ($row,$col,$type) = @_;
	print TAB . "PROTOCOLS loading\n";
    my (%pts,%pt_names);
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$key = trimstr($key) if defined $key;
		$key = convert_dummy('PROTOCOL',$key);
# 		if (defined $key){
# 			if ($key eq ""){
# 				$key = "dummy_protocol.txt";	
# 			}
# 		}
# 		else{
# 			$key = "dummy_protocol.txt";
# 		}
		$pt_names{substr($key,0,index($key,"."))}++;
		$pts{$key}++;
    }
   	print UNLOAD_FILE "-- Rolls back $type PROTOCOLS \n";
	my $ins_pt = Apiis::DataBase::Record->new( tablename => 'protocols', );
	for my $filename ( keys %pts ) {
		my $basename = substr($filename,0,index($filename,"."));
		if (check_existing('protocols','protocol_name',
		    $basename,'$type') eq 'false'){
			$msg = "Loading data for a $type protocol";
			print "\t$msg","."x(LEN-length($msg));
			my $blobid = load_blob($path."/protocols/".$filename);
			if (not defined $blobid){
				print "no \n";
				print "\tProtocol: ../protocols/".$filename." is NOT exist";
				$flag = FILE_NOT_FOUND;
				return $flag;
			} else {
                print "yes\n";
			}
			$msg = "Inserting '$filename'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_pt->column('protocol_id')->extdata($blobid);
			$ins_pt->column('protocol_name')->extdata($basename);
			$ins_pt->column('db_protocol_type')->extdata($type);
			$ins_pt->column('comment')->extdata(COMMENT);
			$ins_pt->insert;
			$ins_pt->check_status (die => 'ERR');
			print "yes\n";
			if (defined $blobid){
                if ($type eq 'Sampling'){
                    $filename = 'sampling__'.$filename;
                }
                elsif ($type eq 'DNA extraction') {
                    $filename = 'dna_extraction__'.$filename;
                }   
				$protocols{$filename} = $blobid;
#                 print "LOAD: ".$filename."->". $protocols{$filename}."\n";
				my $unload_sql  = "";
				$unload_sql .= "DELETE FROM protocols ";
				$unload_sql .= "WHERE protocol_id = '$blobid';\n";
				$unload_sql .= "DELETE FROM blobs ";
				$unload_sql .= "WHERE blob_id = '$blobid';\n";
				print UNLOAD_FILE $unload_sql;
			}
		}
		else{
			print "\tProtocol '$filename' exists, ";
            print "so skip inserting '$filename'\n";
		}
    }
    return NO_ERROR;
}

sub load_primers{
	print TAB . "PRIMERS loading\n";
	my $primer_size = scalar keys %sq_primer_names;
	print UNLOAD_FILE "-- Rolls back PRIMERS \n";
	my $ins_codes=Apiis::DataBase::Record->new( tablename => 'codes', );
	for my $key ( keys %sq_primer_names ) {
		if (check_existing('codes','ext_code',$key,'PRIMER') eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_codes->column('ext_code')->extdata($key);
			$ins_codes->column('class')->extdata('PRIMER');
			$ins_codes->column('short_name')->extdata($key);
			$ins_codes->column('long_name')->extdata($key);
			$ins_codes->column('opening_dt')->extdata($apiis->now);
			$ins_codes->insert;	
			$ins_codes->check_status (die => 'ERR');
			my $cd_guid = $ins_codes->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM codes ";
			$unload_sql .= "WHERE guid = '$cd_guid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			print "\tPrimer '$key' exists, so skip inserting '$key'\n";
		}
    }
	for my $key ( keys %ms_primer_names ) {
		if (check_existing('codes','ext_code',$key,'PRIMER') eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_codes->column('ext_code')->extdata($key);
			$ins_codes->column('class')->extdata('PRIMER');
			$ins_codes->column('short_name')->extdata($key);
			$ins_codes->column('long_name')->extdata($key);
			$ins_codes->column('opening_dt')->extdata($apiis->now);
			$ins_codes->insert;	
			$ins_codes->check_status (die => 'ERR');
			my $cd_guid = $ins_codes->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM codes ";
			$unload_sql .= "WHERE guid = '$cd_guid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			print "\tPrimer '$key' exists, so skip inserting '$key'\n";
		}
    }
    return NO_ERROR;
}

sub load_markers{
	print TAB . "MARKERS loading\n";
	print UNLOAD_FILE "-- Rolls back MARKERS \n";
	my $ins_mk=Apiis::DataBase::Record->new( tablename => 'markers', );
	for my $key ( keys %sq_markers) {
		if (check_existing('markers','marker_name',$sq_markers{$key},'') eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_mk->column('marker_name')->extdata($sq_markers{$key});
			$ins_mk->column('db_marker_type')->extdata('Sequencing');
			$ins_mk->column('db_primer1')->extdata($sq_primer1{$key});
			$ins_mk->column('db_primer2')->extdata($sq_primer2{$key});
			$ins_mk->insert;	
			$ins_mk->check_status (die => 'ERR');
            $marker_ids{$sq_markers{$key}}= 
            	$ins_mk->column('marker_id')->intdata;
            my $mk_guid = $ins_mk->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM markers ";
			$unload_sql .= "WHERE guid = '$mk_guid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			print "\tMarker '$key' exists, so skip inserting '$key'\n";
            my $marker_id =
            get_uniqueID('markers','marker_id','marker_name',$sq_markers{$key});
           # print "marker_id=$marker_id\n";
            $marker_ids{$sq_markers{$key}}=  $marker_id;                                           
        }
    }
	for my $key ( keys %ms_markers) {
		if (check_existing('markers','marker_name',$ms_markers{$key},'') eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_mk->column('marker_name')->extdata($ms_markers{$key});
			$ins_mk->column('db_marker_type')->extdata('Microsatellite');
			$ins_mk->column('db_primer1')->extdata($ms_primer1{$key});
			$ins_mk->column('db_primer2')->extdata($ms_primer2{$key});
			$ins_mk->insert;	
			$ins_mk->check_status (die => 'ERR');
            $marker_ids{$ms_markers{$key}}= 
            	$ins_mk->column('marker_id')->intdata;
            my $mk_guid = $ins_mk->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM markers ";
			$unload_sql .= "WHERE guid = '$mk_guid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			print "\tMarker '$key' exists, so skip inserting '$key'\n";
            my $marker_id =
            get_uniqueID('markers','marker_id','marker_name',$ms_markers{$key});
            # print "marker_id=$marker_id\n";
            $marker_ids{$ms_markers{$key}}= $marker_id;
		}
    }
    return NO_ERROR;
}

sub load_pictures{
	my ($row,$col) = @_;
	print TAB . "PICTURES loading\n";
	my (%photo,%photo_names);
    for (my $i=$row;$i<($row+$record_no);$i++){
		my $key = read_data($sp_file_ref,$i,$col,'yes');
		$key = trimstr($key) if defined $key;
		if ((defined $key) and ($key ne "")){
			$photo_names{substr($key,0,index($key,"."))}++;
			$photo{$key}++;
		}
    }
	print UNLOAD_FILE "-- Rolls back PICTURES \n";
	for my $filename ( keys %photo ) {
		my $basename = substr($filename,0,index($filename,"."));
		$msg = "Inserting '$filename'";
		print "\t$msg","."x(LEN-length($msg));		
		my $blobid = load_blob($path."/pictures/".$filename);
		if (not defined $blobid){
			print "no \n";
			print "\tPicture: ../pictures/".$filename." is NOT exist";
			$flag = FILE_NOT_FOUND;
			return $flag;
		} else {
              print "yes\n";
		}
		my $unload_sql  = "";
		if (defined $blobid){
			$pictures{$filename} = $blobid;
			$unload_sql .= "DELETE FROM blobs ";
			$unload_sql .= "WHERE blob_id = '$blobid';\n";
			print UNLOAD_FILE $unload_sql;
		}
		else{
			$flag = FILE_NOT_FOUND;
			return $flag;
		}
    }
    return NO_ERROR;
}

sub load_individuals{
	my ($row,$col) = @_;
	print TAB . "INDIVIDUALS loading\n";
	my %individuals;
	print UNLOAD_FILE "-- Rolls back ORGANISMS \n";
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $ins_og=Apiis::DataBase::Record->new( tablename => 'organisms', );	
		my $ins_tf=Apiis::DataBase::Record->new( tablename => 'transfer', );
		my $ext_organism 	= read_data($sp_file_ref,$i,$col,'yes');
		my $db_donor_type 	= "Animals";
		my $db_species 		= read_data($sp_file_ref,$i,1,'yes');
		my $db_breed 		= convert_dummy('BREED',read_data($sp_file_ref,$i,2,'yes'));
# 		my $latitude 		= read_data($sp_file_ref,$i,7,'yes'); # need to calculate
# 		my $longitude 		= read_data($sp_file_ref,$i,7,'yes'); # need to calculate
		my $birth_dt		= read_data($sp_file_ref,$i,$col+1,'yes');
		my $db_sex 			= convert_dummy('SEX',read_data($sp_file_ref,$i,$col+2,'yes'));
		my $filename		= read_data($sp_file_ref,$i,$col+3,'yes');
		my $db_organism		= $apiis->DataBase->seq_next_val('seq_transfer__db_organism');
		if(	check_existing('transfer','ext_organism',$ext_organism,'') eq 'false'){
  		  if (not defined $individuals{$ext_organism}) {
			$individuals{$ext_organism}++;	  
			$msg = "Inserting '$ext_organism'";
			print "\t$msg","."x(LEN-length($msg));
			$ins_tf->column('db_organism')->intdata($db_organism);
			$ins_tf->column('db_organism')->encoded(1);
			$ins_tf->column('ext_organism')->extdata($ext_organism);
			$ins_tf->column('db_unit')->extdata("ORGANISMS","ID");
			$ins_tf->column('opening_dt')->extdata($apiis->now);
			$ins_tf->insert;
			$ins_tf->check_status (die => 'ERR');
			my $tf_guid = $ins_tf->column('guid')->intdata;
			$ins_og->column('db_organism')->intdata($db_organism);
			$ins_og->column('db_organism')->encoded(1);
			$ins_og->column('db_sire')->extdata("ORGANISMS","ID","unknown_sire");
			$ins_og->column('db_dam')->extdata("ORGANISMS","ID","unknown_dam");
			$ins_og->column('db_donor_type')->extdata($db_donor_type);
			$ins_og->column('db_species')->extdata($db_species);
			$ins_og->column('db_breed')->extdata($db_breed);
			$ins_og->column('db_sex')->extdata($db_sex);
			$ins_og->column('birth_dt')->extdata($birth_dt);
			if ($filename ne ""){
				$ins_og->column('db_photo')->extdata($pictures{$filename});
			}
			$ins_og->insert;
			$ins_og->check_status (die => 'ERR');

			my $og_guid = $ins_og->column('guid')->intdata;
			print "yes\n";
			my $unload_sql  = "";
			$unload_sql .= "DELETE FROM organisms ";
			$unload_sql .= "WHERE guid = '$og_guid';\n";
			$unload_sql .= "DELETE FROM transfer ";
			$unload_sql .= "WHERE guid = '$tf_guid';\n";
			print UNLOAD_FILE $unload_sql;	
		  }
		}
		else{
			print "\tOrganism '$ext_organism' exists, ";
            print "so skip inserting '$ext_organism'\n";
		}
	}
    return NO_ERROR;
}

sub load_samples{
	my ($row,$col) = @_;
	print TAB . "SAMPLES loading\n";
	# read all samples from column 8 to column 20
	my %organisms;
	print UNLOAD_FILE "-- Rolls back SAMPLES \n";
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $ins_ss=Apiis::DataBase::Record->new( tablename =>'storage_samples',);
		my $ins_sp=Apiis::DataBase::Record->new( tablename =>'samples', );
		my $ins_sc=Apiis::DataBase::Record->new( tablename =>'sample_collection',);
		my $sample_id = next_id('samples','sample_id');
		my $ext_sample		= read_data($sp_file_ref,$i,$col,'yes');
		my $db_organism		= read_data($sp_file_ref,$i,3,'yes');
		my $db_vessel_type	= read_data($sp_file_ref,$i,13,'yes');
		$db_vessel_type = convert_dummy('VESSEL',$db_vessel_type);
		my $db_material_type= read_data($sp_file_ref,$i,9,'yes');
		my $contact_id		= read_data($sp_file_ref,$i,20,'yes');
		$contact_id = convert_dummy('CONTACT',$contact_id);
		my $db_contact		= $users{$contact_id};
		my $no_units		= read_data($sp_file_ref,$i,12,'yes');
		my $protocol_file	= read_data($sp_file_ref,$i,11,'yes');
		$protocol_file = trimstr($protocol_file) if defined $protocol_file;
		$protocol_file = convert_dummy('PROTOCOL', $protocol_file);
       	$protocol_file = 'sampling__'.$protocol_file;
		my $protocol_id;
		$protocol_id = $protocols{$protocol_file} if ($protocol_file ne "");
		my $production_dt	= read_data($sp_file_ref,$i,10,'no');
		my $storage_dt		= read_data($sp_file_ref,$i,14,'no');
		my $key1 			= read_data($sp_file_ref,$i,15,'yes');
		my $key2 			= read_data($sp_file_ref,$i,16,'yes');
		my $key3 			= read_data($sp_file_ref,$i,17,'yes');
		my $key4 			= read_data($sp_file_ref,$i,18,'yes');
		my $key5 			= read_data($sp_file_ref,$i,19,'yes');
		$key1 = convert_dummy('STORAGE',$key1);
		$key2 = convert_dummy('STORAGE',$key2);
		$key3 = convert_dummy('STORAGE',$key3);
		$key4 = convert_dummy('STORAGE',$key4);
		$key5 = convert_dummy('STORAGE',$key5);
		my $loc = "::".$key1."::".$key2."::".$key3."::".$key4."::".$key5;
		my $storage_id		= $sample_storages{$loc};
		if(check_existing('samples','ext_sample',$ext_sample,'') eq 'false'){
			$msg = "Inserting '$ext_sample' ..";
			print "\t$msg";
			$ins_sp->column('sample_id')->intdata($sample_id);
			$ins_sp->column('sample_id')->encoded(1);
			$ins_sp->column('ext_sample')->extdata($ext_sample);
			$ins_sp->column('db_organism')->extdata("ORGANISMS","ID",$db_organism);
			$ins_sp->column('db_vessel_type')->extdata($db_vessel_type);
			$ins_sp->column('db_material_type')->extdata($db_material_type);
			$ins_sp->column('db_user')->extdata($db_contact);
			$ins_sp->column('protocol_id')->extdata($protocol_id) 
			if ($protocol_file ne "");
			$ins_sp->column('production_dt')->extdata($production_dt);
			$ins_sp->column('storage_dt')->extdata($storage_dt);
			$ins_sp->column('comment')->extdata(COMMENT);
			$ins_sp->insert;
			$ins_sp->check_status (die => 'ERR');
            $samples{$ext_sample}=$sample_id;
			my $sp_guid = $ins_sp->column('guid')->intdata;
			print "(sp).."; $msg .= "(sp)..";
			$ins_sc->column('sample_id')->intdata($sample_id);
			$ins_sc->column('sample_id')->encoded(1);
			$ins_sc->column('project_id')->extdata($project_id);
			$ins_sc->column('reused')->intdata('N');
			$ins_sc->column('reused')->encoded(1);
			$ins_sc->insert;
			$ins_sc->check_status (die => 'ERR');
			my $sc_guid = $ins_sc->column('guid')->intdata;
			print "(sc).."; $msg .= "(sc)..";
			$ins_ss->column('sample_id')->intdata($sample_id);
			$ins_ss->column('sample_id')->encoded(1);
			$ins_ss->column('storage_id')->intdata($storage_id);
			$ins_ss->column('storage_id')->encoded(1);
            $ins_ss->column('no_units')->extdata($no_units);
			$ins_ss->insert;
			$ins_ss->check_status (die => 'ERR');
			my $ss_guid = $ins_ss->column('guid')->intdata;
			print "(ss).."; $msg .= "(ss)..";
            my $unload_sql = "";
			$unload_sql .= "DELETE FROM storage_samples ";
			$unload_sql .= "WHERE guid = '$ss_guid';\n";
            $unload_sql .= "DELETE FROM sample_collection ";
			$unload_sql .= "WHERE guid = '$sc_guid';\n";
			$unload_sql .= "DELETE FROM samples ";
			$unload_sql .= "WHERE guid = '$sp_guid';\n";
            print UNLOAD_FILE $unload_sql;
            print "."x(LEN-length($msg)),"yes\n";
		}
		else{
			$msg = "Sample '$ext_sample' exists, ";
			$msg .="so skip inserting '$ext_sample'";
			print "\t$msg","."x(LEN-length($msg)),"\n";
		}
	}
    return NO_ERROR;
}

sub load_dna{
	my ($row,$col) = @_;
	print TAB . "DNA loading\n";
	print UNLOAD_FILE "-- Rolls back DNA \n";
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $ins_ss=Apiis::DataBase::Record->new(tablename =>'storage_samples',);
		my $ins_sp=Apiis::DataBase::Record->new(tablename =>'samples', );
		my $ins_de=Apiis::DataBase::Record->new(tablename =>'dna_extraction',);
		my $dna_id 				= next_id('samples','sample_id');
        my $material_type		= "DNA";
        my $ext_sample			= read_data($sp_file_ref,$i,8,'yes');
        my $ext_dna				= read_data($sp_file_ref,$i,21,'yes');
        my $protocol_file		= read_data($sp_file_ref,$i,23,'yes');
		$protocol_file = trimstr($protocol_file) if defined $protocol_file;
		$protocol_file = convert_dummy('PROTOCOL',$protocol_file);
       	$protocol_file = 'dna_extraction__'.$protocol_file;
        my $no_units			= read_data($sp_file_ref,$i,24,'yes');
        my $dna_concentration 	= read_data($sp_file_ref,$i,25,'yes');
        my $dna_purity			= read_data($sp_file_ref,$i,26,'yes');
        my $vessel_type			= read_data($sp_file_ref,$i,27,'yes');
        $vessel_type   = convert_dummy('VESSEL',$vessel_type);
        my $production_dt		= read_data($sp_file_ref,$i,22,'no');
        my $storage_dt			= read_data($sp_file_ref,$i,28,'no');
        my $key1 				= read_data($sp_file_ref,$i,29,'yes');
        my $key2 				= read_data($sp_file_ref,$i,30,'yes');
        my $key3 				= read_data($sp_file_ref,$i,31,'yes');
        my $key4 				= read_data($sp_file_ref,$i,32,'yes');
        my $key5 				= read_data($sp_file_ref,$i,33,'yes');
		$key1 = convert_dummy('STORAGE',$key1);
		$key2 = convert_dummy('STORAGE',$key2);
		$key3 = convert_dummy('STORAGE',$key3);
		$key4 = convert_dummy('STORAGE',$key4);
		$key5 = convert_dummy('STORAGE',$key5);
		my $contact_id		= read_data($sp_file_ref,$i,34,'yes');
		$contact_id 		= convert_dummy('CONTACT',$contact_id);
		my $db_contact		= $users{$contact_id};
        my $loc = "::".$key1."::".$key2."::".$key3."::".$key4."::".$key5;
        my $protocol_id;
        $protocol_id = $protocols{$protocol_file} if ($protocol_file ne "");
        my $storage_id		= $dna_storages{$loc};
        if(check_existing('samples','ext_sample',$ext_dna,'') eq 'false'){
			$msg = "Inserting '$ext_dna'..";
			print "\t$msg";
            $ins_sp->column('sample_id')->intdata($dna_id);
            $ins_sp->column('sample_id')->encoded(1);
            $ins_sp->column('ext_sample')->extdata($ext_dna);
            $ins_sp->column('db_vessel_type')->extdata($vessel_type);
            $ins_sp->column('db_material_type')->extdata($material_type);
            $ins_sp->column('db_user')->extdata($db_contact);
            $ins_sp->column('protocol_id')->extdata($protocol_id)
            	if ($protocol_file ne "");
            $ins_sp->column('production_dt')->extdata($production_dt);
            $ins_sp->column('storage_dt')->extdata($storage_dt);
            $ins_sp->column('comment')->extdata(COMMENT);
            $ins_sp->insert;
            $ins_sp->check_status (die => 'ERR');
            print "(sp).."; $msg .= "(sp)..";
            $dna_ids{$ext_dna}=$dna_id;
            my $sp_guid = $ins_sp->column('guid')->intdata;
            $ins_de->column('dna_id')->intdata($dna_id);
            $ins_de->column('dna_id')->encoded(1);
            $ins_de->column('sample_id')->intdata($samples{$ext_sample});
            $ins_de->column('sample_id')->encoded(1);
            $ins_de->column('project_id')->extdata($project_id);
            $ins_de->column('reused')->intdata('N');
            $ins_de->column('reused')->encoded(1);
            $ins_de->column('dna_concentration')->extdata($dna_concentration)
            	if (defined $dna_concentration);
			$ins_de->column('dna_purity')->extdata($dna_purity)
			    if (defined $dna_purity);
            $ins_de->insert;
            $ins_de->check_status (die => 'ERR');
            print "(de).."; $msg .= "(de)..";
            my $de_guid = $ins_de->column('guid')->intdata;
            $ins_ss->column('sample_id')->intdata($dna_id);
            $ins_ss->column('sample_id')->encoded(1);
            $ins_ss->column('storage_id')->intdata($storage_id);
            $ins_ss->column('storage_id')->encoded(1);
            $ins_ss->column('no_units')->extdata($no_units);
            $ins_ss->insert;
            $ins_ss->check_status (die => 'ERR');
            print "(ss).."; $msg .= "(ss)..";
            my $ss_guid = $ins_ss->column('guid')->intdata;
            my $unload_sql  = "";
            $unload_sql .= "DELETE FROM storage_samples ";
            $unload_sql .= "WHERE guid = '$ss_guid';\n";
            $unload_sql .= "DELETE FROM dna_extraction ";
            $unload_sql .= "WHERE guid = '$de_guid';\n";
            $unload_sql .= "DELETE FROM samples ";
            $unload_sql .= "WHERE guid = '$sp_guid';\n";
            print UNLOAD_FILE $unload_sql;
            print "."x(LEN-length($msg)),"yes\n";
        }
        else{
        	$msg  = "Sample '$ext_sample' exists, ";
            $msg .= " so skip inserting '$ext_sample'";
            print "\t$msg","."x(LEN-length($msg)),"No\n";
		}
	}
    return NO_ERROR;
}

sub load_storage{
	my ($row,$col,$type) = @_;
	print TAB . "STOGARE loading\n";
	my (%level1,%level2,%level3,%level4,%level5,%samples);
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $key1 = read_data($sp_file_ref,$i,$col  ,'yes');
		my $key2 = read_data($sp_file_ref,$i,$col+1,'yes');
		my $key3 = read_data($sp_file_ref,$i,$col+2,'yes');
		my $key4 = read_data($sp_file_ref,$i,$col+3,'yes');
		my $key5 = read_data($sp_file_ref,$i,$col+4,'yes');
		$key1 = convert_dummy('STORAGE',$key1);
		$key2 = convert_dummy('STORAGE',$key2);
		$key3 = convert_dummy('STORAGE',$key3);
		$key4 = convert_dummy('STORAGE',$key4);
		$key5 = convert_dummy('STORAGE',$key5);
        my $sample_id 	= read_data($sp_file_ref,$i,10,'yes');
		my $l1 = $key1;
		my $l2 = "::".$key1."::".$key2;
		my $l3 = "::".$key1."::".$key2."::".$key3;
		my $l4 = "::".$key1."::".$key2."::".$key3."::".$key4;
		my $l5 = "::".$key1."::".$key2."::".$key3."::".$key4."::".$key5;
		$level1{$l1}++;$level2{$l2}++;
		$level3{$l3}++;$level4{$l4}++;$level5{$l5}++;
	}
	print UNLOAD_FILE "-- Rolls back STORAGES \n";
	print TAB."  Level 1 \n";
	my $ins_sr = Apiis::DataBase::Record->new(tablename => 'storage',);
	my $unload_sql  = "";
	for my $key ( keys %level1) {
		if(check_existing('storage','storage_name',$key,'0') eq 'false'){
			$msg = "Inserting '$key'";
			print "\t$msg","."x(LEN-length($msg));
			my $storage_id = next_id('storage','storage_id');
			$level1{$key} = $storage_id;
			$ins_sr->column('storage_id')->extdata($storage_id);
			$ins_sr->column('storage_name')->extdata($key);
			$ins_sr->column('parent_id')->extdata('0');
			$ins_sr->insert;
			$ins_sr->check_status (die => 'ERR');
			my $sp_guid = $ins_sr->column('guid')->intdata;
			print "yes\n";
			$unload_sql .= "DELETE FROM storage ";
			$unload_sql .= "WHERE guid = '$sp_guid';\n";
		}
		else{
			my $sql = "SELECT storage_id FROM storage ";
			$sql   .= "WHERE storage_name ='$key' AND parent_id ='0'";
			my $storage_id = get_storage_id($sql);
			print "\tLocation '$key' exists, so skip inserting ";
            print "'$key' --> $storage_id\n";
			$level1{$key} = $storage_id;
		}
    }
	print TAB."  Level 2 \n";
	for my $key ( keys %level2) {
		my ($node_name, $parent_name, $head) = get_names($key);
		my $parent_id = $level1{$parent_name};
		if(check_existing('storage','storage_name',$node_name,$parent_id)
		eq 'false'){
			$msg = "Inserting '$node_name'";
			print "\t$msg","."x(LEN-length($msg));
			my $storage_id = next_id('storage','storage_id');
			$level2{$key} = $storage_id;
			$ins_sr->column('storage_id')->extdata($storage_id);
			$ins_sr->column('storage_name')->extdata($node_name);
			$ins_sr->column('parent_id')->extdata($parent_id);
			$ins_sr->insert;
			$ins_sr->check_status (die => 'ERR');
			my $sp_guid = $ins_sr->column('guid')->intdata;
			print "yes\n";
			$unload_sql .= "DELETE FROM storage ";
			$unload_sql .= "WHERE guid = '$sp_guid';\n";
		}
		else{
			my $sql = "SELECT storage_id FROM storage ";
			$sql   .= "WHERE storage_name ='$node_name' ";
			$sql   .= "AND parent_id ='$parent_id'";
			my $storage_id = get_storage_id($sql);
			print "\tLocation '$key' exists, so skip inserting ";
            print "'$node_name' --> $storage_id\n";
			$level2{$key} = $storage_id;
		}
    }
	print TAB."  Level 3 \n";
	for my $key ( keys %level3) {
		my ($node_name, $parent_name, $head) = get_names($key);
		my $parent_id = $level2{$head};
		if(check_existing('storage','storage_name',$node_name,$parent_id)
		eq 'false'){
			$msg = "Inserting '$node_name'";
			print "\t$msg","."x(LEN-length($msg));
			my $storage_id = next_id('storage','storage_id');
			$level3{$key} = $storage_id;
			$ins_sr->column('storage_id')->extdata($storage_id);
			$ins_sr->column('storage_name')->extdata($node_name);
			$ins_sr->column('parent_id')->extdata($parent_id);
			$ins_sr->insert;
			$ins_sr->check_status (die => 'ERR');
			my $sp_guid = $ins_sr->column('guid')->intdata;
			print "yes\n";
			$unload_sql .= "DELETE FROM storage ";
			$unload_sql .= "WHERE guid = '$sp_guid';\n";
		}
		else{
			my $sql = "SELECT storage_id FROM storage ";
			$sql   .= "WHERE storage_name ='$node_name' ";
			$sql   .= "AND parent_id ='$parent_id'";
			my $storage_id = get_storage_id($sql);
			print "\tLocation '$key' exists, so skip inserting ";
            print "'$node_name' --> $storage_id\n";
			$level3{$key} = $storage_id;
		}
    }
	print TAB."  Level 4 \n";
	for my $key ( keys %level4) {
		my ($node_name, $parent_name, $head) = get_names($key);
		my $parent_id = $level3{$head};
		if(check_existing('storage','storage_name',$node_name,$parent_id)
		eq 'false'){
			$msg = "Inserting '$node_name'";
			print "\t$msg","."x(LEN-length($msg));
			my $storage_id = next_id('storage','storage_id');
			$level4{$key} = $storage_id;
			$ins_sr->column('storage_id')->extdata($storage_id);
			$ins_sr->column('storage_name')->extdata($node_name);
			$ins_sr->column('parent_id')->extdata($parent_id);
			$ins_sr->insert;
			$ins_sr->check_status (die => 'ERR');
			print "yes\n";
			my $sp_guid = $ins_sr->column('guid')->intdata;
			$unload_sql .= "DELETE FROM storage ";
			$unload_sql .= "WHERE guid = '$sp_guid';\n";
		}
		else{
			my $sql = "SELECT storage_id FROM storage ";
			$sql   .= "WHERE storage_name ='$node_name' ";
			$sql   .= "AND parent_id ='$parent_id'";
			my $storage_id = get_storage_id($sql);
			print "\tLocation '$key' exists, so skip inserting ";
            print "'$node_name' --> $storage_id\n";
			$level4{$key} = $storage_id;
		}
    }
	print TAB."  Level 5 \n";
	for my $key ( keys %level5) {
		my ($node_name, $parent_name, $head) = get_names($key);
		my $parent_id = $level4{$head};
		if(check_existing('storage','storage_name',$node_name,$parent_id) eq 'false'){
			$msg = "Inserting '$node_name'";
			print "\t$msg","."x(LEN-length($msg));
			my $storage_id = next_id('storage','storage_id');
			$level5{$key} = $storage_id;
			$ins_sr->column('storage_id')->extdata($storage_id);
			$ins_sr->column('storage_name')->extdata($node_name);
			$ins_sr->column('parent_id')->extdata($parent_id);
			$ins_sr->insert;
			$ins_sr->check_status (die => 'ERR');
			my $sp_guid = $ins_sr->column('guid')->intdata;
			if (uc($type) eq "SAMPLE"){
				$sample_storages{$key} = $storage_id;
			}
			else{
				$dna_storages{$key} = $storage_id;
			}
			print "yes\n";
			$unload_sql .= "DELETE FROM storage ";
			$unload_sql .= "WHERE guid = '$sp_guid';\n";
		}
		else{
			my $sql = "SELECT storage_id FROM storage ";
			$sql   .= "WHERE storage_name ='$node_name' ";
			$sql   .= "AND parent_id ='$parent_id'";
			my $storage_id = get_storage_id($sql);
			print "\tLocation '$key' exists, so skip inserting ";
            print "'$node_name' --> $storage_id\n";
			$level5{$key} = $storage_id;
            $sample_storages{$key} = $storage_id;
		}
    }
    print UNLOAD_FILE $unload_sql;
    return NO_ERROR;
}
sub load_sequences{
	my ($row,$col) = (2,2);
	my $filename = "datafile_sq.". get_file_extension($sq_datafile_format);
	$sq_file_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");
	print TAB . "Reading from $path/$filename \n";
# 	my $number_of_sq = $record_no;
	print TAB . "SEQUENCES loading $record_no, $sq_marker_no\n";
	my %file_names;		# ext_dna --> filename
	print UNLOAD_FILE "-- Rollback SEQUENCES \n";
	my $usql  = "";
	print "\t     ";
	for (my $j=1;$j<=$sq_marker_no;$j++){
		print ($j%10);
		print "  ";
	}
	print "   DNA IDs\n";
	for (my $i=$row;$i<($row+$record_no);$i++){
		my $ins_sq = Apiis::DataBase::Record->new(tablename =>'sequences',);
		my $ext_dna = read_data($sq_file_ref,$i,1,"yes");
		if ($ext_dna eq ""){
			next;
		}
		print "\t".($i-1).".", " "x(4-length($i-1));
		for (my $j=1;$j<=$sq_marker_no;$j++){
			my $file_name 	= read_data($sq_file_ref,$i,$j+1,"yes");
# 			print "file name: ".$file_name."\n";
			my $marker 		= read_data($sq_file_ref,1,$j+1,"yes");
			if ($file_name eq ""){ $flag = EMPTY_SQ; return $flag; }
			$file_names{$ext_dna}= $file_name;
			print "ok ";
			my $sq_id =	next_id('sequences','sequence_id');
			my $dna_id; 				# get dna_id from ext_dna	
			if ($sp_loading eq 'yes'){
				$dna_id = $dna_ids{$ext_dna};
			}
			else{
				$dna_id = $dummy_dna_ids{$ext_dna};
			}
			my $marker_id = $marker_ids{$marker};
			my $filename = "$path/final_sq/$file_names{$ext_dna}";
			my $sq_data = read_sequence($filename);
			$ins_sq->column('sequence_id')->intdata($sq_id);
			$ins_sq->column('sequence_id')->decoded(1);
			$ins_sq->column('dna_id')->intdata($dna_id);
			$ins_sq->column('dna_id')->decoded(1);
			$ins_sq->column('project_id')->intdata($project_id);
			$ins_sq->column('project_id')->decoded(1);
			$ins_sq->column('marker_id')->intdata($marker_id);
			$ins_sq->column('marker_id')->decoded(1);
			$ins_sq->column('sequence')->extdata($sq_data);
			$ins_sq->insert;
			$ins_sq->check_status (die => 'ERR');
			my $sq_guid = $ins_sq->column('guid')->intdata;
			$usql .= "DELETE FROM sequences ";
			$usql .= "WHERE guid = '$sq_guid';\n";
		}
		print "  $ext_dna\n";
    }
	print UNLOAD_FILE $usql;
    return NO_ERROR;
}

sub load_microsatellites{
	my ($row,$col) = (2,1);
	print TAB . "MICROSATELLITES loading\n";
	print UNLOAD_FILE "-- Rollback MICROSATELLITES \n";
	my $filename = "final_ms/$ms_datafile";
	$ms_file_ref = ReadData("$path/$filename",dtfmt =>"dd-mm-yyyy",sep =>",");
	print "\t     ";
	for (my $j=1;$j<=$ms_marker_no;$j++){
		print ($j%10);
		print " ";
	}
	print "   DNA IDs\n";
	$record_no = 0;
	while (1){
		my $ins_ms = Apiis::DataBase::Record->new(tablename => 'microsatellites',);
		last if (read_data($ms_file_ref,$row,1,'yes') eq '');
		$record_no++;
		my $ext_dna 	= read_data($ms_file_ref,$row,1,"yes");
		print "\t".$record_no.".", " "x(4-length($record_no));
  		# print $ext_dna . " ";
		for (my $j=1;$j<=$ms_marker_no;$j++){
			my ($marker,$allele1,$allele2);
			if (($ms_data_file_type eq '2')
				or (uc(trimstr($ms_data_file_type)) eq 'DIPLOIT-TWO-COLUMN')){
				$marker  = trimstr(read_data($ms_file_ref,1,$j*2,"yes"));
                #print $marker . " |";
				$allele1 = read_data($ms_file_ref,$row,$j*2,"yes");
				$allele2 = read_data($ms_file_ref,$row,$j*2+1,"yes");
			} elsif (($ms_data_file_type eq '1')
				or (uc(trimstr($ms_data_file_type)) eq 'DIPLOIT-ONE-COLUMN')){
				$marker  = trimstr(read_data($ms_file_ref,1,$j+1,"yes"));
                #print $marker . " =";
				$allele1 = read_data($ms_file_ref,$row,$j+1,"yes");
				$allele2 = read_data($ms_file_ref,$row+1,$j+1,"yes");
			}
			my $ms_id =	next_id('microsatellites','microsatellite_id');
			my $dna_id;		# get dna_id from ext_dna
			if ($sp_loading eq 'yes'){
				$dna_id = $dna_ids{$ext_dna};
			}
			else{
				$dna_id = $dummy_dna_ids{$ext_dna};
			}
			my $marker_id = $marker_ids{$marker};	# get id from marker_name
			$ins_ms->column('microsatellite_id')->intdata($ms_id);
			$ins_ms->column('microsatellite_id')->decoded(1);
			$ins_ms->column('dna_id')->intdata($dna_id);
			$ins_ms->column('dna_id')->decoded(1);
			$ins_ms->column('project_id')->intdata($project_id);
			$ins_ms->column('project_id')->decoded(1);
			$ins_ms->column('marker_id')->intdata($marker_id);
			$ins_ms->column('marker_id')->decoded(1);
			$ins_ms->column('allele1')->extdata($allele1);
			$ins_ms->column('allele2')->extdata($allele2);
			$ins_ms->column('comment')->extdata(COMMENT);
			$ins_ms->insert;
			$ins_ms->check_status (die => 'ERR');
			if (($allele1 ne "")and($allele2 ne "")){
				if ($allele1 eq $allele2){
					print "+ ";
				}
				else{
					print "- ";
				}
			}
			else{
				print "? ";
			}
			my $ms_guid = $ins_ms->column('guid')->intdata;
			my $usql  = "";
			$usql .= "DELETE FROM microsatellites ";
			$usql .= "WHERE guid = '$ms_guid';\n";
			print UNLOAD_FILE $usql;
		}
		print "  $ext_dna\n";
		if (($ms_data_file_type eq '2')
			or (uc(trimstr($ms_data_file_type)) eq 'DIPLOIT-TWO-COLUMN')){
			$row++;
		} elsif (($ms_data_file_type eq '1')
			or (uc(trimstr($ms_data_file_type)) eq 'DIPLOIT-ONE-COLUMN')){
        	$row += 2;
		}
	}
    return NO_ERROR;
}

sub check_existing{
	my ($table_name,$field_name,$value,$extra_value) = @_;
	my $sql = "";
	$value = sprintf "%s",$apiis->DataBase->dbh->quote($value);
	$extra_value = sprintf "%s",$apiis->DataBase->dbh->quote($extra_value);
	if ($table_name eq 'projects'){
		$sql .= "SELECT count(*) FROM $table_name ";
		$sql .= "WHERE $field_name = $value";
	}
	elsif ($table_name eq 'codes'){
		$sql .= "SELECT count(*) FROM $table_name ";
		$sql .= "WHERE class = $extra_value AND $field_name = $value";
	}
	elsif ($table_name eq 'contacts'){
		$sql .= "SELECT count(*) FROM $table_name ";
		$sql .= "WHERE $field_name = $value";
	}
	elsif ($table_name eq 'protocols'){
		$sql .= "SELECT count(*) FROM $table_name p, codes c ";
		$sql .= "WHERE p.db_protocol_type = c.db_code "; 
		$sql .= "AND $field_name = $value ";
		$sql .= "AND ext_code =  $extra_value ";
	}
	elsif ($table_name eq 'storage'){
		$sql .= "SELECT count(*) FROM $table_name sp ";
		$sql .= "WHERE  $field_name = $value ";
		$sql .= "AND parent_id =  $extra_value ";
	}
	else {
		$sql .= "SELECT count(*) FROM $table_name ";
		$sql .= "WHERE $field_name = $value";
	}
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
	my @data_array = $sql_ref->handle->fetchrow_array;
#   	print $sql . "\n";
	if ($data_array[0] eq 0){
    	return 'false';
    }
    else{
    	return 'true';
    }
}

sub get_storage_id{
	my $sql = shift;
	my $sql_ref = $apiis->DataBase->user_sql($sql);
	$apiis->check_status;
	my @data_array = $sql_ref->handle->fetchrow_array;
	if (defined $data_array[0]){
		return $data_array[0];
	}
	else{
		return "0";
	}
}

sub usage {
    print "Usage:\n"
      . "    -h    help, know how to use \n\n"
      . "    You must supply 5 required parameters:\n"
      . "    -n <> database name\n"
      . "    -u <> database user\n"
      . "    -p <> database password\n"
      . "    -m <> action mode (values:load,check)\n"
      . "    -d <> directory path to external data source\n\n";
    die "\n";
}

sub convert_dummy{
	my ($data_type,$data_str) = @_;
    if ($data_str eq ""){
		return "unknown_breed" 		if ($data_type eq 'BREED');
		return "u"             		if ($data_type eq 'SEX');
		return "unknown"       		if ($data_type eq 'STORAGE');
 		return "unknown_type"  		if ($data_type eq 'VESSEL');
		return "unknown_contact" 	if ($data_type eq 'CONTACT');
		return "dummy_protocol.txt" if ($data_type eq 'PROTOCOL');
    } else {
    	return $data_str;
    }
}

sub read_data {
	my ($datafile_ref,$row,$col,$trimming,$data_obj) = @_;
	print "\t* DEBUG: Reading data from a data file at row $row,col $col: " if $debug;
	my $datastr = $datafile_ref->[1]{cell}[$col][$row];
	if (defined $datastr){
		$datastr = Encode::decode_utf8($datastr);
	}
	if ($trimming eq 'yes'){
		$datastr = trimstr($datastr) if defined $datastr;	
	}
#    $datastr =~ s/'/''/;
	if (defined $datastr){
		print "$datastr\n" if $debug;	
		return $datastr;
	} else {
		print "empty\n" if $debug;	
		return "";
	}
}

sub read_sequence {
	my ($filename) = @_;
	my $format = "fasta";
	my $seq_obj = Bio::SeqIO->new(-file=>$filename, -format=>$format);
	my $seq     = $seq_obj->next_seq();
    my $sq_data = "";
    if (defined $seq->seq){
		$sq_data = $seq->seq;
	}
	if ($sq_data eq ""){
		$format = "raw";
		$seq_obj = Bio::SeqIO->new(-file=>$filename, -format=>$format);
		$seq     = $seq_obj->next_seq();
		$sq_data = $seq->seq;		
	}
	return $sq_data;
}

sub trimstr{
	my $str = shift;
	$str =~ s/^\s+//;
	$str =~ s/\s+$//;
	return $str;
}

sub get_names{
	my $fullname = shift;
    my ($parent_name, $node_name) = ("","");
    my $pos = rindex($fullname,"::");
    $node_name = substr($fullname,$pos+2,length($fullname)-$pos);
    $fullname = substr($fullname,0,$pos);
    $pos = rindex($fullname,"::");
    $parent_name = substr($fullname,$pos+2,length($fullname)-$pos);
	return ($node_name,$parent_name,$fullname);
}

sub next_id{
	my ($table_name,$column_name) = @_;
	return $apiis->DataBase->seq_next_val("seq_".$table_name."__".$column_name);
}

sub get_file_extension {
	my $format = shift;
	return "xls" if (uc($format) eq "EXCEL");
	return "xls" if (uc($format) eq "XLS");
	return "csv" if (uc($format) eq "CSV");
	return "ods" if (uc($format) eq "ODS");
	return "xls";
}

sub get_uniqueID {
	my ($table_name,$output_field,$input_field,$input_value) = @_;
    my $sql = "select $output_field from $table_name where $input_field = '$input_value'";
    # print "get_uniqueID $sql\n";
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @data_array = $sql_ref->handle->fetchrow_array;
    my $output_value = -1;
    $output_value = $data_array[0] if defined $data_array[0];
    return $output_value;
}

=head1 AUTHOR

Truong Van Chi Cong <cong.chi@fli.bund.de>

=cut
