#!/usr/bin/env perl

=head1 NAME

A Perl program to load initial data for MolabIS database

=head1 DESCRIPTION

This program inserts a number of required records into MolabIS database. It
should be run at first before working with the web application. A data file
mimetypes.dat which stays in the same folder with this script will be read.
So, make sure that this file is available for reading.

=cut

=head1 SYNOPSIS

perl B<initiate_database.pl> B<-n> <database> B<-u> <username> B<-p> <password>

=cut

BEGIN {
   use Env qw( APIIS_HOME );
   die "\n\tAPIIS_HOME is not set!\n\n" unless $APIIS_HOME;
   push @INC, "$APIIS_HOME/lib";
}

use strict;
use warnings;
use Apiis;
use Data::Dumper;
use Getopt::Std;
Apiis->initialize;
my ($db_name,$db_user,$db_pass,$flag,$msg);
my ($db_language,$db_country);

use constant NO_ERROR	=> 0;
use constant LEN		=> 50;
# messages which are output when there is a runtime error
my @em;
$em[0]  = "There is no error.                                       ";

# check command line parameters
use vars qw($opt_h $opt_n $opt_u $opt_p);

run();

sub init{
	getopts('hn:u:p:');
	usage() if $opt_h;
	# required parameters
	usage() if ! $opt_n;
	usage() if ! $opt_u;
	usage() if ! $opt_p;
	$db_name   = $opt_n;
	$db_user   = $opt_u;
	$db_pass   = $opt_p;
	# SQL commands for data rollback
	open UNLOAD_FILE, ">", "clear_database.sql";
	require Apiis::DataBase::User;
	my $dummy = Apiis::DataBase::User->new(
		id => $db_user, password => $db_pass,
	);
	$apiis->join_model($db_name, userobj => $dummy);
	$apiis->check_status(die => 'ERR');
	$flag = NO_ERROR;
}

sub run{
	init();
	my $error_value = load_data();
	if ($error_value eq NO_ERROR){
		$apiis->DataBase->dbh->commit;
		print "\n";
		print "\t-------------------- Data Commit --------------------\n";
		print "\t- Initial data was successfully loaded to Database  -\n";
		print "\t-----------------------------------------------------\n";
		print "\n";
	}
	else {
		print "\t----------------------- Error -----------------------\n";
		print "\t- " .            $em[$error_value]              . " -\n";
		print "\t-----------------------------------------------------\n\n";
	}
}

sub load_data{
	# required records
	open (MIMEFILE, "<mimetypes.dat") or die "Not Found mimetypes.dat $!\n";
	while (my $code = <MIMEFILE>) {
		chomp($code);
		my $ext_code	=$code;
		my $short_name	=$code;
		my $long_name	=$code;
		return $flag if (load_codes('MIMETYPE',$ext_code,$short_name,$long_name) ne NO_ERROR);
	}
	return $flag if (load_codes('MATERIAL_TYPE','DNA','DNA','DNA') ne NO_ERROR);
	return $flag if (load_codes('DONOR_TYPE','Animals','Animals','Animals') ne NO_ERROR);
	return $flag if (load_codes('DONOR_TYPE','Plants','Plants','Plants') ne NO_ERROR);
	return $flag if (load_codes('PROTOCOL_TYPE','Sampling','Sampling','Sampling') ne NO_ERROR);
	return $flag if (load_codes('PROTOCOL_TYPE','DNA extraction','DNA extraction','DNA extraction') ne NO_ERROR);
	return $flag if (load_codes('PROTOCOL_TYPE','DNA validation','DNA validation','DNA validation') ne NO_ERROR);
	return $flag if (load_codes('PROTOCOL_TYPE','PCR amplification','PCR amplification','PCR amplification') ne NO_ERROR);
	return $flag if (load_codes('PROTOCOL_TYPE','PCR validation','PCR validation','PCR validation') ne NO_ERROR);
	return $flag if (load_codes('PROTOCOL_TYPE','Electrophoresis','Electrophoresis','Electrophoresis') ne NO_ERROR);
	return $flag if (load_codes('METHOD','CE','Capillary Electrophoresis','Capillary Electrophoresis') ne NO_ERROR);
	return $flag if (load_codes('METHOD','GE','Gel Electrophoresis','Gel Electrophoresis') ne NO_ERROR);
	return $flag if (load_codes('PURPOSE','S','Sequencing','Sequencing') ne NO_ERROR);
	return $flag if (load_codes('PURPOSE','G','Genotyping','Genotyping') ne NO_ERROR);
	return $flag if (load_codes('MARKER_TYPE','Microsatellite','Microsatellite','Microsatellite') ne NO_ERROR);
	return $flag if (load_codes('MARKER_TYPE','Sequencing','Sequencing','Sequencing') ne NO_ERROR);
	return $flag if (load_codes('COUNTRY','Germany','Germany','Germany') ne NO_ERROR);
 	return $flag if (load_codes('COUNTRY','Vietnam','Vietnam','Vietnam') ne NO_ERROR);
	return $flag if (load_codes('LANGUAGE','EN','EN','English') ne NO_ERROR);
 	return $flag if (load_codes('LANGUAGE','DE','DE','German') ne NO_ERROR);
   	return $flag if (load_codes('LANGUAGE','VI','VI','Vietnamese') ne NO_ERROR);
	return $flag if (load_codes('SOFTWARE','unknown_software','unknown_software','unknown_software') ne NO_ERROR);
	return $flag if (load_codes('BREED','unknown_breed','unknown_breed','unknown_breed') ne NO_ERROR);
	return $flag if (load_codes('SPECIES','unknown_species','unknown_species','unknown_species') ne NO_ERROR);
	return $flag if (load_codes('PRIMER','Default','Default','Default') ne NO_ERROR);
	return $flag if (load_codes('VESSEL_TYPE','Default','Default','Default') ne NO_ERROR);
	return $flag if (load_codes('SEX','m','m','male') ne NO_ERROR);
	return $flag if (load_codes('SEX','f','f','female') ne NO_ERROR);
	return $flag if (load_codes('SEX','u','u','unknown_sex') ne NO_ERROR);
	return $flag if (load_units('USERS','unknown_contact') ne NO_ERROR);
	return $flag if (load_units('ORGANISMS','ID') ne NO_ERROR);
	return $flag if (load_transfers() ne NO_ERROR);
	return $flag;
}

sub load_codes{
	my ($class_name,$ext_code,$short_name,$long_name) = @_;
	my $ins_codes=Apiis::DataBase::Record->new(tablename=>'codes', );
	if(check_existing('codes','ext_code',$ext_code,$class_name) eq 'false'){
		$msg = "Inserting '$ext_code'";
		print "\t$msg", "."x(LEN-length($msg));
		$ins_codes->column('ext_code')->extdata($ext_code);
		$ins_codes->column('class')->extdata($class_name);
		$ins_codes->column('short_name')->extdata($short_name);
		$ins_codes->column('long_name')->extdata($long_name);
		$ins_codes->column('opening_dt')->extdata($apiis->now);
		$ins_codes->insert;
		$ins_codes->check_status (die => 'ERR');
		print "Yes\n";
       	$db_country =  $ins_codes->column('db_code')->intdata
			if ($class_name eq 'COUNTRY');
       	$db_language =  $ins_codes->column('db_code')->intdata
			if ($class_name eq 'LANGUAGE');
		my $usql = "";
		$usql .= "DELETE FROM codes ";
		$usql .= "WHERE class = '$class_name' AND ext_code='$ext_code';\n";
		print UNLOAD_FILE $usql;
	}
	else{
		$msg = "$class_name '$ext_code' exists, so skip inserting '$ext_code'";
		print "\t$msg", "."x(LEN-length($msg));
	}
	return NO_ERROR;
}

sub load_units{
	my ($ext_unit,$ext_id) = @_;
	my $ins_unit=Apiis::DataBase::Record->new(tablename=>'unit', );
	my $ins_contacts=Apiis::DataBase::Record->new(tablename=>'contacts', );
	if(check_existing('unit','ext_unit',$ext_unit,$ext_id) eq 'false'){
		$msg = "Inserting '".$ext_unit."::".$ext_id."'";
		print "\t$msg", "."x(LEN-length($msg));
 		my $db_contact = $apiis->DataBase->seq_next_val("seq_contacts__db_contact");
		$ins_contacts->column('db_contact')->extdata($db_contact);
		$ins_contacts->column('first_name')->extdata('Default');
		$ins_contacts->column('second_name')->extdata('Default');
		$ins_contacts->column('db_country')->intdata($db_country);
		$ins_contacts->column('db_country')->decoded(1);
		$ins_contacts->column('db_language')->intdata($db_language);
		$ins_contacts->column('db_language')->decoded(1);
		$ins_contacts->insert;
		$ins_contacts->check_status(die=>'ERR');
		my $guid_ct = $ins_contacts->column('guid')->intdata;
		$ins_unit->column('ext_unit')->extdata($ext_unit);
		$ins_unit->column('ext_id')->extdata($ext_id);
		$ins_unit->column('opening_dt')->extdata($apiis->now);
		$ins_unit->column('db_contact')->extdata($db_contact);
		$ins_unit->insert;
		$ins_unit->check_status(die=>'ERR');
		my $guid_un = $ins_unit->column('guid')->intdata;
		print "Yes\n";
		my $usql = "";
		$usql .= "DELETE FROM contacts ";
		$usql .= "WHERE guid = '$guid_ct'\n";
		$usql .= "DELETE FROM unit ";
		$usql .= "WHERE guid = '$guid_un'\n";
		print UNLOAD_FILE $usql;
	}
	else{
		$msg = "Unit '$ext_unit'::'$ext_id' exists, so skip inserting '$ext_unit'::'$ext_id'";
		print "\t$msg", "."x(LEN-length($msg));
	}
	return NO_ERROR;
}

sub load_transfers{
	# Insert unknown sire (1) and unknown dam (2)
	my %unknown=('1'=>'unknown_sire','2'=>'unknown_dam',);
	for my $i (1,2) {
		my $ins_transfer=Apiis::DataBase::Record->new( tablename =>'transfer',);
		$ins_transfer->column('db_unit')->extdata('ORGANISMS','ID');
		$ins_transfer->column('ext_organism')->extdata($unknown{$i});
		$ins_transfer->column('db_organism')->intdata($i);
		$ins_transfer->column('db_organism')->encoded(1);
		$ins_transfer->column('opening_dt')->extdata($apiis->now);
		$ins_transfer->insert;
		$ins_transfer->check_status (die => 'ERR');
		my $guid = $ins_transfer->column('guid')->intdata;
		my $usql = "";
		$usql .= "DELETE FROM transfer ";
		$usql .= "WHERE guid = '$guid';\n";
		print UNLOAD_FILE $usql;
	}
	return NO_ERROR;
}

sub check_existing{
	my ($table_name,$field_name,$value,$extra_value) = @_;
	my $sql = "";
	if ($table_name eq 'codes'){
		$sql .= "SELECT count(*) FROM $table_name ";
		$sql .= "WHERE class = '$extra_value' AND $field_name = '$value'";
	}
	else {
		$sql .= "SELECT count(*) FROM $table_name ";
		$sql .= "WHERE $field_name = '$value'";
	}
	my $sql_ref = $apiis->DataBase->user_sql($sql);
	$apiis->check_status;
	my @data_array = $sql_ref->handle->fetchrow_array;
	if ($data_array[0] eq '0'){
		return 'false';
	}
	else{
		return 'true';
	}
}

sub usage {
	print "Usage:\n"
		. "    -h    help, know how to use \n\n"
		. "    You must supply 3 required parameters:\n"
		. "    -n <> database name\n"
		. "    -u <> database user\n"
		. "    -p <> database password\n\n";
	die "\n";
}

=head1 AUTHOR

Truong Van Chi Cong <cong.chi@fli.bund.de>

=cut
