package Molab::Ajax; 

=head1 NAME

Molab::Ajax -- Handling all Asynchronous requests

=head1 DESCRIPTION

This module consists of libraries of Ajax functions used to care of all Ajax
operations in MolabIS.  

=head1 METHODS

Available subroutines:

=cut

##############################################################################
use strict;
use warnings;
use Apiis;
use CGI qw/:standard/;
use Data::Dumper;
use Spreadsheet::Read;
use Bio::SeqIO;
use Bio::AlignIO;
use Bio::Tools::GuessSeqFormat;
use CGI qw/escapeHTML/;


##############################################################################
# INITIALIZING GLOBAL VALUES
##############################################################################
my $tmpl_dir    = "../../templates/";			# path to HTML templates
my $debug_path  = "/tmp/molab_debug_file";  	# file for debugging

##############################################################################

=head2 search - Look for a set of records matching with given conditions

Molab::Ajax::search is an Ajax function invoked from main.cgi after an user 
clicks on the button B<'Search'> on the form. The subroutine will look for all 
records affected in the query, and then make a list of rows. Each row is a result
record shown on a separate line in the 'Result Part'. Each result record consists
of several actions such as View, Update, Delete, Move, Remove which allow users
to carry out different functions.

  ---------------------------------------------------------
  .  Filter1 .......    Filter2 .......   [SEARCH]         . <--- Search Part
  ---------------------------------------------------------
  .                                                        .
  .  Results:                                              .
  .  <Col1>    <Col2>    <Coln>          <Actions>         . <--- Result Part
  .   ...       ...       ...          ... |...|...        .
  .   ...       ...       ...          ... |...|...        .
  ----------------------------------------------------------

=over 4

=item I<usage>:

    require Ajax;
    my $ajax = &Molab::Ajax::search($sql,$form_ref,$layout);

=item I<inputs>:
	
    1. String: SQL query
    2. Array ref: point to array of input fields on the form
    3. String: layout name

=item I<output>:

    1. String: AJAX string formated in HTML including two parts
        part 1: result table
        part 2: number of records

=back

=cut

sub search {
	my ($sql, $form_fields_ref, $layout) = @_;
    $apiis->log('notice',"Ajax::search: ".$sql.",".$layout);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        $row_no++;
        $hash_ref->{'no'} = $row_no;
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        # convert all isodate fields to extdate
		foreach my $key (keys %$hash_ref){
			if ($key =~ /(.*)\_[d|D][t|T]/){
				my $extdate = $apiis->iso2extdate($hash_ref->{$key});
				if (defined $extdate){
					$hash_ref->{$key} = $extdate;
				}
			}
# 			if ($hash_ref->{$key} ne ''){
# 				$hash_ref->{$key} = $t2h->parse($hash_ref->{$key});
# 			}
 			$hash_ref->{$key} = escapeHTML($hash_ref->{$key});
        }
        push(@record_sets, \%$hash_ref);
		# prepare a list of data fields used to fill data into the form
        my @form_fields;
        for(my $i = 0; $i<@{$form_fields_ref}; $i++){
            $form_fields[$i] = "'" . @{$form_fields_ref}[$i] . "'";
        }
        my $form_str = join(",", @form_fields);
		# prepare Ajax strings
		if ($layout eq "specified_contact"){
            my $report_pdf = " onclick=\"
            	showLoadingStatus(this);
            	external(['action__ajax_reporting','sid','layout__$layout','guid__$ID','NO_CACHE'],";
       		$report_pdf .= "['download_area']); \"";
			$hash_ref->{report_pdf} 	= $report_pdf;
		}
		elsif ($layout eq "analysis_by_organism"){
            my $report_pdf = " onclick=\"
            	showLoadingStatus(this);
            	external(['action__ajax_reporting','sid','layout__$layout','project_id__$hash_ref->{'project_id'}','organism_id__$hash_ref->{'organism_id'}','NO_CACHE'],";
       		$report_pdf .= "['download_area']); \"";
			$hash_ref->{report_pdf} 	= $report_pdf;
		}
		else {
	        my $view_str = " onclick=\"
	            changeSubmitLabel(\'View\',document.forms[0]);
	            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
	        my $update_str = " onclick=\"
	            changeSubmitLabel(\'Update\',document.forms[0]);
	            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
	        my $delete_str = " onclick=\"
	            changeSubmitLabel(\'Delete\',document.forms[0]);
	            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
	        # make a complete Ajax string for each action
		    $view_str   .= "[" . $form_str . "]); \"";
	        $update_str .= "[" . $form_str . "]); \"";
	        $delete_str .= "[" . $form_str . "]); \"";
	        $hash_ref->{view} 	= $view_str;
	        $hash_ref->{update} = $update_str;
	        $hash_ref->{delete} = $delete_str;
		}
    }
    # open the template for the result table
    my $tmpl_name = $layout . "_result.tmpl";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
    # fill in parameters for the template
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
#         $apiis->log('info', $_ . "----" . Dumper($tmpl_form));
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $result .= "__pjx__" . $row_no;
    $result .= (($row_no eq 0) || ($row_no eq 1)) ? " record" : " records";
    $apiis->log('debug', 'Ajax::search return ' . $result);
    return $result;
}

sub search_sequences {
	my ($sql, $form_fields_ref, $layout) = @_;
    $apiis->log('notice',"Ajax::search_sequences: " . $sql ."," . $layout);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    my $dna_id = "";
    my $seq_no = 0;
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        if ($hash_ref->{'result_col1'} ne $dna_id){
			$dna_id = $hash_ref->{'result_col1'};
			$row_no++;
			$hash_ref->{'no'} = $row_no;
			$seq_no = 1;
        }
		else {
			$hash_ref->{'result_col1'} = "";
			$hash_ref->{'no'} = "";
			$seq_no++;
		}
		$hash_ref->{'result_col2'} = "seq ".$seq_no;        
        
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        
        # convert all isodate fields to extdate
        foreach my $key (%{$hash_ref}){
             if ($key =~ /(.*)\_[d|D][t|T]/){
              my $extdate = $apiis->iso2extdate($hash_ref->{$key});
                if (defined $extdate){
                    $hash_ref->{$key} = $extdate;
                }
            }
        }
        push(@record_sets, \%$hash_ref);
		# prepare a list of data fields used to fill data into the form
        my @form_fields;
        for(my $i = 0; $i<@{$form_fields_ref}; $i++){
            $form_fields[$i] = "'" . @{$form_fields_ref}[$i] . "'";
        }
        my $form_str = join(",", @form_fields);
		# prepare Ajax strings
        my $view_str = " onclick=\"
            changeSubmitLabel(\'View\',document.forms[0]);
            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
        my $update_str = " onclick=\"
            changeSubmitLabel(\'Update\',document.forms[0]);
            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
        my $delete_str = " onclick=\"
            changeSubmitLabel(\'Delete\',document.forms[0]);
            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
        # make a complete Ajax string for each action
	    $view_str   .= "[" . $form_str . "]); \"";
        $update_str .= "[" . $form_str . "]); \"";
        $delete_str .= "[" . $form_str . "]); \"";
        # put them to the hash with the coresponding actions
        $hash_ref->{view} 	= $view_str;
        $hash_ref->{update} = $update_str;
        $hash_ref->{delete} = $delete_str;
    }
    # open the template for the result table
    my $tmpl_name = $layout . "_result.tmpl";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
    # fill in parameters for the template
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        $apiis->log('debug', $_);
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $result .= "__pjx__" . $row_no;
    $result .= (($row_no eq 0) || ($row_no eq 1)) ? " record" : " records";
    $apiis->log('debug', 'Ajax::search return ' . $result);
    return $result;
}

sub search_microsatellites {
	my ($sql, $form_fields_ref, $layout) = @_;
    $apiis->log('notice',"Ajax::search_microsatellites: " . $sql ."," . $layout);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    my $dna_id = "";
    my $marker = "";
    my $ms_no = 0;
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        if ($hash_ref->{'result_col1'} ne $dna_id){
			$dna_id = $hash_ref->{'result_col1'};
			$marker = $hash_ref->{'result_col4'};
			$row_no++;
			$hash_ref->{'no'} = $row_no;
        }
		else {
			if ($hash_ref->{'result_col4'} eq $marker){
				$hash_ref->{'result_col1'} = "";	
				$hash_ref->{'no'} = "";
			}
			else{
				$row_no++;
				$hash_ref->{'no'} = $row_no;
			}
		}
#		$hash_ref->{'result_col2'} = " ".$seq_no;        
        
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        
        # convert all isodate fields to extdate
        foreach my $key (%{$hash_ref}){
             if ($key =~ /(.*)\_[d|D][t|T]/){
              my $extdate = $apiis->iso2extdate($hash_ref->{$key});
                if (defined $extdate){
                    $hash_ref->{$key} = $extdate;
                }
            }
        }
        push(@record_sets, \%$hash_ref);
		# prepare a list of data fields used to fill data into the form
        my @form_fields;
        for(my $i = 0; $i<@{$form_fields_ref}; $i++){
            $form_fields[$i] = "'" . @{$form_fields_ref}[$i] . "'";
        }
        my $form_str = join(",", @form_fields);
		# prepare Ajax strings
        my $view_str = " onclick=\"
            changeSubmitLabel(\'View\',document.forms[0]);
            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
        my $update_str = " onclick=\"
            changeSubmitLabel(\'Update\',document.forms[0]);
            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
        my $delete_str = " onclick=\"
            changeSubmitLabel(\'Delete\',document.forms[0]);
            external(['action__ajax_fill','sid','layout__$layout','ID__$ID','NO_CACHE'],";
        # make a complete Ajax string for each action
	    $view_str   .= "[" . $form_str . "]); \"";
        $update_str .= "[" . $form_str . "]); \"";
        $delete_str .= "[" . $form_str . "]); \"";
        # put them to the hash with the coresponding actions
        $hash_ref->{view} 	= $view_str;
        $hash_ref->{update} = $update_str;
        $hash_ref->{delete} = $delete_str;
    }
    # open the template for the result table
    my $tmpl_name = $layout . "_result.tmpl";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
    # fill in parameters for the template
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        $apiis->log('debug', $_);
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $result .= "__pjx__" . $row_no;
    $result .= (($row_no eq 0) || ($row_no eq 1)) ? " record" : " records";
    $apiis->log('debug', 'Ajax::search_microsatellites return ' . $result);
    return $result;
}

=head2 fill_form - Retrieve data from database and fill results into the form

This is a Ajax subroutine called after the user clicks on B<View>, B<Update> or
B<Delete> icon in 'Result Part'. It will look for the detail of a given record,
and return the result as a string formated by CGI::Ajax. Normally, the result 
is shown in datafields (data inputs) of the 'Detail Part'.

  ---------------------------------------------------------
  .  Filter1 .......    Filter2 .......   [SEARCH]         . <--- Search Part
  ---------------------------------------------------------
  .                                                        .
  .  Results:                                              .
  .  <Col1>    <Col2>    <Coln>          <Actions>         . <--- Result Part
  .   ...       ...       ...         View|Update|Delete   .
  .   ...       ...       ...         View|Update|Delete   .
  ----------------------------------------------------------
  .  Form                                                  .
  .   -------------------------------------------------    .
  .   .   Label1: [  ]          Label2: [  ]          .    .
  .   .   Label3: [  ]          Label4: [  ]          .    .
  .   .   Labeln: [  ]                                .    . <--- Detail Part
  .   .                                               .    .
  .   .                   [BUTTON]                    .    .
  .   -------------------------------------------------    .


=over 4

=item I<usage>:

    require Ajax;
    my $ajax = &Molab::Ajax::fill_form($sql,$form_ref);

=item I<inputs>:
	
    1. String: SQL query used to retrieve from the database
    2. Array ref: point to a hash of data fields on the form

=item I<output>:

    1. String: A set of values formated in CGI::Ajax as follows
    
    value1__pjx__value2__pjx__value3__pjx__ ... __pjx__valueN 

=back

=cut

sub fill_form {
    my ($layout, $sql, $form_fields_ref) = @_;
    $apiis->log('notice',"Ajax::fill_form: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my $hash_ref = $sql_ref->handle->fetchrow_hashref;
    my $result = "";
    # convert all isodate fields to extdate
    foreach my $key (%{$hash_ref}){
        if ($key =~ /(.*)\_[d|D][t|T]/){
            my $extdate = $apiis->iso2extdate($hash_ref->{$key});
            if (defined $extdate){
                $hash_ref->{$key} = $extdate;
            }
        }
    }
    foreach my $key (@{$form_fields_ref}){
    	$apiis->log('info', "(" . $key . ":" . $hash_ref->{$key} . ")");
        # make drop-down list for storage places
        # this case is applied only for the form 'samples' 
        if (($key =~ /sp1_(\d)_/) and ($hash_ref->{$key})) {
        	my $level = $1;
        	# get all children nodes with the same parent
        	my $parent_id = get_parentID_by_storageID($hash_ref->{$key});
            $apiis->log('debug', "storage place ".$key." level ".$level." with (storage_id,parrent_id)=(".$hash_ref->{$key}.",".$parent_id.")");
            my $select_name = "sp1_" . $1;
            my $select_str = make_storage_combobox($parent_id,$select_name,$hash_ref->{$key});
			$result .= $select_str . "__pjx__";
        }  
        # make drop-down list for other elements  	
        elsif ($key =~ /(.*)\_[l|L][o|O][o|O][p|P]\_[d|D][i|I][v|V]/){
            my $select_id=$1.'_loop';
            $apiis->log('info', "View, select id= " . $select_id);
            my $sql = Molab::Params::get_list_sql($layout,$select_id);
            my $option_str;
            if (defined $sql){
              $option_str= Molab::Ajax::get_dropdown_list($sql,$hash_ref->{$key},$select_id);
            }
            $result .=  $option_str . "__pjx__";
        }
        elsif ($key =~ /(.*)\_[i|I][m|M][a|A][g|G] [e|E]/){
            $result .= $hash_ref->{$key} . "__pjx__";
        }
        else {
            $result .= $hash_ref->{$key} . "__pjx__";
        }
    }
    $apiis->log('info', 'Ajax::fill_form:return ' .$result);
    return ($result);
}

=head2 get_parentID_by_storageID - Get parent ID from the storage ID of a node 

This subroutine returns the parent ID of a node. To use this function you must
supply the storage ID of that node.

=over 4

=item I<usage>:

    require Ajax;
    my $parent_id = &Molab::Ajax::get_parentID_by_storageID($storage_id);

=item I<inputs>:
	
    1. String: Storage ID of the node which you want to find the parent ID

=item I<output>:

    1. String: Parent ID

=back

=cut

sub get_parentID_by_storageID {
    my $storage_id = shift;
    my $sql = "select parent_id from storage where storage_id = $storage_id";
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @data_array = $sql_ref->handle->fetchrow_array;
    my $parent_id = -1;
    $parent_id = $data_array[0] if defined $data_array[0];
    return $parent_id;
}

=head2 get_uniqueID - Get sample ID from the guid in sample 

This subroutine returns the sample ID of a sampple when we know guid of 
that sample.

=over 4

=item I<usage>:

    require Ajax;
    # e.g. we can know where a sample is stored via sample_id
    my $storage_id =  Molab::Ajax::get_uniqueID('storage_samples','storage_id','sample_id',$sample_id);

=item I<inputs>:
	
    1. String: Guid of a given sample

=item I<output>:

    1. String: Sample ID

=back

=cut

sub get_uniqueID {
	my ($table_name,$output_field,$input_field,$input_value) = @_;
    my $sql = "select $output_field from $table_name where $input_field = '$input_value'";
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @data_array = $sql_ref->handle->fetchrow_array;
    my $output_value = -1;
    $output_value = $data_array[0] if defined $data_array[0];
    return $output_value;
}

sub is_empty {
	my ($table_name,$input_field,$input_value) = @_;
    my $sql = "select count(*) from $table_name where $input_field = $input_value";
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @data_array = $sql_ref->handle->fetchrow_array;
    my $output_value = -1;
    $output_value = $data_array[0] if defined $data_array[0];
    $apiis->log('notice', 'COUNT:' . $output_value);
    if ($output_value eq '1'){
    	return 1;
    }
    else{
    	return 0;
    }
}

=head2 get_markers - Get PCR markers in a string

This subroutine returns a string of PCR markers from a given PCR amplification

=over 4

=item I<usage>:

    require Ajax;
    # e.g. we can know where a sample is stored via sample_id
    my $markers =  Molab::Ajax::get_markers('storage_samples','storage_id','sample_id',$sample_id);

=item I<inputs>:
	
    1. String: Guid of a given sample

=item I<output>:

    1. String: Sample ID

=back

=cut

sub get_markers {
	my $ext_amplification_value = shift;
    my $sql;
    $sql  = "SELECT mk.marker_name||'('||cd1.short_name||'-'||cd2.short_name||')' as marker_name  ";
    $sql .= "FROM pcr_amplification pa, pcr_markers pm, markers mk ";
    $sql .= "LEFT JOIN codes cd1 ON mk.db_primer1=cd1.db_code ";
    $sql .= "LEFT JOIN codes cd2 ON mk.db_primer2=cd2.db_code ";
    $sql .= "WHERE pa.amplification_id = pm.amplification_id ";
    $sql .= "AND pm.marker_id = mk.marker_id ";
    $sql .= "AND pa.ext_amplification = '".$ext_amplification_value."'";
    $apiis->log('notice',"SQL:".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my $markers_str = "";
    my $markers_number = 0;
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
    	$markers_number++;
        $markers_str .= $hash_ref->{'marker_name'} . ", ";
    }    
    return ($markers_str,$markers_number);
}

sub get_markers_to_array {
	my $ext_amplification_value = shift;
    my $sql;
    $sql  = "SELECT m.marker_name  ";
    $sql .= "FROM pcr_amplification pa, pcr_markers pm, markers m ";
    $sql .= "WHERE pa.amplification_id = pm.amplification_id ";
    $sql .= "AND pm.marker_id = m.marker_id ";
    $sql .= "AND pa.ext_amplification = '".$ext_amplification_value."'";
    $apiis->log('notice',"SQL:".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @data_array;
    my $markers_number = 0;
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
    	$markers_number++;
        push(@data_array,$hash_ref->{'marker_name'});
    }    
    return (\@data_array,$markers_number);
}

=head2 make_storage_combobox - Make a dropdown list of storage places 

This subroutine creates a HTML string to make a drop-down list of all nodes having the same
parent in the hierarchy of storage places.

=over 4

=item I<usage>:

    require Ajax;
    my $select_string = &Molab::Ajax::make_storage_combobox($parent_id,$select_name,$selected);

=item I<inputs>:
	
    1. String: Parent ID
    2. String: Name of the drop-down list
    2. String: Storage ID will be selected as the defauld option in the drop-down list

=item I<output>:

    1. String: HTML string formatted as the following sample:
    
    <select name='sp_1_loop' id='sp_1_loop'>
      <option value='val1'>name 1</option>
      <option value='val2' selected>name 2</option>
      <option value='val3'>name 3</option>
      ...
      <option value='valN'>name N</option>
    </select> 

=back

=cut

sub make_storage_combobox {
    my ($parent_id,$select_name,$selected_id) = @_;
    my $sql = "SELECT storage_id,storage_name FROM storage WHERE parent_id=$parent_id ORDER BY storage_name";
    $apiis->log('info', "make_storage_combox with SQL= " . $sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @data_array;
    my $select_str = '<select id="'.$select_name.'" name="'.$select_name.'" style="visibility: visible;" class="collect_samples__storage_places" ';
    if ($select_name ne 'sp1_3'){
    	$select_str .= 'onChange="setOptions(this,1)"'; 
    } 
    $select_str .= '>';
	my $option_str = "<option value='0'>select one</option>";
#    my $option_str = "";
    while ( @data_array = $sql_ref->handle->fetchrow_array ) {
    	if ($data_array[0] eq $selected_id){
    		$option_str .= "<option value='$data_array[0]' selected=\"selected\">$data_array[1]</option>";
    	}
    	else {
          $option_str .= "<option value='$data_array[0]'>$data_array[1]</option>";
    	}
    }
	$select_str .= $option_str . '</select>';
    return $select_str;
}

=head2 get_dropdown_list - Make HTML codes for a drop-down list

This subroutine extracts data in database to make HTML codes for a given
drop-down list on the form. Data is extracted from database by using a given SQL
string as the first parameter. The selected option is indentified by the second
parameter. An empty value is used for the default mode (no selected option). The
third paramater provides an identification of the list.

=over 4

=item I<usage>:

    require Ajax;
    my $ajax = &Molab::Ajax::get_dropdown_list($sql,$current_val);

=item I<inputs>:
	
    1. String: SQL string
    2. String: the current option of the drop-down list
    3. String: the name of the drop-down list

=item I<output>:

    1. String: HTML codes

=back

=cut

sub get_dropdown_list{
    my ($sql, $current_opt, $select_id) = @_;
    $apiis->log('notice','Ajax::get_dropdown_list: '.$sql.','.$current_opt);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @col_values = ( );
    my $option_str = "";
    $option_str .= '<option value="-1">select one</option>';
    while ( my @data_array = $sql_ref->handle->fetchrow_array ) {
        my $selected;
        if ($data_array[2] eq $current_opt){
            $selected = " selected=\"selected\"";
        }
        else {
            $selected = "";
        }
        $option_str .= 
          "<option value='$data_array[0]'$selected>$data_array[1]</option>";
    }
    my $HTML_str = '';
    if (defined $select_id){
        $HTML_str = '<select name="' . $select_id . '" id="' . $select_id.'"
                     class="form_input">' . $option_str . '</select>';
    }
    else {
        $HTML_str = $option_str;
    }
    $apiis->log('info', 'Ajax::get_dropdown_list result'. $HTML_str);
    return $HTML_str;
}

=head2 get_stock - Build a drop-down list of storage places

This sub-routine makes a drop-down list of storage places for a given level
in the storage hierarchy. Data is extracted from the table "storage" in the
database. 

=over 4

=item B<input>:

    1. A string of SQL query
    2. A number of the level in the storage hierarchy (0,1,2,3,4)
    3. The name of the drop-down list to be filled

=item B<output>:

    1. A string of HTML codes which are used to fill the drop-down list

=item B<Synopsis>:

    my $ajax = &Molab::Ajax::get_stock($sql,$form_input->{'level'},$select_id);

=back

=cut

sub get_stock {
    my ($sql,$level,$select_id) = @_;
    my $template_level = $level+1;
    my $next_level=$template_level+1;
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->log('notice',"Ajax::get_stock: ".$sql);
    my @data_array;
    my $result = "";
    if ($level == 0){
    	if (defined $select_id and $select_id) {
    		$result  = '<select ';
    		$result .= ' name="'.$select_id.'" ';
    		$result .= ' id="'.$select_id.'" ';
    		$result .= ' class="form_input_storage" ';
    		$result .= ' onchange="external([ 
    			\'action__ajax_get_stock\',
    			\'sid\',
    			\'level__'.$template_level.'\',
    			\'l'.$template_level.'_storage_name_loop\',
    			\'select_id__l'.$next_level.'_storage_name_loop\',
    			\'NO_CACHE\'
    		],[
    			\'l'.$template_level.'_guid\',
    			\'l'.$template_level.'_storage_name\',
    			\'l'.$template_level.'_comment\',
    			\'l'.$next_level.'_parent_id\',
    			\'l'.$next_level.'_storage_name_loop_div\'
    		]);';
    		$result .= 'checkLocation(this);"' .'>'; 
    	};
        my $HTML_str = "<option value='...'>...</option>";
        while ( @data_array = $sql_ref->handle->fetchrow_array ) {
          $HTML_str .= "<option value='$data_array[0]'>$data_array[1]</option>";
        }
        $result .= $HTML_str;
    }
    else{
        @data_array = $sql_ref->handle->fetchrow_array;
        $result .= $data_array[0] . "__pjx__" . $data_array[1] . "__pjx__";
        $result .= $data_array[2] . "__pjx__" . $data_array[3] . "__pjx__";
        if (defined $select_id and $select_id) {
           if ($level >=3) {
                $result .= '<select name="'.$select_id.'" ';
                $result .= ' id="'.$select_id.'" ';
                $result .= ' class="form_input_storage" ';
                $result .= ' onchange="external(
                [
                	\'action__ajax_get_stock\',
                	\'sid\',
                	\'level__'.$template_level.'\',
                	\'l'.$template_level.'_storage_name_loop\',
                	\'select_id__l'.$next_level.'_storage_name_loop\',
                	\'NO_CACHE\'
                ],[
                	\'l'.$template_level.'_guid\',
                	\'l'.$template_level.'_storage_name\',
                	\'l'.$template_level.'_comment\'
                ]);';
                $result .= 'checkLocation(this); "' .'>';
           } else {
                $result .= '<select name="'.$select_id. '" ';
                $result .= ' id="'.$select_id.'" ';
                $result .= ' class="form_input_storage" ';
                $result .= ' onchange="external(
                [
                	\'action__ajax_get_stock\',
                	\'sid\',
                	\'level__'.$template_level.'\',
                	\'l'.$template_level.'_storage_name_loop\',
                	\'select_id__l'.$next_level.'_storage_name_loop\',
                	\'NO_CACHE\'
                ],[
                	\'l'.$template_level.'_guid\',
                	\'l'.$template_level.'_storage_name\',
                	\'l'.$template_level.'_comment\',
                	\'l'.$next_level.'_parent_id\',
                	\'l'.$next_level.'_storage_name_loop_div\'
                ]);';
                $result .= 'checkLocation(this); "' .'>';
           }
        };
        $sql  = "SELECT storage_id,storage_name,storage_id FROM storage ";
        $sql .= "WHERE parent_id= '$data_array[3]' ";
        $sql .= "ORDER BY storage_name";
        my $sql_ref = $apiis->DataBase->user_sql($sql);
        my $HTML_str = "<option value='...'>...</option>";
        while ( @data_array = $sql_ref->handle->fetchrow_array ) {
          $HTML_str .= "<option value='$data_array[0]'>$data_array[1]</option>";
        }
        $result .= $HTML_str;
    }
    if (defined $select_id and $select_id) {
      $result .= '</select>';
    }
    $apiis->log('debug', "Ajax::get_stock returns: " . $result);
    return $result;
}

sub make_form {
	my ($form_input,$session,$cgi) = @_;
    $apiis->log('notice',"Ajax::make_form");
    my $layout = $form_input->{'layout'};
    my $page = $form_input->{'page'};
    my $action = $form_input->{'action'};
    my $sid  = $form_input->{'sid'}; 
    my $samples_number = $form_input->{'samples_number'};
    
    # modify the page number
    if ($action eq "ajax_next"){
    	$page++;
    }
    elsif ($action eq "ajax_back"){
    	$page--;
    }
    
    # open the template and translate labels
    $apiis->log('debug',"opening template ".$layout.$page.".html");
    my $tmpl_name = $layout . $page . ".html";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
	
	# make form STEP 2 for collecting samples
	if ($layout eq "collect_samples"){
		my @record_sets = ( );
		my $ajax_output = "[";
		for (my $row_no=1; $row_no<=$samples_number; $row_no++){
	    	my %hash_ref; 
	    	$hash_ref{'no'} = $row_no;
	        # set styles for rows in the result table
	        if ($row_no % 2){
	        	$hash_ref{'row_style'} = "table_row_even";
	        }
	        else {
	        	$hash_ref{'row_style'} = "table_row_odd";
	        }
	        $hash_ref{'organism_id'} = "organism_id__". $row_no;
	        if ($row_no eq "1"){
	        	$ajax_output .= "'organism_id__$row_no'";
	        }
	        else {
	        	$ajax_output .= ",'organism_id__$row_no'";	
	        }
	        push(@record_sets, \%hash_ref);
	    }
	    $tmpl_form->param( result_loop => \@record_sets);
	    my $ajax_onclick = " onclick= \"submit_form(); external(
          	['action__ajax_upload','layout','sid',
          	'page',
          	'blobfile',
          	'NO_CACHE'] , ";
        $ajax_output  .= "],";
        $ajax_onclick .= $ajax_output;
        $ajax_onclick .= " 'POST' );\"";
        
        $tmpl_form->param( ajax_onclick => $ajax_onclick);
	}
	        
    # fill in parameters for the template
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        if (/^[l|L]\_(.*)/) {
        	$apiis->log('notice', "LABEL" . $_ . ":" .$form_labels{$_});
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    # set hidden values (action and session) in the content template	
	$tmpl_form->param('v_action' => $action);
	$tmpl_form->param('v_session_id' => $sid);
    
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $apiis->log('notice', 'Ajax::next_form return ' . $result);
    return $result;
}

sub upload_file {
	my $cgi = shift;
	my $filename;
	my $file_info;
	my $upload_dir = "/tmp";
	
	$apiis->log("debug", Dumper($cgi));
	if ($cgi->param("blobfile") ne "") {
		
		$filename = $cgi->param("blobfile");
		$file_info=$cgi->uploadInfo($filename);
#		$apiis->log("notice", "file info: " . $file_info);
		# get the file type
		my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
		$filename =~ s/.*[\/\\](.*)/$1/;
		my $upload_filehandle = $cgi->upload("blobfile");
		my $buffer='';
				
		unless (defined $type) {
			open UPLOADFILE, ">$upload_dir/$filename";
			binmode UPLOADFILE;
			while (<$upload_filehandle>){
				print UPLOADFILE;
			}
			close UPLOADFILE;
		}
		else {
			# read the file from the query object file handle
			my $data;
			while (my $bytesread=read($upload_filehandle,$data,1024)) {
				$buffer.=$data;
			}
		}
	}
	my $result = ""; 
	if ($cgi->param("layout") eq "collect_samples") {
		$apiis->log("notice", "Reading from: " . "$upload_dir/$filename");
		my $ref = ReadData ("$upload_dir/$filename", sep => ",");
		$apiis->log("notice", "This file contains ", $ref->[1]{maxcol}, " columns (fields) and ", $ref->[1]{maxrow}, " rows (lines).\n");
		for (my $row = 1; $row <= $ref->[1]{maxrow}; $row++){
			$result .= $ref->[1]{cell}[1][$row] . "__pjx__"
#    		for (my $col = 1; $col <= $ref->[1]{maxcol}; $col++){
#        		print $ref->[1]{cell}[$col][$row], " | ";
#    		}
		}		
	}
	$apiis->log("notice", "Ajax::upload_file returns " . $result);
	return $result;
}

=head2 upload_sequence - Upload a fasta file and return the final sequence

This function is used to upload a fasta file to a temporary folder on the server.
It resurns the content of the file as a final sequence data.

=over 4

=item B<input>:

    1. CGI object
    2. The name of the fasta file

=item B<output>:

    1. Final sequence string

=item B<Synopsis>:

    my $sq_data = &Molab::Ajax::upload_sequence($cgi,$filename);

=back

=cut

sub upload_sequence {
	my ($cgi,$input_file) = @_;
	my $filename;
	my $file_info;
	my $upload_dir = "/tmp";
	my $sq_data;
	if ($cgi->param($input_file) ne "") {
		$filename = $cgi->param($input_file);
		$file_info=$cgi->uploadInfo($filename);
		# get the file type
		my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
		$filename =~ s/.*[\/\\](.*)/$1/;
		my $upload_filehandle = $cgi->upload($input_file);
		$apiis->log("notice", "uploading file: ". $filename);
		open UPLOADFILE, ">$upload_dir/$filename";
		binmode UPLOADFILE;
		while (<$upload_filehandle>){
			print UPLOADFILE;
		}
		close UPLOADFILE;
   		# check if it is a fasta format
		my $filename = "$upload_dir/$filename";
        my $guesser = new Bio::Tools::GuessSeqFormat( -file => $filename );
	    my $data_format  = $guesser->guess;
		$apiis->log('notice',"Format: " .$data_format);
		if ($data_format eq 'fasta'){
            # read sequences from a fasta file
			my $seq_obj  = Bio::SeqIO->new(-file=>$filename,-format=>$data_format);
			my $seq      = $seq_obj->next_seq();
			$sq_data  = $seq->seq;
		} else {
			# read sequences from a text file (without header)
			open FILE, $filename or die "Couldn't open file: $!";
 			binmode FILE;
  			$sq_data = <FILE>;
  			close FILE;
		}
	}
	$apiis->log("notice", "Sequence: " . $sq_data);
	return $sq_data;
}

sub get_samples {
	my ($project_id, $material_type,$layout,$final_concentration,$volume_type,$volume,) = @_;
    $apiis->log('notice',"Ajax::get_samples: ".$project_id.",".$material_type.",".$layout);
    # prepare SQL to retrieve all samples in specified project
    my $sql;
    if ($layout eq "prepare_samples1"){
	    $sql = "select p.project_name as result_col1, c.ext_code as result_col2, s.ext_sample as result_col3 ";
	    $sql   .= "from samples s, sample_collection sc, projects p, codes c ";
	    $sql   .= "where s.sample_id = sc.sample_id ";
	    $sql   .= "and sc.project_id = p.project_id ";
	    $sql   .= "and s.db_material_type = c.db_code ";
	    $sql   .= "and c.ext_code = '" . $material_type. "' ";
	    $sql   .= "and p.project_id = '" . $project_id . "' ";
    }
    elsif ($layout eq "prepare_working_sol1"){
	    $sql = "select s.ext_sample as result_col1, ss.no_units as result_col2, de.dna_concentration as result_col5 ";
	    $sql   .= "from samples s, dna_extraction de, projects p, storage_samples ss, codes c ";
	    $sql   .= "where de.dna_id = s.sample_id ";
	    $sql   .= "and de.project_id = p.project_id ";
	    $sql   .= "and s.db_material_type = c.db_code ";
	    $sql   .= "and de.dna_id = ss.sample_id ";
	    $sql   .= "and c.ext_code = '" . $material_type. "' ";
	    $sql   .= "and p.project_id = '" . $project_id . "' ";
    }
    elsif ($layout eq "amplify_pcr1"){
	    $sql = "select s.ext_sample as result_col1, c.ext_code as result_col2 ";
	    $sql   .= "from samples s, dna_extraction de, projects p, codes c ";
	    $sql   .= "where s.sample_id = de.dna_id ";
	    $sql   .= "and de.project_id = p.project_id ";
	    $sql   .= "and s.db_material_type = c.db_code ";
	    $sql   .= "and c.ext_code = '" . $material_type. "' ";
	    $sql   .= "and p.project_id = '" . $project_id . "' ";
    }
    $apiis->log('notice',"SQL: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        $row_no++;
        $hash_ref->{'no'} = $row_no;
        $hash_ref->{'ext_sample'} = "ext_sample__" . $row_no;
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        $hash_ref->{'selected_sample'} = "selected__".$row_no;
        if ($layout eq "prepare_working_sol1"){
			$hash_ref->{'result_col6'}           = $final_concentration;
			# C1:col5, C2:col6, V1:col7, V2:col8, water:col9
			if ($volume_type eq "initial_volume"){
				$hash_ref->{'result_col7'} = sprintf("%.2f",$volume);
				# calculate V2 = C1xV1/C2
				$hash_ref->{'result_col8'} = sprintf("%.2f",$hash_ref->{'result_col5'}*$hash_ref->{'result_col7'}/$hash_ref->{'result_col6'});
				# water = V2-V1
				$hash_ref->{'result_col9'} = sprintf("%.2f",$hash_ref->{'result_col8'} - $hash_ref->{'result_col7'});
			}
			elsif ($volume_type eq "final_volume"){
				$hash_ref->{'result_col8'} = sprintf("%.2f",$volume);	
				# calculate V1 = C2xV2/C1
				$hash_ref->{'result_col7'} = sprintf("%.2f",$hash_ref->{'result_col6'}*$hash_ref->{'result_col8'}/$hash_ref->{'result_col5'});  
				# water = V2-V1
				$hash_ref->{'result_col9'} = sprintf("%.2f",$hash_ref->{'result_col8'} - $hash_ref->{'result_col7'});
			}
			else {
				$apiis->log('err', "Cannot find the type of volume!")
			}
			$hash_ref->{'result_col3'}           = sprintf("%.2f",$hash_ref->{'result_col7'});
			$hash_ref->{'result_col4'}           = sprintf("%.2f",$hash_ref->{'result_col2'} - $hash_ref->{'result_col3'});
			$hash_ref->{'current_volume'}        = "current_volume__".$row_no;
			$hash_ref->{'current_volume_val'}    = $hash_ref->{'result_col2'};
			$hash_ref->{'ss_concentration'}      = "ss_concentration__".$row_no;
			$hash_ref->{'ss_concentration_val'}  = $hash_ref->{'result_col5'};
			$hash_ref->{'ws_concentration'}      = "ws_concentration__".$row_no;
			$hash_ref->{'ws_concentration_val'}  = $hash_ref->{'result_col6'};
			$hash_ref->{'ss_volume'}             = "ss_volume__".$row_no;
			$hash_ref->{'ss_volume_val'}         = $hash_ref->{'result_col7'};
			$hash_ref->{'ws_volume'}             = "ws_volume__".$row_no;
			$hash_ref->{'ws_volume_val'}         = $hash_ref->{'result_col8'};
			$hash_ref->{'water_volume'}          = "water_volume__".$row_no;
			$hash_ref->{'water_volume_val'}      = $hash_ref->{'result_col9'};
        }
        push(@record_sets, \%$hash_ref);
    }
    # open the template for the result table
    my $tmpl_name = $layout . "_result.tmpl";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
    # fill in parameters for the template
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        $apiis->log('debug', $_);
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    $tmpl_form->param( samples_number => $row_no);
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $apiis->log('debug', 'Ajax::get_samples return ' . $result);
#    $result = "<h1>Hello</h1>";
    return $result;
}

# read data for checking PCR
sub read_data {
	my $sql = shift;
    $apiis->log('notice',"Ajax::read_data: " . $sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my $hash_ref = $sql_ref->handle->fetchrow_hashref;
    return $hash_ref;
}

sub get_samples_for_reusing {
	my ($layout,$filter_name,$project_id,$species,$breed,$reusing_type) = @_;
    
    my $sql;
    if ($reusing_type eq 'SAMPLE'){
    	$sql = "SELECT sp.ext_sample as sample_id, ext_sample
            FROM samples sp, sample_collection sc, organisms og 
              LEFT JOIN codes c1 ON c1.db_code = og.db_species
              LEFT JOIN codes c2 ON c2.db_code = og.db_breed
            WHERE sp.sample_id = sc.sample_id
              AND sp.db_organism = og.db_organism ";
         $sql .= sprintf "  AND project_id::text like %s ",
                 $apiis->DataBase->dbh->quote($project_id);
         $sql .= sprintf "  AND c1.ext_code like %s ",
                 $apiis->DataBase->dbh->quote($species);
         $sql .= sprintf "  AND c2.ext_code like %s ",
                 $apiis->DataBase->dbh->quote($breed);
    }
    if ($reusing_type eq 'DNA'){
    	$sql = "SELECT sp.ext_sample as sample_id, sp.ext_sample
            FROM samples sp, dna_extraction de 
              LEFT JOIN samples sp2 ON sp2.sample_id = de.sample_id
              LEFT JOIN organisms og ON og.db_organism = sp2.sample_id
              LEFT JOIN codes c1 ON c1.db_code = og.db_species
              LEFT JOIN codes c2 ON c2.db_code = og.db_breed
            WHERE sp.sample_id = de.dna_id ";
         $sql .= sprintf "  AND project_id::text like %s ",
                 $apiis->DataBase->dbh->quote($project_id);
         $sql .= sprintf "  AND c1.ext_code like %s ",
                 $apiis->DataBase->dbh->quote($species);
         $sql .= sprintf "  AND c2.ext_code like %s ",
                 $apiis->DataBase->dbh->quote($breed);
    }
	$apiis->log('notice',"Ajax::get_samples_for_reusing: " . $sql);
	
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @col_values = ( );
    my $option_str = "";
    while ( my @data_array = $sql_ref->handle->fetchrow_array ) {
        $option_str .= 
          "<option value='".$data_array[0]."'>$data_array[1]</option>";
    }
    my $HTML_str = '';
    
    $HTML_str = '<select name="'.$filter_name . '" id="' . $filter_name .
                 '" class="form_input" size="10" multiple="multiple">' . $option_str . '</select>';
    $apiis->log('info', 'Ajax::get_samples_for_reusing'. $HTML_str);
    return $HTML_str;
}

sub get_samples_for_electrophoresis {
	my ($project_id, $pcr_name, $layout) = @_;
    $apiis->log('notice',"Ajax::get_samples_for_electrophoresis:");
    $apiis->log('notice', $project_id.",".$pcr_name.",".$layout);
    # prepare SQL to retrieve all samples
    my $sql;
    $sql  = "SELECT sp.ext_sample as result_col2, pa.ext_amplification as result_col1 ";
    $sql .= "FROM samples sp, dna_extraction de, projects pj, codes cd, ";
    $sql .= "     amplified_samples ap, pcr_amplification pa ";
    $sql .= "WHERE sp.sample_id = de.dna_id ";
    $sql .= "AND de.project_id = pj.project_id ";
    $sql .= "AND pa.amplification_id = ap.amplification_id ";
    $sql .= "AND ap.dna_id = de.dna_id ";
    $sql .= "AND sp.db_material_type = cd.db_code ";
    $sql .= "AND cd.ext_code = 'DNA' ";
    $sql .= "AND pj.project_id =" . $project_id . " ";
    $sql .= "AND pa.amplification_id::text like '" . $pcr_name . "' ";
    $sql .= "ORDER BY result_col1, result_col2";
    
    $apiis->log('notice',"SQL: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        $row_no++;
        $hash_ref->{'no'} = $row_no;
        $hash_ref->{'ext_sample'} = "ext_sample__" . $row_no;
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        $hash_ref->{'selected_sample'} = "selected__".$row_no;
        my ($markers_str,$markers_number) = get_markers($hash_ref->{'result_col1'});
        $hash_ref->{'result_col3'} = $markers_str;
       	$hash_ref->{'pcrname'} = "pcrname__" . $row_no;
       	$hash_ref->{'markers'} = "markers__" . $row_no; 
        push(@record_sets, \%$hash_ref);
    }
    # open the template for the result table
    my $tmpl_name = $layout . "_result.tmpl";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
    # fill in parameters for the template
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        $apiis->log('debug', $_);
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    $tmpl_form->param( samples_number => $row_no);
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $apiis->log('debug', 'Ajax::get_samples return ' . $result);
    return $result;
}

# retrieve a list of DNA samples for exporting data 
sub get_dna_samples {
	my ($project_id, $marker_id, $layout) = @_;
    $apiis->log('notice',"Ajax::get_dna_samples:");
    $apiis->log('notice', $project_id.",".$marker_id.",".$layout);
    # prepare SQL to retrieve all samples
    my $sql;
	if ($layout eq 'ms_toolkit'){
	    $sql  = "SELECT DISTINCT pj.project_name as result_col1, ";
	    $sql .= "       cd.short_name   as result_col2, ";
	    $sql .= "       dna.ext_sample  as result_col3 ";
	    $sql .= "FROM projects pj, microsatellites ms, dna_extraction de ";
	    $sql .= " LEFT JOIN samples dna ON de.dna_id = dna.sample_id ";
	    $sql .= " LEFT JOIN samples sp ON de.sample_id = sp.sample_id ";
	    $sql .= " LEFT JOIN organisms og ON sp.db_organism = og.db_organism ";
	    $sql .= " LEFT JOIN codes cd ON og.db_breed = cd.db_code ";
	    $sql .= "WHERE pj.project_id = ms.project_id ";
	    $sql .= "  AND ms.dna_id     = de.dna_id ";
	    $sql .= "  AND pj.project_id::text = '" . $project_id . "' ";
	    $sql .= "ORDER BY result_col3";
	}
	else {
	    $sql  = "SELECT pj.project_name as result_col1, ";
	    $sql .= "       mk.marker_name  as result_col2, ";
	    $sql .= "       sp.ext_sample   as result_col3 ";
	    $sql .= "FROM projects pj, sequences sq, markers mk, samples sp ";
	    $sql .= "WHERE pj.project_id = sq.project_id ";
	    $sql .= "  AND sq.marker_id  = mk.marker_id ";
	    $sql .= "  AND sq.dna_id     = sp.sample_id ";
	    $sql .= "  AND pj.project_id::text = '" . $project_id . "' ";
	    $sql .= "  AND mk.marker_id::text  = '". $marker_id . "' ";
	    $sql .= "ORDER BY result_col3";
	}
    
    $apiis->log('notice',"SQL: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        $row_no++;
        $hash_ref->{'no'} = $row_no;
        if ($layout ne "sequence_formats"){
			$hash_ref->{'ext_sample'} = "ext_sample_" . $row_no;
		}
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        $hash_ref->{'selected_sample'} = "selected__".$row_no;
        push(@record_sets, \%$hash_ref);
    }
    # fill parameters in the template
    my $tmpl_name = $layout . "_result.tmpl";
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    $tmpl_form->param( samples_number => $row_no);
    # return result as a AJAX string
    my $result = $tmpl_form->output;
    $apiis->log('debug', 'Ajax::get_dna_samples return ' . $result);
    return $result;
}

sub get_samples_for_genotyping {
	my ($project_id, $marker_id, $layout) = @_;
    $apiis->log('notice',"Ajax::get_samples_for_genotyping: ".$project_id.",".$marker_id);
    # prepare SQL to retrieve all samples for a given pcr amplification
    my $sql;
    $sql  = "SELECT s.ext_sample as result_col1, ";
    $sql .= "       e.ext_electrophoresis as result_col2, ";
    $sql .= "       m.marker_name as result_col3 ";
    $sql .= "FROM samples s, dna_extraction de, projects p, ";
    $sql .= "  electrophoresises e, analyzed_samples a, markers m ";
    $sql .= "WHERE s.sample_id = de.dna_id ";
    $sql .= "AND de.project_id = p.project_id ";
    $sql .= "AND e.electrophoresis_id = a.electrophoresis_id ";
    $sql .= "AND a.dna_id = de.dna_id ";
    $sql .= "AND m.marker_id = a.marker_id ";
    $sql .= "AND p.project_id::text = '" . $project_id . "' ";
    $sql .= "AND a.marker_id::text = '" . $marker_id . "' "; 
    
    $apiis->log('notice',"SQL: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    my @record_sets = ( );
    my $row_no = 0;		# used to count the number of records
	# process and put data records to the '@record_sets' array 
    while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
        $row_no++;
        $hash_ref->{'no'} = $row_no;
        $hash_ref->{'ext_sample'} = "ext_sample__" . $row_no;
        my $ID = $hash_ref->{'guid'};
        # set styles for rows in the result table
        if ($row_no % 2){
        	$hash_ref->{'row_style'} = "result_even_row";
        }
        else {
        	$hash_ref->{'row_style'} = "result_odd_row";
        }
        $hash_ref->{'selected_sample'} = "selected__".$row_no;
        
       	$hash_ref->{'result_col4'}     = "result__" .$row_no;
       	$hash_ref->{'result_col4_val'} = 1;
        push(@record_sets, \%$hash_ref);
    }
    # open the template for the result table
    my $tmpl_name = $layout . "_result.tmpl";
	$apiis->log('notice',"Open the template: ".$tmpl_name);
    my $tmpl_form = HTML::Template->new(filename=>$tmpl_dir .$tmpl_name);
    # fill in parameters for the template
	my $rs_ref = \@record_sets;
    my %form_labels = %{ Molab::Labels::get_labels(uc($layout))};
    foreach ( $tmpl_form->param ) {
        $apiis->log('debug', $_);
        if (/^[l|L]\_(.*)/) {
            if ( $form_labels{$_} ) {
                $tmpl_form->param( $_ => $form_labels{$_} );
            }
        }
    }
    $tmpl_form->param( result_loop => $rs_ref);
    $tmpl_form->param( samples_number => $row_no);
    # make the AJAX string to return
    my $result = $tmpl_form->output;
    $apiis->log('debug', 'Ajax::get_samples return ' . $result);
    return $result;
}

# get all PCRs within a project 
sub list_of_pcrs {
    my ($select_id,$project_id,$class) = @_;
	my $sql = "";
	$sql .= "SELECT amplification_id, ext_amplification ";
	$sql .= "FROM pcr_amplification pa, projects pj ";
	$sql .= "WHERE pj.project_id = pa.project_id ";
	$sql .= "  AND pj.project_id::text like '" . $project_id ."' ";
    $apiis->log('notice', "Ajax::list_of_pcrs: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    my @data_array;
    
    my $result = "";
    $result = '<select name="' . $select_id . '" id="' . $select_id . '" ';
	if ($class eq ""){
		$result .= '>';
	}else{
		$result .= 'class="'. $class . '" >';
	}
    my $HTML_str = ""; 
	if ($select_id eq "pcr_amplification_loop"){
	   $HTML_str = "<option value='-1'>select one</option>";
	} else {
       $HTML_str = "<option value='%'>All</option>";
	}
    while ( @data_array = $sql_ref->handle->fetchrow_array ) {
    	$HTML_str .= "<option value='$data_array[0]'>$data_array[1]</option>";
    }
    $result .= $HTML_str;
    $result .= '</select>';
    $apiis->log('notice', "Ajax::list_of_pcrs: ".$result);
    return $result;
}

# get all electrophoresises within a project 
sub list_of_electrophoresises {
    my ($select_id,$project_id) = @_;
    $apiis->log('notice', "Ajax::list_of_electrophoresises:");
	my $sql = "";
	$sql .= "SELECT electrophoresis_id, ext_electrophoresis ";
	$sql .= "FROM electrophoresises el, projects pj ";
	$sql .= "WHERE pj.project_id = el.project_id ";
	$sql .= "  AND el.project_id::text = '" . $project_id ."' ";
    $apiis->log('notice', "Ajax::list_of_electrophoresises: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    my @data_array;
    
    my $result = "";
    $result = '<select name="'.$select_id.'" id="'.$select_id.'" class="validate-selection" '.'>';
    my $HTML_str = "<option value='-1'>select one</option>";
    while ( @data_array = $sql_ref->handle->fetchrow_array ) {
    	$HTML_str .= "<option value='$data_array[0]'>$data_array[1]</option>";
    }
    $result .= $HTML_str;
    $result .= '</select>';
    $apiis->log('notice', "Ajax::list_of_electrophoresises: ".$result);
    return $result;
}

# get all markers within a project 
sub list_of_markers {
    my ($select_id,$project_id,$marker_type,$layout) = @_;
    $apiis->log('notice', "Ajax::list_of_markers:");
	my $sql = "";
	if ($marker_type eq 'sq'){
		$sql .= "SELECT DISTINCT mk.marker_id, mk.marker_name ";
		$sql .= "FROM projects pj, sequences sq, markers mk ";
		$sql .= "WHERE pj.project_id = sq.project_id ";
		$sql .= "  AND sq.marker_id = mk.marker_id ";
		$sql .= "  AND pj.project_id::text = '" . $project_id ."' ";
	}
	elsif($marker_type eq 'ms'){
		$sql .= "SELECT DISTINCT mk.marker_id, mk.marker_name ";
		$sql .= "FROM projects pj, microsatellites ms, markers mk ";
		$sql .= "WHERE pj.project_id = ms.project_id ";
		$sql .= "  AND ms.marker_id = mk.marker_id ";
		$sql .= "  AND pj.project_id::text = '" . $project_id ."' ";
	} 
    $apiis->log('notice', "Ajax::list_of_markers: ".$sql);
    my $sql_ref = $apiis->DataBase->user_sql($sql);
    my @data_array;
    
    my $result = "";
    my $HTML_str = "";
    if ($layout eq 'ms_toolkit'){
    	$result = '<select name="'.$select_id.'" id="'.$select_id.'" class="validate-selection" '.' multiple="multiple" size="8">';
    	$HTML_str .= "<option value='-1'>select one or more</option>"; 
    }
    else{
    	$result = '<select name="'.$select_id.'" id="'.$select_id.'" class="validate-selection" '.'>';
    	$HTML_str .= "<option value='-1'>select one</option>"; 
    }
    
    while ( @data_array = $sql_ref->handle->fetchrow_array ) {
    	$HTML_str .= "<option value='$data_array[0]'>$data_array[1]</option>";
    }
    $result .= $HTML_str;
    $result .= '</select>';
    $apiis->log('notice', "Ajax::list_of_markers: ".$result);
    return $result;
}

=head2 compile_report - Compile a report from a given template

The process of compiling a report is carried out as follows:
- read data inputs from Web browser
- compile the report design (XML file was designed in iReport)
- fill values taken in form into report parameters
- generate a new report in the required format (e.g. pdf, html)
- make a link to download that report

=over 4

=item B<input>:

    1. Name of report or layout name
    2. Language (e.g. en, de, vi)
    3. Format of report (e.g. pdf, html, doc)
    4. Path to where the report design stored (input files)
    5. Path to where a new report will be generated (output files)
    4. An array of parameters for filling the report

=item B<output>:

    1. A instant link for downloading the report

=back

=cut

use Inline (
  Java => <<'END_OF_JAVA_CODE',
      import java.util.*;
      import java.io.*;
      import java.util.Date;
      import java.text.SimpleDateFormat;
      import java.sql.Connection;
      import java.sql.DriverManager;
      import java.sql.SQLException;
      import java.util.HashMap;
      import java.util.Locale;
      import net.sf.jasperreports.engine.JREmptyDataSource;
      import net.sf.jasperreports.engine.JRException;
      import net.sf.jasperreports.engine.JasperCompileManager;
      import net.sf.jasperreports.engine.JasperExportManager;
      import net.sf.jasperreports.engine.JasperFillManager;
      import net.sf.jasperreports.engine.JRParameter;
      class Report {
        private Connection connection;
        private String reportName;
        private String reportLanguage;
        private String reportFormat;
        private String inputPath;
        private String outputPath;
        private String[] paraNames;
        private String[] paraValues;
        private String driver;
        private String url;
        private String steps;
        public Report(){}
        public Report(String driver,  String url){
            this.driver = driver;
            this.url    = url;
        }
        public String compileReport(String args[]){
          try
          {
            long start = System.currentTimeMillis();
			long lastPoint = start;
            int paraNumber  = args.length;
            if (paraNumber >= 4){
              // Get the report name and the language
              reportName        = args[0];
              reportLanguage    = args[1];
              reportFormat      = args[2];
              inputPath         = args[3];
              outputPath        = args[4];
              // Read the list of parameters
              paraNumber -= 5;
              if (paraNumber % 2 == 0){
                if (paraNumber > 0){
                  paraNames  = new String[paraNumber/2];
                  paraValues = new String[paraNumber/2];
                  for(int i=0; i<=paraNumber/2; i=i+2){
                    paraNames[i/2] = args[i+5];
                    paraValues[i/2]= args[i+6];
                  }
                }
              }
              else {
                return 
                  "Error: Missing in (Name,Value) used to fill to report";
              }
            }
            else {
              return "Error: Invalid number of parameters";
            }
            Date currentDate = new Date();
            SimpleDateFormat sdf = new SimpleDateFormat ("yyyyMMddHHmmssSS"); 
            String stringStamp = sdf.format(currentDate);
            String reportIn  = inputPath  + reportName;
            String reportOut = outputPath + stringStamp;
            System.err.println("1. Preparing params : " +
                ((System.currentTimeMillis() - lastPoint)/1000.0) + " s");
            lastPoint = System.currentTimeMillis();
            // STEP 1: Compile JasperDesign into JasperReport
            JasperCompileManager.compileReportToFile(
                reportIn + ".jrxml",reportOut + ".jasper");
            System.err.println("2. Compiling report : " +
                ((System.currentTimeMillis() - lastPoint)/1000.0) + " s");
			lastPoint = System.currentTimeMillis();
            // STEP 2: Create a map of parameters to pass to the report
            HashMap params = new HashMap();
            params.put(JRParameter.REPORT_LOCALE, new Locale(reportLanguage));
            for(int i=0; i<paraNumber/2; i++){
               params.put(paraNames[i], paraValues[i]);
            }
            System.err.println("3. Mapping params   : " +
                ((System.currentTimeMillis() - lastPoint)/1000.0) + " s");
			lastPoint = System.currentTimeMillis();				
            // STEP 3: Get a database connection
            Class.forName(driver);
            connection = DriverManager.getConnection(url);
            System.err.println("4. Connecting to DB : " +
                ((System.currentTimeMillis() - lastPoint)/1000.0) + " s");
			lastPoint = System.currentTimeMillis();				
            // STEP 4: Fill the jasper report with parameters done in step 2
            JasperFillManager.fillReportToFile(
                reportOut+".jasper",reportOut+".jrprint",params,connection);
            System.err.println("5. Filling data     : " +
                ((System.currentTimeMillis() - lastPoint)/1000.0) + " s");
			lastPoint = System.currentTimeMillis();
            // STEP 5: Exporting to other formats
            if (reportFormat.equals("pdf")){
              JasperExportManager.exportReportToPdfFile(
                reportOut+".jrprint",reportOut+".pdf");
            }
            else if (reportFormat.equals("html")){
              JasperExportManager.exportReportToHtmlFile(
                reportOut+".jrprint",reportOut+".html");
            }
            else if (reportFormat.equals("xml")){
              JasperExportManager.exportReportToXmlFile(
                reportOut+".jrprint",reportOut+".xml",true);
            }
            else if (reportFormat.equals("csv")){
              JasperExportManager.exportReportToXmlFile(
                reportOut+".jrprint",reportOut+".csv",true);
            }
            System.err.println("6. Exporting file   : " +
                ((System.currentTimeMillis() - lastPoint)/1000.0) + " s");
			lastPoint = System.currentTimeMillis();
            // STEP 6: Close the connection
            connection.close();
            System.err.println("    TOTAL TIME      : " +
                ((System.currentTimeMillis() - start)/1000.0) + " s");
			return (stringStamp + "." + reportFormat);
          }
          catch(Exception e){
            System.err.println("Error: " + steps + e.getMessage());
            return "Error: " + steps + e.getMessage();
          }
        }
      }
END_OF_JAVA_CODE
  SHARED_JVM => 0,
  DIRECTORY => '/tmp',
);

# reporting
sub compile_report {
	my $debug = 1;
    # prepare to make run time for making report
	if ($debug) { open DBG_FILE, ">>", $debug_path; }
	use Time::HiRes::Value;
	my $begin_time = Time::HiRes::Value->now();
	use Data::Dumper;
    my ($report_name, $report_language, $report_format,
        $report_in_path, $report_out_path, $para_names, $para_inputs,$server_name, $user_name, $password) = @_;
    my $result = "";
    my $tbody = "";
    my $driver = "org.postgresql.Driver";
    my $url= "jdbc:postgresql://".$apiis->Model->db_host.":".$apiis->Model->db_port."/".$apiis->Model->db_name;
    $url .= "?user=".$user_name."&password=$password";
    $apiis->log("notice", "extract_report $url");
    my $report = Molab::Ajax::Report->new($driver,$url);
	# time stamp 1
	my $end_time = Time::HiRes::Value->now();
	print DBG_FILE "\n\t"."+ connecting to database: ".$end_time->sub($begin_time). " s" if ($debug);
	my @para_arr;
    push @para_arr, ($report_name,$report_language,$report_format);
    push @para_arr, ($report_in_path,$report_out_path);
    foreach my $key (@{$para_names}){
        $apiis->log("notice", "param ". $key.":".$para_inputs->{$key});
        if ($para_inputs->{$key}) {
	  $apiis->log("notice", "extract_report parameters ". $key ." ".$para_inputs->{$key});
            push @para_arr, ($key,$para_inputs->{$key});
        }
    }
   	# time stamp 2
	$end_time = Time::HiRes::Value->now();
	print DBG_FILE "\n\t"."+ preparing parameters: ".$end_time->sub($begin_time). " s"  if ($debug);
	$apiis->log("debug", "Report parameters: ". Dumper(\@para_arr));
    my $generated_report = $report->compileReport(\@para_arr);
    $apiis->log("info", "Info:" . $generated_report);
    $tbody = q{
        <img src="/images/download_pdf.png" alt="Download"
          border="0" align="absmiddle">
        <a href="http://}.$server_name. q{/cgi-bin/download_report.cgi?report_name=} 
        . $generated_report . q{"><h3>}. $generated_report  .q{</h3> </a> };
    $result .= q{
      <table width="100%"  border="0" cellpadding="1" 
        cellspacing="1" class="fr">
        <tr><td align="center"><br/>} . $tbody . q{<br/></td></tr>
        <tr><td align="center"><br/></td></tr>
      </table>
    };
   	# time stamp 3
	$end_time = Time::HiRes::Value->now();
	print DBG_FILE "\n\t"."+ compiling the report: ".$end_time->sub($begin_time). " s"  if ($debug);
    return ($result);
}

# exporting data to a required format (fasta, nexus dna, ms toolkit ..)
sub export_data {
	use Spreadsheet::Write;
	my ($layout,$format,$file_name,$form_input_ref,$server_name) = @_;
	my $output_dir  = "/tmp/";  # store the output file
	my $result = "";
	my $tbody = "";
	# prepare a unique filename by using time stamp
	$apiis->log("notice","Ajax::export_data('".$layout."')");
	my($sec,$min,$hrs,$day,$month,$year) = (localtime)[0,1,2,3,4,5];
	my $stamp = ($year+1900).($month+1).$day.$hrs.$min.$sec;
	my $output_file;
	if ($file_name ne ""){
		$output_file = $file_name."_".$stamp.".";
	} else {
		$output_file = $stamp.".";
	}
	# the file will be stored in a temp folder at the server
	my $output_path = $output_dir . $output_file .$format;
	my $data_str = "";
	my $selected_samples = " AND sp.ext_sample IN (";
	$apiis->log('notice', "received ajax params:");
	foreach my $key (keys %$form_input_ref) {
		$apiis->log('notice',"\t$key => ". $form_input_ref->{$key});
		if ($key =~ /^ext_sample(.*)/){
			$selected_samples .= "'" .$form_input_ref->{$key} ."',";
		}
	}
	$selected_samples .= "'')";	 # there is a dummy id at the end
    # retrieve sequences to convert data to fasta format
	if ($layout eq 'sequence_formats'){
		my $sql = "";
		$sql .= "SELECT pj.project_id, mk.marker_name, sp.ext_sample, sq.sequence ";
		$sql .= "FROM projects pj, sequences sq, markers mk, samples sp ";
		$sql .= "WHERE pj.project_id = sq.project_id ";
		$sql .= " AND sq.marker_id = mk.marker_id ";
		$sql .= " AND sq.dna_id = sp.sample_id ";
# 		$sql .= " AND pj.project_id = ".$form_input_ref->{'project_name_loop'};
		$sql .= " AND sq.marker_id = " .$form_input_ref->{'marker_loop'};
		$sql .= $selected_samples;
		$apiis->log('notice', "Ajax SQL:".$sql);
		my $sql_ref = $apiis->DataBase->user_sql($sql);
		my @data_array;
		# make a fasta file
		my $fasta_str = "";
		while ( @data_array = $sql_ref->handle->fetchrow_array ) {
			$fasta_str .= ">".$data_array[2] . " | " .$data_array[1] ."\n";
			$fasta_str .= $data_array[3] . "\n";
		}
		my $fasta_file = $output_dir . $output_file."fasta";
		open (FASTAFILE, ">",$fasta_file) or die("Cannot Open File");
		print FASTAFILE $fasta_str;
		close (FASTAILE);
		# small hack to access the file again
		open (FASTAFILE, "<",$fasta_file) or die("Cannot Open File");
		close (FASTAILE);
		if ($format ne 'fasta'){
			my $in  = Bio::AlignIO->new(
				'-file' => $fasta_file,'-format' => 'fasta');
			open (OUTFILE, ">",$output_path) or die("Cannot Open File");
			close (OUTFILE);
			my $out = Bio::AlignIO->new(
				'-file' => ">$output_path",'-format' => $format);
			while ( my $aln = $in->next_aln() ) {
				$out->write_aln($aln);
			}
		}
	}
    elsif ($layout eq 'aligned_sequences'){
		my $fasta_str = $form_input_ref->{'sequence'};
		#make a fasta file
		my $fasta_file = $output_dir . $output_file."fasta";
		open (FASTAFILE, ">",$fasta_file) or die("Cannot Open File");
		print FASTAFILE $fasta_str;
		close (FASTAILE);
		# small hack to access the file again
		open (FASTAFILE, "<",$fasta_file) or die("Cannot Open File");
		close (FASTAILE);
		# check if it is a fasta format
        my $guesser = new Bio::Tools::GuessSeqFormat( -file => $fasta_file );
	    my $data_format  = $guesser->guess;
		$apiis->log('notice',"Format: " .$data_format);
		if ($data_format eq 'fasta'){
			my $in  = Bio::AlignIO->new(
			'-file' => $fasta_file,'-format' => 'fasta');
			open (OUTFILE, ">",$output_path) or die("Cannot Open File");
			close (OUTFILE);
			my $out = Bio::AlignIO->new(
				'-file' => ">$output_path",'-format' => $format);
			while ( my $aln = $in->next_aln() ) {
				$out->write_aln($aln);
			}
		}
		else {
		$result .= q{
			<br/>
			<table width="100%"  border="0" cellpadding="1" cellspacing="1" class="fr">
			<tr><td align="center" class="form_input_error"><br/>Incorrect data format was found! Please input FASTA formatted sequences<br/><br/></td></tr>
			<tr><td align="center"><br/></td></tr>
			</table>
	    	};
			return $result;
		}
	}
	# CREATING THE FILE
	if (($layout eq 'sequence_formats') or ($layout eq 'aligned_sequences')){
		$tbody = q{
		<img src="/images/}.$format.q{.png" alt="Download" border="0" align="absmiddle">
		<a href="http://}.$server_name. q{/cgi-bin/download_report.cgi?report_name=}
		. $output_file.$format . q{"><h3>}. $output_file.$format  .q{</h3> </a> };
    }
    else{
	    $tbody = q{
        <img src="/images/icon_fasta.png" alt="Download" border="0" align="absmiddle">
        <a href="http://}.$server_name. q{/cgi-bin/download_report.cgi?report_name=}
        . $output_file.$format . q{"><h3>}. $output_file.$format  .q{</h3> </a> };
    }
    $result .= q{
      <table width="100%"  border="0" cellpadding="1" cellspacing="1" class="fr">
        <tr><td align="center"><br/>} . $tbody . q{<br/></td></tr>
        <tr><td align="center"><br/></td></tr>
      </table>
    };
    return ($result);
}

# exporting data to MS toolkit format
sub export_mstoolkit {
    use Spreadsheet::WriteExcel;
    my ($layout, $file_name,$file_extension, $form_input_ref, $server_name) = @_;
    my $output_dir  = "/tmp/";  # store the output file
    my $result = "";
    my $tbody = "";
	# prepare a unique filename by using time stamp 
    $apiis->log("notice", "Ajax::export_mstoolkit('" .$layout."')");
    my($sec, $min, $hrs, $day, $month, $year) = (localtime) [0,1,2,3,4,5];
	my $stamp =  ($year+1900).($month+1).$day.$hrs.$min.$sec;
    my $output_file = $file_name."_".$stamp.".".$file_extension;
    # the file will be stored in a temp folder at the server
    my $output_path = $output_dir . $output_file;
    my $data_str = "";
    my $data_format = $form_input_ref->{"data_format_loop"};
	# handle with selected samples
	my $selected_samples = " AND sp.ext_sample IN ("; 
	$apiis->log('notice', "received ajax params:");
	foreach my $key (keys %$form_input_ref) {
		$apiis->log('notice', "\t$key => ". $form_input_ref->{$key});
		if ($key =~ /^ext_sample(.*)/){
			$selected_samples .= "'" .$form_input_ref->{$key} ."',";
		}
	}
	$selected_samples .= "'') ";	 # there is a dummy id at the end
	# create a new Excel workbook
#     my $workbook = Spreadsheet::WriteExcel->new($output_path);
#     my $worksheet = $workbook->add_worksheet($data_format."-column");
# 	$apiis->log('notice', $output_path . " was created");
    my $marker_arr = $form_input_ref->{"marker_loop"};
	$apiis->log('notice', ref($marker_arr));
    my $markers_number = 0;
	if (ref($marker_arr) eq "ARRAY"){
		$markers_number = @$marker_arr;
	} elsif ($marker_arr ne ''){
		$markers_number = 1;
	}
	my $k = 0;
	my $sql1 = "";
	my $sql1_ref;
    my @cells;
	
	# write all selected samples to the first column
 	$sql1 .= "SELECT DISTINCT pj.project_id, sp.ext_sample ";
   	$sql1 .= "FROM projects pj, microsatellites ms, markers mk, samples sp ";
   	$sql1 .= "WHERE pj.project_id = ms.project_id ";
   	$sql1 .= "  AND ms.marker_id = mk.marker_id ";
   	$sql1 .= "  AND ms.dna_id = sp.sample_id ";
   	$sql1 .= "  AND pj.project_id = ". $form_input_ref->{'project_name_loop'};
   	$sql1 .= $selected_samples;
   	$sql1 .= "ORDER BY sp.ext_sample";
	$apiis->log('notice', "SQL1:".$sql1);
 	$sql1_ref = $apiis->DataBase->user_sql($sql1);
	my @data_array1;
	my ($row,$col) = (0,0);
    while ( @data_array1 = $sql1_ref->handle->fetchrow_array() ) {
		$row++;
#     	$worksheet->write_string($row, $col, $data_array1[1]);
		$cells[$row][$col] = $data_array1[1];
    	if ($data_format eq 'diploid-one'){
    		$row++;
    	}
    }
	# make data with each selected marker
	$col = 1;
 	if ($markers_number > 1){
		foreach my $marker_id (@$marker_arr){
			$row = 0;
			# get the marker name from marker_id
			my $sql2 = "SELECT marker_id,marker_name FROM markers WHERE marker_id::text = '".$marker_id."'";
# 			$apiis->log('notice', "Ajax SQL2:".$sql2);
			my $sql2_ref = $apiis->DataBase->user_sql($sql2);
			$apiis->check_status;
			my @data_array2 = $sql2_ref->handle->fetchrow_array;
			my $marker_name = $data_array2[1] if defined $data_array2[1];
			# make the header
			$apiis->log('notice', "Data format:".$data_format);
# 			$worksheet->write_string($row, $col, $marker_name);
			$cells[$row][$col] = $marker_name;
			# get alleles
			my $sql3 = "";
			$sql3 .= "SELECT pj.project_id, mk.marker_name, sp.ext_sample, ms.allele1, ms.allele2 ";
			$sql3 .= "FROM projects pj, microsatellites ms, markers mk, samples sp ";
			$sql3 .= "WHERE pj.project_id = ms.project_id ";
			$sql3 .= "  AND ms.marker_id = mk.marker_id ";
			$sql3 .= "  AND ms.dna_id = sp.sample_id ";
			$sql3 .= "  AND pj.project_id = ". $form_input_ref->{'project_name_loop'};
			$sql3 .= "  AND ms.marker_id = " . $marker_id;
			$sql3 .= $selected_samples;
			$sql3 .= "ORDER BY sp.ext_sample";
			$apiis->log('notice', "Ajax SQL3:".$sql3);
			my $sql_ref3 = $apiis->DataBase->user_sql($sql3);
			$row = 1;
			while ( my @data_array3 = $sql_ref3->handle->fetchrow_array() ) {
				my $allele1 = $data_array3[3];
				my $allele2 = $data_array3[4];
				if ($data_format eq 'diploid-one'){
# 					$worksheet->write_string($row, $col, $allele1);
					$cells[$row][$col] = $allele1;
# 					$worksheet->write_string($row+1, $col, $allele2);
					$cells[$row+1][$col] = $allele2;
				}
				elsif ($data_format eq 'diploid-two'){
# 					$worksheet->write_string($row, $col, $data_array3[3]);
					$cells[$row][$col] = $data_array3[3];
# 					$worksheet->write_string($row, $col+1, $data_array3[4]);
					$cells[$row][$col+1] = $data_array3[4];
				}
				else{
# 					$worksheet->write_string($row, $col, $data_array3[3]);
					$cells[$row][$col] = $data_array3[3];
				}
				if ($data_format eq 'diploid-one'){
					$row = $row + 2;
				}
				else{
					$row = $row + 1;
				}
			}
			if ($data_format eq 'diploid-two'){
				$col = $col + 2;
			}
			else{
				$col = $col + 1;
			}
		}
	} elsif ($markers_number == 1) {
		$row = 0;
		# get the marker name from marker_id
		my $sql2 = "SELECT marker_id,marker_name FROM markers WHERE marker_id = ".$marker_arr;
		$apiis->log('notice', "Ajax SQL2:".$sql2);
		my $sql2_ref = $apiis->DataBase->user_sql($sql2);
		$apiis->check_status;
		my @data_array2 = $sql2_ref->handle->fetchrow_array;
		my $marker_name = $data_array2[1] if defined $data_array2[1];
		# make the header
		$apiis->log('notice', "Data format:".$data_format);
# 		$worksheet->write_string($row, $col, $marker_name);
		$cells[$row][$col] = $marker_name;
		# get alleles
		my $sql3 = "";
		$sql3 .= "SELECT pj.project_id, mk.marker_name, sp.ext_sample, ms.allele1, ms.allele2 ";
		$sql3 .= "FROM projects pj, microsatellites ms, markers mk, samples sp ";
		$sql3 .= "WHERE pj.project_id = ms.project_id ";
		$sql3 .= "  AND ms.marker_id = mk.marker_id ";
		$sql3 .= "  AND ms.dna_id = sp.sample_id ";
		$sql3 .= "  AND pj.project_id = ". $form_input_ref->{'project_name_loop'};
		$sql3 .= "  AND ms.marker_id = " . $marker_arr;
		$sql3 .= $selected_samples;
		$sql3 .= "ORDER BY sp.ext_sample";
		$apiis->log('notice', "Ajax SQL3:".$sql3);
		my $sql_ref3 = $apiis->DataBase->user_sql($sql3);
		$row = 1;
		while ( my @data_array3 = $sql_ref3->handle->fetchrow_array() ) {
			my $allele1 = $data_array3[3];
			my $allele2 = $data_array3[4];
			if ($data_format eq 'diploid-one'){
# 				$worksheet->write_string($row, $col, $allele1);
				$cells[$row][$col] = $allele1;
# 				$worksheet->write_string($row+1, $col, $allele2);
				$cells[$row+1][$col] = $allele2;
			}
			elsif ($data_format eq 'diploid-two'){
# 				$worksheet->write_string($row, $col, $data_array3[3]);
				$cells[$row][$col] = $data_array3[3];
# 				$worksheet->write_string($row, $col+1, $data_array3[4]);
				$cells[$row][$col+1] = $data_array3[4];
			}
			else{
# 				$worksheet->write_string($row, $col, $data_array3[3]);
				$cells[$row][$col] = $data_array3[3];
			}
			if ($data_format eq 'diploid-one'){
				$row = $row + 2;
			}
			else{
				$row = $row + 1;
			}
		}
		if ($data_format eq 'diploid-two'){
			$col = $col + 2;
		}
		else{
			$col = $col + 1;
		}
	}
		
	if ($file_extension eq "csv"){
		open (OUTPUT, ">>$output_path") or die "$!";
		foreach $_ (@cells){
			my $str = join(",",@$_);
        	print OUTPUT "$str\n";
		}
	}
	elsif ($file_extension eq "xls"){
	   	my $workbook = Spreadsheet::WriteExcel->new($output_path);
		my $worksheet = $workbook->add_worksheet($data_format."-column");
		$worksheet->write_col('A1', \@cells);
	}
	$tbody = q{
        <img src="/images/}.$file_extension.q{.png" alt="Download" border="0" align="absmiddle">
        <a href="http://}.$server_name. q{/cgi-bin/download_report.cgi?report_name=} 
        . $output_file . q{"><h3>}. $output_file  .q{</h3> </a> };
    $result .= q{
      <table width="100%"  border="0" cellpadding="1" cellspacing="1" class="fr">
        <tr><td align="center"><br/>} . $tbody . q{<br/></td></tr>
        <tr><td align="center"><br/></td></tr>
      </table>
    };
    return ($result);
}

1;


=head1 AUTHOR

Truong Van Chi Cong <cong.chi@fli.bund.de>

=cut

__END__
