package Molab::LoadObjects;

=head1 NAME

Molab::LoadObjects - Provide a library of Load Objects

=head1 SYNOPSIS

   use LoadObjects;
   # call the load object to handle data for Codes
   my ($err_status,$err_ref)= &LO_codes(\%$form_input);

=head1 DESCRIPTION

This module supports various B<Load Objects> for processing data in the input forms.
Each LO will be linked to one menu item in the menu bar to execute the proper data
function. They are invoked from the Molab::Main module when the user clicks on the 
submit buttons named 'Insert', 'Update', or 'Delete'

=head1 List of Load Objects

  + LO_codes
  + LO_contacts
  + LO_protocols
  + LO_storages
  + LO_projects
  + LO_define_projects
  + LO_organisms
  + LO_insert_organisms
  + LO_samples
  + LO_collect_samples

=cut

###############################################################################
BEGIN {
    use Env qw( APIIS_HOME);
    die "\n\tAPIIS_HOME is not set!\n\n" unless $APIIS_HOME;
    push @INC, "$APIIS_HOME/lib";
}
use warnings;
use strict;
use Env qw( APIIS_HOME );
use Apiis::DataBase::Record;
use Apiis::Init;
use CGI qw/:standard/;
require apiis_alib;
our $statement_nr=0;
our $delete_data_hash;
our $delete_pseudosql;

###############################################################################

=head1 SUBROUTINES


=head2 LO_codes - Process data for Codes

This is the Load Object for inserting, updating, and deleting external codes

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects

=cut

sub LO_codes {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_codes...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            guid
            classname
            ext_code
            short_name
            long_name
            opening_dt
            closing_dt
            description
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        my $today = $apiis->today;
		$data_hash{opening_dt} = $today;
        EXIT:{
            my @pseudo_sql;
            
            if ($data_hash{'bt_submit'} eq "Insert"){
	      		# inserts one record in table codes
	      		# minimum required data: ext_code, class and opening_dt

                $apiis->log('info', "Inserting codes...");
                $pseudo_sql[0] = (
                    'INSERT INTO codes(
                        ext_code,
                        class,
                        short_name,
                        long_name,
                        opening_dt,
                        description
                    )
                    VALUES (
                        $short_name,
                        $classname,
                        $short_name,
                        $long_name,
                        $opening_dt,
                        $description
                    )'
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Update"){
	      		# updates one record in table codes
	      		# all columns are updated
                $apiis->log('info',"Updating codes...");
                $pseudo_sql[0] = (
                    'UPDATE codes SET
                        ext_code        = $short_name,
                        class           = $classname,
                        short_name      = $short_name,
                        long_name       = $long_name,
                        description     = $description
                     WHERE guid = $guid
                    '
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
	        	# deletes one record in table codes
	      		# to be deleted only if it is not referenced by other records
                $apiis->log('info',"Deleting codes...");
                $pseudo_sql[0] = (
                    'DELETE FROM codes
                     WHERE guid = $guid
                    '
                );
            }

            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_markers - Process data for Markers

This is the Load Object for inserting, updating, and deleting markers

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects

=cut

sub LO_markers {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_markers...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            guid
            marker_type
            marker_name
            db_primer1_loop
            db_primer2_loop
            comment
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        my $today = $apiis->today;
		$data_hash{opening_dt} = $today;
        EXIT:{
            my @pseudo_sql;
            
            if ($data_hash{'bt_submit'} eq "Insert"){
	      		# inserts one record in the 'markers' table
                $apiis->log('info', "Inserting markers...");
                $pseudo_sql[0] = (
                    'INSERT INTO markers(
                        marker_name,
                        db_marker_type,
                        db_primer1,
                        db_primer2,
                        comment
                    )
                    VALUES (
                        $marker_name,
                        $marker_type,
                        $db_primer1_loop,
                        $db_primer2_loop,
                        $comment
                    )'
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Update"){
	      		# updates one record in the 'markers' table
	      		# all columns are updated
                $apiis->log('info',"Updating codes...");
                $pseudo_sql[0] = (
                    'UPDATE markers SET
                        marker_name     = $marker_name,
                        db_marker_type  = $marker_type,
                        db_primer1      = $db_primer1_loop,
                        db_primer2      = $db_primer2_loop,
                        comment         = $comment
                     WHERE guid = $guid
                    '
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
	        	# deletes one record in the 'markers' table
	      		# to be deleted only if it is not referenced by other records
                $apiis->log('info',"Deleting codes...");
                $pseudo_sql[0] = (
                    'DELETE FROM markers
                     WHERE guid = $guid
                    '
                );
            }

            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_contacts - Process data for Contacts and Unit

This is the Load Object for inserting, updating, and deleting records in unit and contacts table (in relation 1:1)

  Input:   1. Reference to the data hash with defined keys (@LO_keys)

  Output:  1. Error status (0 for OK, otherwise errors)
           2. An array of error objects

=cut

sub LO_contacts {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors, $hash_ref );
    $apiis->log('info',"Loading LO_contacts...");

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            ext_unit
            ext_id
	    	unit__guid
            phone
            phone2
            fax
            email
            account
            bank_name
            bank_zip
            contacts__guid
            title
            salutation
            birth_dt
            first_name
            second_name
            third_name
            db_language_loop
            street
            town
            db_country_loop
            zip
            label
            comment
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        my $today = $apiis->today;
		my $db_contact=$apiis->DataBase->seq_next_val('seq_contacts__db_contact');
		$data_hash{'db_contact'}=$db_contact;
		$apiis->log('notice', 'db_contact='.$db_contact);

        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Insert"){
	      
	      # inserts one record in contacts and 
	      # one referencing record in unit tables
	      # the order is important: contacts, then unit

                $apiis->log('notice', "Inserting...");
                $pseudo_sql[0] = (
                    'INSERT INTO contacts(
                        db_contact,
                        title,
                        salutation,
                        birth_dt,
                        first_name,
                        second_name,
                        third_name,
                        db_language,
                        phone,
                        phone2,
                        fax,
                        email,
                        account,
                        bank_name,
                        bank_zip,
                        street,
                        town,
                        db_country,
                        zip,
                        label,
                        comment
                    )
                    VALUES (
                        $db_contact,
                        $title,
                        $salutation,
                        $birth_dt,
                        $first_name,
                        $second_name,
                        $third_name,
                        $db_language_loop,
                        $phone,
                        $phone2,
                        $fax,
                        $email,
                        $account,
                        $bank_name,
                        $bank_zip,
                        $street,
                        $town,
                        $db_country_loop,
                        $zip,
                        $label,
                        $comment
                    )'
                );
                $pseudo_sql[1] = (
                    'INSERT INTO unit(
                        ext_unit,
                        ext_id,
                        db_contact
                    )
                    VALUES (
                        $ext_unit,
                        $ext_id,
                        $db_contact
                    )'
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Update"){
	      
	      #updates one record in unit and one in contacts
	      #here the order does not metter

                $apiis->log('info',"Updating...");
                $pseudo_sql[0] = (
                    'UPDATE unit SET
                        ext_unit    = $ext_unit,
                        ext_id      = $ext_id
                     WHERE guid = $unit__guid'
                );
                $pseudo_sql[1] = (
                    'UPDATE contacts SET
                        phone       = $phone,
                        phone2      = $phone2,
                        fax         = $fax,
                        email       = $email,
                        account     = $account,
                        bank_name   = $bank_name,
                        bank_zip    = $bank_zip,
                        title       = $title,
                        salutation  = $salutation,
                        birth_dt    = $birth_dt,
                        first_name  = $first_name,
                        second_name = $second_name,
                        third_name  = $third_name,
                        db_language = $db_language_loop,
                        street      = $street,
                        town        = $town,
                        db_country  = $db_country_loop,
                        zip         = $zip,
                        label       = $label,
                        comment     = $comment
                     WHERE guid = $contacts__guid'
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){

	      #deletes a record from contacts and unit
	      #the referencing record (unit) must be deleted first

                 $apiis->log('info',"Deleting...");
                 $pseudo_sql[0] = (
                    'DELETE FROM unit
                     WHERE guid = $unit__guid
                    '
                 );
                 $pseudo_sql[1] = (
                    'DELETE FROM contacts
                     WHERE guid = $contacts__guid
                    '
                 );
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);

            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_define_projects - Insert data for Projects

This is the Load Object only for inserting a new project

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects

=cut

sub LO_define_projects {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_define_projects...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            project_name
            keyword
            begin_dt
            finish_dt
            description
            comment
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Insert"){
	        	# inserts one record in table PROJECTS
	     		# minimum required data: project_name, keyword
                $apiis->log('info', "Inserting...");
                $pseudo_sql[0] = (
                    'INSERT INTO projects(
						project_name,
						keyword,
						begin_dt,
						finish_dt,
                        description,
                        comment
                    )
                    VALUES (
                        $project_name,
                        $keyword,
                        $begin_dt,
                        $finish_dt,
                        $description,
                        $comment
                    )'
                );
            }

            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_protocols - Process data for Protocols

This is the Load Object for inserting, updating, and deleting protocols

  Input:   1. Reference to the data hash with defined keys (@LO_keys)

  Output:  1. Error status (0 for OK, otherwise errors)
           2. An array of error objects

=cut

=head2 LO_projects - Process data for Projects

This is the Load Object for inserting, updating, and deleting projects

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects

=cut

sub LO_projects {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_projects...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            project_name
            keyword
            begin_dt
            finish_dt
            description
            comment
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Insert"){
	        	# inserts one record in table PROJECTS
	     		# minimum required data: project_name, keyword
                $apiis->log('info', "Inserting projects...");
                $pseudo_sql[0] = (
                    'INSERT INTO projects(
						project_name,
						keyword,
						begin_dt,
						finish_dt,
                        description,
                        comment
                    )
                    VALUES (
                        $project_name,
                        $keyword,
                        $begin_dt,
                        $finish_dt,
                        $description,
                        $comment
                    )'
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Update"){
	      		# updates one record in table projects
	      		# all columns are updated
                $apiis->log('info',"Updating projects...");
                $pseudo_sql[0] = (
                    'UPDATE projects SET
                        project_name    = $project_name,
                        keyword         = $keyword,
                        begin_dt	    = $begin_dt,
                        finish_dt       = $finish_dt,
                        description     = $description,
                        comment         = $comment
                     WHERE guid = $guid
                    '
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
	        	# deletes one record in table codes
	      		# to be deleted only if it is not referenced by other records
                $apiis->log('info',"Deleting projects...");
                $pseudo_sql[0] = (
                    'DELETE FROM projects
                     WHERE guid = $guid
                    '
                );
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_protocols {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info',"Loading LO_protocols...");

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
		my @LO_keys;
		if ($data_hash{'bt_submit'} eq "Insert"){
			@LO_keys = qw (
	            protocol_name
	            db_protocol_type
	            comment
	        );
		} 
		else {
	    	@LO_keys = qw (
	            guid
	            protocol_name
	            db_protocol_type
	            comment
	        );
	    }
    	# some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
        EXIT:{
	        my @pseudo_sql;
	        # BEGIN - prepare SQL to INSERT data
	        if ($data_hash{'bt_submit'} eq "Insert"){
	            # Uploading file and getting a new blobid
		    	if ($data_hash{'blobfile'} ne "") {
					my $upload_dir = "/tmp";
					my $filename = $query->param("blobfile");
					my $file_info= $query->uploadInfo($filename);
					my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
					$filename =~ s/.*[\/\\](.*)/$1/;
	                my $upload_filehandle = $query->upload("blobfile");
					my $buffer='';
					$apiis->log('notice', "Prepare uploading the file $filename,type $type");					
					unless (defined $type) {
			  			open UPLOADFILE, ">$upload_dir/$filename";
			  			binmode UPLOADFILE;
			  			while (<$upload_filehandle>){
			    			print UPLOADFILE;
			  			}
			  			close UPLOADFILE;
					} 
					else {
			  			my $data;
			  			while (my $bytesread=read($upload_filehandle,$data,1024)) {
			    			$buffer.=$data;
			  			}
					}
					#if file content is empty raise error
					if ($buffer eq '') {
				  		$apiis->log('err', "Empty file");
				  		$hash_ref->{err_status}=1;
				  		push @errors,  
				  		Apiis::Errors->new(
							type      => 'DATA',
							severity  => 'ERR',
							ext_fields=> [qw/ blobfile /],
							from      => 'LO_protocols',
							msg_short => __("The file is empty, does not exist, or was not uploaded"),
						);
				  		last EXIT;
					}
					# load file into the BLOBS table
					$apiis->log('debug', 'mimetype: ' . $type);
					my $blobid;
					if (defined $type) {
				  		$blobid = load_blob($buffer,$type);
					} 
					else {
				  		$blobid = load_blob("$upload_dir/$filename");
					}
					if(defined $blobid){
				  		$apiis->log('info', "The protocol was stored to database");
				  		$apiis->log('debug', "blobid=" . $blobid . "will be used to insert to Protocols");
		                $data_hash{protocol_id} = $blobid;
					}
					else {
				  		$apiis->log('err', "Cannot load file to database");
				  		$hash_ref->{err_status}=1;
				  		push @errors, Apiis::Errors->new(
							    type      => 'DATA',
							    severity  => 'ERR',
							    ext_fields=> [qw/ blobfile /],
							    from      => 'LO_protocols',
							    msg_short => "Cannot insert the protocol file in the database",
						);
				  		last EXIT;
					}
	    		} 
		    	else {
					$apiis->log('err', "Protocol file name missing on insert");
					$hash_ref->{err_status}=1;
					push @errors,  Apiis::Errors->new(
							  type      => 'DATA',
							  severity  => 'ERR',
							  ext_fields=> [qw/ blobfile /],
							  from      => 'LO_protocols',
							  msg_short => "Cannot insert protocol without uploaded file",
					);
				  	last EXIT;
		   		}
	            
	            $pseudo_sql[0] = (
	                'INSERT INTO protocols(
	                   protocol_id,
	                   protocol_name,
	                   db_protocol_type,
	                   comment
	                 )
	                 VALUES (
	                   $protocol_id,
	                   $protocol_name,
	                   $db_protocol_type,
	                   $comment
	                 )'
	            );
	        }
	        # END - prepare SQL to INSERT data
	        
	        # BEGIN - prepare SQL to UPDATE data
	        elsif ($data_hash{'bt_submit'} eq "Update"){
	        	$apiis->log('info',"Updating...");
	            # Uploading file and getting a new blobid
		      	if ($data_hash{'blobfile'} ne "") {
					my $upload_dir = "/tmp";
					my $filename = $query->param("blobfile");
					my $file_info= $query->uploadInfo($filename);
					my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
					$filename =~ s/.*[\/\\](.*)/$1/;
			                my $upload_filehandle = $query->upload("blobfile");
					my $buffer='';
					unless (defined $type) {
						open UPLOADFILE, ">$upload_dir/$filename";
					  	binmode UPLOADFILE;
					  	while (<$upload_filehandle>){
					    	print UPLOADFILE;
					  	}
					  	close UPLOADFILE;
					} 
					else {
			  			my $data;
			  			while (my $bytesread=read($upload_filehandle,$data,1024)) {
			    			$buffer.=$data;
			  			}
					}
					#if file content is empty raise error
					if ($buffer eq '') {
			  			$apiis->log('err', "Empty file");
			  			$hash_ref->{err_status}=1;
			  			push @errors,  Apiis::Errors->new(
							    type      => 'DATA',
							    severity  => 'ERR',
							    ext_fields=> [qw/ blobfile /],
							    from      => 'LO_protocols',
							    msg_short => "The file is empty, does not exist, or was not uploaded",
						);
			  			last EXIT;
					}
					# load file into the BLOBS table
					my $protocol_id = $data_hash{protocol_id};
					$apiis->log('info', 'mimetype: ' . $type . $protocol_id);
					my $blobid;
					if (defined $type) {
			  			$blobid = load_blob($buffer,$type,$protocol_id);
					}
					else {
			  			$blobid = load_blob('$upload_dir/$filename',$protocol_id);
					}
					if($blobid eq $protocol_id){
			  			$apiis->log('info', "The protocol was updated to database");
			  			$apiis->log("debug", "blobid=" . $blobid);
					}
					else {
			  			$apiis->log('err', "Cannot load file to database");
			  			$hash_ref->{err_status}=1;
			  			push @errors,  Apiis::Errors->new(
						    type      => 'DATA',
						    severity  => 'ERR',
						    ext_fields=> [qw/ blobfile /],
						    from      => 'LO_protocols',
						    msg_short => "Cannot update the protocol file in the database",
						);
			  			last EXIT;
					}
	      		}
		      	$pseudo_sql[0] = (
					'UPDATE protocols SET 
					 	protocol_name    = $protocol_name,
	                    db_protocol_type = $db_protocol_type,
	                    comment          = $comment
	                 WHERE guid = $guid'
				);
		    }
			# END - prepare SQL to UPDATE data
			
			# BEGIN - prepare SQL to DELETE data
		    elsif ($data_hash{'bt_submit'} eq "Delete"){
		    	$apiis->log('info',"Deleting...");
		      	$pseudo_sql[0] = (
					'DELETE FROM protocols WHERE guid = $guid'
				);
		      	if (exists $data_hash{'protocol_id'} and $data_hash{'protocol_id'}) {
					$pseudo_sql[1] = (
					  'DELETE FROM blobs WHERE blob_id = $protocol_id'
					);
		      	}
	        }
	        # END - prepare SQL to DELETE data
	
	        $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
	        if($hash_ref->{err_status}){
	            $apiis->DataBase->dbh->rollback;
	            push @errors, @{ $hash_ref->{err_ref} };
	        }
	        else {
	        	$apiis->DataBase->dbh->commit;
	        }
      	}
    	return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_storages - Process data for Storage management

This is the Load Object for inserting, updating, and deleting storage locations

  Input:   1. Reference to the data hash with defined keys (@LO_keys)

  Output:  1. Error status (0 for OK, otherwise errors)
           2. An array of error objects

=cut

sub LO_storages {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info',"Loading LO_storages...");

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            l1_guid
            l1_storage_name
            l1_parent_id
            l1_comment
            l2_guid
            l2_storage_name
            l2_parent_id
            l2_comment
            l3_guid
            l3_storage_name
            l3_parent_id
            l3_comment
            l4_guid
            l4_storage_name
            l4_parent_id
            l4_comment
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        EXIT:{
            my @pseudo_sql;

	        #get the value for the new stock_id from sequence
            my $seqval = $apiis->DataBase->seq_next_val('seq_storage__storage_id');
            $data_hash{stock_id} = $seqval;

		    #check which submit button was pressed: 
		    #Storage place (level1) Insert1, Update1,Delete1
		    #Tank (level2) Insert2, Update2,Delete2
		    #Canister (level3) Insert3, Update3,Delete3
		    #Canister level (level4) Insert4, Update4,Delete4

            if ($data_hash{'bt_submit'} =~ /^Insert(\d)/){
	        	#getting the level (1,2,3 or 4)
	        	my $level=$1;
	      
		        # inserts one record in table storage
		        # minimum required data storage_name, parent_id

                $apiis->log('info', "Inserting".$level."...");
                $pseudo_sql[0] = (
                    'INSERT INTO storage(
                        storage_id,
                        storage_name,
                        parent_id,
                        comment
                    )
                    VALUES (
                        $stock_id,
                        $l'.$level.'_storage_name,
                        $l'.$level.'_parent_id,
                        $l'.$level.'_comment
                    )'
                );
            }

            elsif ($data_hash{'bt_submit'} =~ /^Update(\d)/){
	      #getting the level (1,2,3 or 4)
	      my $level=$1;

                $apiis->log('info',"Updating".$level."...");

                $pseudo_sql[0] = (
                    'UPDATE storage SET
                        storage_name  = $l'.$level.'_storage_name,
                        parent_id   = $l'.$level.'_parent_id,
                        comment     = $l'.$level.'_comment
                     WHERE guid = $l'.$level.'_guid'
                );
            }

            elsif ($data_hash{'bt_submit'} =~ /^Delete(\d)$/){
	      
	      # delete a storage record from storage table
	      # and all the records down this hierarchy tree branch
	      # all the delete pseudo statements are generated by 
	      # the delete_storage subroutine

	        my $num=$1;
		my $guid_name='l'.$num.'_guid';
                $apiis->log('info',"Deleting".$num."...");
		my $sql="SELECT storage_id FROM storage WHERE guid=$data_hash{$guid_name}";
		my $sql_ref = $apiis->DataBase->user_sql($sql);
		my $storage_id;
		while ( my @data_array = $sql_ref->handle->fetchrow_array ) {
		  $storage_id=$data_array[0];
		}

		delete_storage($storage_id);

		#push the statements generated by delete_storage in the @pseudo_sql
		foreach my $statement (@$delete_pseudosql) {
		  push @pseudo_sql,$statement;
		};

		#merging  the original data hash and the one
		#from delete_storage
		%data_hash =
		  ( %data_hash, %{ $delete_data_hash } ); 

	    }

	    $hash_ref = meta_db(\@pseudo_sql,\%data_hash);

            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }

    # delete_storage is recursive subroutine that traverses 
    # the tree starting from this node and fills the pseudo 
    # statements in the local  @pseudo array referenced by $pseudo

    sub delete_storage {
      my $storage_id=shift;
      my @storage_ids;

      #get the next storage_ids in the hierarchy
      my $sql="SELECT storage_id FROM storage WHERE parent_id=$storage_id";
      my $sql_ref = $apiis->DataBase->user_sql($sql);
      while ( my @data_array = $sql_ref->handle->fetchrow_array ) {
	push @storage_ids,$data_array[0];
      }

      if (not $storage_ids[0]) {
	#generate storage key by prepending 'a', to 
	#prevent duplication in the data hash
	my $storage_key='storage_id'.'a'.$statement_nr;
	#store this key in separate hash (hashes will be merged later)
        $delete_data_hash->{$storage_key}=$storage_id;
	$delete_pseudosql->[$statement_nr] = (
			  'DELETE FROM storage 
                           WHERE storage_id =$'.$storage_key
			 );
	$statement_nr++;
	return;
      } 

      foreach my $strg (@storage_ids) {
	delete_storage($strg);
      }
      my $storage_key='storage_id'.'a'.$statement_nr;
      $delete_data_hash->{$storage_key}=$storage_id;
      $delete_pseudosql->[$statement_nr] = (
			  'DELETE FROM storage 
                           WHERE storage_id =$'.$storage_key
			 );
      $statement_nr++;
      print STDERR $delete_pseudosql->[$statement_nr],"\n";
      return;
    }
}

=head2 LO_organisms - Process data for organisms

This is the Load Object for inserting, updating, and deleting animals

  Input:   1. Reference to the data hash with defined keys (@LO_keys)

  Output:  1. Error status (0 for OK, otherwise errors)
           2. An array of error objects

=cut
sub LO_organisms {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info',"Loading LO_organisms...");

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys;
        @LO_keys = qw (
        	transfer__guid
            my_guid
			ext_organism
            db_species_loop
            db_breed_loop
            latitude
            longitude
            db_donor_type_loop
            db_sire_loop
            db_dam_loop
            db_sex_loop
            birth_dt
            birth_year
            comment
            image_id
        );
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
        my $seqval;
        $data_hash{opening_dt} = $apiis->today;
#        $apiis->log('debug',"my_guid=". $data_hash{'my_guid'});
		
        EXIT:{
            my @pseudo_sql;
            
            # BEGIN: prepare SQL to insert new organism
            if ($data_hash{'bt_submit'} eq "Insert"){
                $apiis->log('info', "Inserting Organisms...");
                $seqval=$apiis->DataBase->seq_next_val('seq_transfer__db_organism');              
                $data_hash{seqval} = $seqval;
                
                # BEGIN: process file uploading
                if ($data_hash{'blobfile'} ne "") {
                    # Uploading file and getting a new blobid
                    my $upload_dir = "/tmp";
                    my $filename = $query->param("blobfile");
                    my $file_info= $query->uploadInfo($filename);
                    my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
                    $filename =~ s/.*[\/\\](.*)/$1/;
                    my $upload_filehandle = $query->upload("blobfile");

                    my $buffer='';
                    # get the file type
                    unless (defined $type) {
                        open UPLOADFILE, ">$upload_dir/$filename";
                        binmode UPLOADFILE;
                        while (<$upload_filehandle>){
                            print UPLOADFILE;
                        }
                        close UPLOADFILE;
                    }
                    else {
		      			#read the file from the query object file handle
                        my $data;
                        while (my $bytesread=read($upload_filehandle,$data,1024)) {
                            $buffer.=$data;
                        }
                    }

                    #if file content is empty raise error
                    if ($buffer eq '') {
                        $apiis->log('err', "Empty image");
                        $hash_ref->{err_status}=1;
                        push @errors,  Apiis::Errors->new(
                            type      => 'DATA',
                            severity  => 'ERR',
                            ext_fields=> [qw/ blobfile /],
                            from      => 'LO_organisms',
                            msg_short => "The file is empty, does not exist, or was not uploaded",
                        );
                        last EXIT;
                    }
                    # send the file content from file system to Blob table
                    my $blobid;
                    if (defined $type) {
                        $blobid = load_blob($buffer,$type);
                    }
                    else {
                        $blobid = load_blob("$upload_dir/$filename");
                    }
                    # check if image file is loaded successfully
                    if(defined $blobid){
                        $apiis->log("debug", "blobid=" . $blobid);
                        $data_hash{image_id} = $blobid;
                    }
                    else {
		      			#raise error and exit
                        $apiis->log('err', "Cannot insert the image file in the database");
                        $hash_ref->{err_status}=1;
                        push @errors,  Apiis::Errors->new(
                            type      => 'DATA',
                            severity  => 'ERR',
                            ext_fields=> [qw/ blobfile /],
                            from      => 'LO_organisms',
                            msg_short => "Cannot insert the image file in the database",
                        );
                        last EXIT;
                    }
                }
				# END: process file uploading

	            $pseudo_sql[0] = (
	                'INSERT INTO transfer(
	                    db_organism,
	                    db_unit,
	                    ext_organism,
	                    opening_dt
	                )
	                VALUES (
	                    $seqval,
	                    concat("ORGANISMS","ID"),
	                    $ext_organism,
	                    $opening_dt
	                )'
	            );
                if ($data_hash{'blobfile'} eq "") {
			        $pseudo_sql[1] = (
	                    'INSERT INTO organisms(
	                        db_organism,
	                        db_donor_type,
	                        db_breed,
	                        db_species,
	                        latitude,
	                        longitude,
	                        db_sire,
	                        db_dam,
	                        db_sex,
	                        birth_dt,
	                        birth_year,
	                        comment
	                    )
	                    VALUES (
	                        concat("ORGANISMS","ID",$ext_organism),
	                        $db_donor_type_loop,
	                        $db_breed_loop,
	                        $db_species_loop,
	                        $latitude,
	                        $longitude,
		                    concat("ORGANISMS","ID",$db_sire_loop),
	        	            concat("ORGANISMS","ID",$db_dam_loop),
	        	            $db_sex_loop,
		                    $birth_dt,
		                    $birth_year,
	                        $comment
	                    )'
	                );
                }
                else {
			        $pseudo_sql[1] = (
	                    'INSERT INTO organisms(
	                        db_organism,
	                        db_donor_type,
	                        db_breed,
	                        db_species,
	                        latitude,
	                        longitude,
	                        db_sire,
	                        db_dam,
	                        db_sex,
	                        birth_dt,
	                        birth_year,
	                        comment,
	                        db_photo
	                    )
	                    VALUES (
	                        concat("ORGANISMS","ID",$ext_organism),
	                        $db_donor_type_loop,
	                        $db_breed_loop,
	                        $db_species_loop,
	                        $latitude,
	                        $longitude,
		                    concat("ORGANISMS","ID",$db_sire_loop),
	        	            concat("ORGANISMS","ID",$db_dam_loop),
	        	            $db_sex_loop,
		                    $birth_dt,
		                    $birth_year,
	                        $comment,
	                        $image_id
	                    )'
	                );

                }
            }
			# END: prepare SQL to insert new organism
			
			# BEGIN: prepare SQL to update the existing organism
            elsif ($data_hash{'bt_submit'} eq "Update"){
                $apiis->log('info',"Updating Organisms ...");
                $apiis->log('debug', "hash_blobfile:". $data_hash{'blobfile'});
                if ($data_hash{'blobfile'} ne "") {
		  			# in case user have uploaded file
                    my $upload_dir = "/tmp";
                    my $filename = $query->param("blobfile");
                    my $blobid;
                    # check if there is filename in the field
                    # only then try to insert or update animal image
                    if ($filename) {
                        $apiis->log('debug', "Filename:". $filename);
                        my $file_info=$query->uploadInfo($filename);
                        my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
                        $apiis->log('debug', "Type:". $type);
                        $filename =~ s/.*[\/\\](.*)/$1/;
                        my $upload_filehandle = $query->upload("blobfile");
                        my $buffer='';
					unless (defined $type) {
			  			open UPLOADFILE, ">$upload_dir/$filename";
			  			binmode UPLOADFILE;
			  			while (<$upload_filehandle>){
			    			print UPLOADFILE;
			  			}
			  			close UPLOADFILE;
					}
		     		else {
                        my $data;
                        while (my $bytesread=read($upload_filehandle,$data,1024)) {
                            $buffer.=$data;
                        }
                    }
                    # if file content is empty raise error
                    if ($buffer eq '') {
                        $apiis->log('err', "Empty image");
                        $hash_ref->{err_status}=1;
                            push @errors,  Apiis::Errors->new(
                                type      => 'DATA',
                                severity  => 'ERR',
                                ext_fields=> [qw/ blobfile /],
                                from      => 'LO_organisms',
                                msg_short => "The file is empty, does not exist, or was not uploaded",
                            );
                            last EXIT;
                    }
                    # check if there is already existing image_id
                    # if image_id exists load file into the BLOBS table with this image_id (update)
                    # if image_id does not exist load file into the BLOBS table with new image_id (insert)
  		    		my $image_id;
		    		$image_id = $data_hash{image_id};
                    $apiis->log('debug', 'mimetype: ' . $type . $image_id);
                    if ($image_id) {
                        # update the existing image
                        if (defined $type) {
                            $blobid = load_blob($buffer,$type,$image_id);
                        }
                        else {
                            $blobid = load_blob('$upload_dir/$filename',,$image_id);
                        }
                        if($blobid eq $image_id){
                            $apiis->log('info', "The image was updated in the database");
                            $apiis->log("debug", "blobid=" . $blobid);
                        }
                        else {
                            $apiis->log('err', "Cannot update the image file in the database");
                            $hash_ref->{err_status}=1;
                            push @errors,  Apiis::Errors->new(
                                type      => 'DATA',
                                severity  => 'ERR',
                                ext_fields=> [qw/ blobfile /],
                                from      => 'LO_organisms',
                                msg_short => "Cannot update the image file in the database",
                            );
                            last EXIT;
                        }
                    }
                    else {
                        #load as new image
                        if (defined $type) {
                            $blobid = load_blob($buffer,$type);
                        }
                        else {
                            $blobid = load_blob('$upload_dir/$filename');
                        }
                        if($blobid){
                            $apiis->log('debug', "The image was inserted in the database");
                            $apiis->log("debug", "blobid=" . $blobid);
                        }
                        else {
                            $apiis->log('err', "Cannot insert the image file database");
                            $hash_ref->{err_status}=1;
                            push @errors,  Apiis::Errors->new(
                                type      => 'DATA',
                                severity  => 'ERR',
                                ext_fields=> [qw/ blobfile /],
                                from      => 'LO_organisms',
                                msg_short => "Cannot insert the image file in the database",
                            );
                            last EXIT;
                        }
                    }
                    # create error object and exit
                    $data_hash{image_id}=$blobid if ($blobid);
                } # end if defined filename
            } # end if ($data_hash{'blobfile'} ne "")
            $pseudo_sql[0] = (
                'UPDATE transfer SET
                    db_unit      = concat("ORGANISMS","ID"),
                    ext_organism = $ext_organism,
                    opening_dt = $opening_dt
                 WHERE guid = $transfer__guid'
            );
            if ($data_hash{'blobfile'} ne "") {
		      	# if there is image, update the image_id 
	      		# as the user may have not uploaded the image on the 
                $apiis->log('debug',"Updating with image ...");
                $pseudo_sql[1] = (
                    'UPDATE organisms SET
                    	db_donor_type = $db_donor_type_loop,
                        db_sire      = concat("ORGANISMS","ID",$db_sire_loop),
                        db_dam       = concat("ORGANISMS","ID",$db_dam_loop),
                        db_sex       = $db_sex_loop,
                        db_breed     = $db_breed_loop,
                        birth_dt     = $birth_dt,
                        birth_year   = $birth_year,
                        latitude     = $latitude,
                        longitude    = $longitude,
                        db_species   = $db_species_loop,
                        comment	     = $comment,
                        db_photo     = $image_id
                     WHERE guid = $my_guid'
                )
            }
            else {
            	$apiis->log('debug',"Updating without image ...");
                $pseudo_sql[1] = (
                    'UPDATE organisms SET
                        db_donor_type = $db_donor_type_loop,
                        db_sire      = concat("ORGANISMS","ID",$db_sire_loop),
                        db_dam       = concat("ORGANISMS","ID",$db_dam_loop),
                        db_sex       = $db_sex_loop,
                        db_breed     = $db_breed_loop,
                        birth_dt     = $birth_dt,
                        birth_year   = $birth_year,
                        latitude     = $latitude,
                        longitude    = $longitude,
                        db_species   = $db_species_loop,
                        comment	     = $comment
                     WHERE guid = $my_guid'
                 );

            }
        }
        # END: prepare SQL to update the existing organism
        # BEGIN: prepare SQL to delete the existing organism
        elsif ($data_hash{'bt_submit'} eq "Delete"){
        	# delete the record from organisms and related records in transfer and blobs table
        	# note1: the order is important first organisms, then the others
        	# (order of blobs and transfer does not matter)
        	# note2: only one record is deleted from transfer (no multiple channels)
            $apiis->log('info',"Deleting Organisms...");
            my @pseudo_select;
            $pseudo_select[0] = 'SELECT $db_photo FROM organisms WHERE guid = $my_guid';
            $hash_ref = meta_db( \@pseudo_select, \%data_hash );
            if ( $hash_ref->{err_status} ) {    # error handling
                push @errors, @{ $hash_ref->{err_ref} };
                last EXIT;
            }
            %data_hash = ( %data_hash, %{ $hash_ref->{data_ret} } );    # merging hashes
                $pseudo_sql[0] = (
                     'DELETE FROM organisms
                      WHERE guid = $my_guid
                     '
                );
                $pseudo_sql[1] = (
                     'DELETE FROM transfer
                      WHERE guid = $transfer__guid
                     '
                );
                if (exists $data_hash{'image_id'} and $data_hash{'image_id'}) {
                    $pseudo_sql[2] = (
                        'DELETE FROM blobs
                         WHERE blob_id = $image_id
                        '
                    );
                }
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);

            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_insert_organisms - Insert data for organisms

This is the Load Object for inserting, updating, and deleting animals

  Input:   1. Reference to the data hash with defined keys (@LO_keys)

  Output:  1. Error status (0 for OK, otherwise errors)
           2. An array of error objects

=cut
sub LO_insert_organisms {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info',"Loading LO_insert_organisms...");

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys;
        @LO_keys = qw (
        	transfer__guid
            my_guid
			ext_organism
            db_species_loop
            db_breed_loop
            latitude
            longitude
            db_donor_type_loop
            db_sire_loop
            db_dam_loop
            db_sex_loop
            birth_dt
            birth_year
            comment
        );
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
        my $seqval;
        $data_hash{opening_dt} = $apiis->today;
		
        EXIT:{
            my @pseudo_sql;
            
            # BEGIN: prepare SQL to insert new organism
            if ($data_hash{'bt_submit'} eq "Insert"){
                $apiis->log('info', "Inserting Organisms...");
                $seqval=$apiis->DataBase->seq_next_val('seq_transfer__db_organism');              
                $data_hash{seqval} = $seqval;
                
                # BEGIN: process file uploading
                if ($data_hash{'blobfile'} ne "") {
                    # Uploading file and getting a new blobid
                    my $upload_dir = "/tmp";
                    my $filename = $query->param("blobfile");
                    my $file_info=$query->uploadInfo($filename);
                    my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
                    $filename =~ s/.*[\/\\](.*)/$1/;
                    my $upload_filehandle = $query->upload("blobfile");

                    my $buffer='';
                    # get the file type
                    unless (defined $type) {
                        open UPLOADFILE, ">$upload_dir/$filename";
                        binmode UPLOADFILE;
                        while (<$upload_filehandle>){
                            print UPLOADFILE;
                        }
                        close UPLOADFILE;
                    }
                    else {
		      			#read the file from the query object file handle
                        my $data;
                        while (my $bytesread=read($upload_filehandle,$data,1024)) {
                            $buffer.=$data;
                        }
                    }

                    #if file content is empty raise error
                    if ($buffer eq '') {
                        $apiis->log('err', "Empty image");
                        $hash_ref->{err_status}=1;
                        push @errors,  Apiis::Errors->new(
                            type      => 'DATA',
                            severity  => 'ERR',
                            ext_fields=> [qw/ blobfile /],
                            from      => 'LO_organisms',
                            msg_short => "The file is empty, does not exist, or was not uploaded",
                        );
                        last EXIT;
                    }
                    # send the file content from file system to Blob table
                    my $blobid;
                    if (defined $type) {
                        $blobid = load_blob($buffer,$type);
                    }
                    else {
                        $blobid = load_blob("$upload_dir/$filename");
                    }
                    # check if image file is loaded successfully
                    if(defined $blobid){
                        $apiis->log("debug", "blobid=" . $blobid);
                        $data_hash{image_id} = $blobid;
                    }
                    else {
		      			#raise error and exit
                        $apiis->log('err', "Cannot insert the image file in the database");
                        $hash_ref->{err_status}=1;
                        push @errors,  Apiis::Errors->new(
                            type      => 'DATA',
                            severity  => 'ERR',
                            ext_fields=> [qw/ blobfile /],
                            from      => 'LO_organisms',
                            msg_short => "Cannot insert the image file in the database",
                        );
                        last EXIT;
                    }
                }
				# END: process file uploading

	            $pseudo_sql[0] = (
	                'INSERT INTO transfer(
	                    db_organism,
	                    db_unit,
	                    ext_organism,
	                    opening_dt
	                )
	                VALUES (
	                    $seqval,
	                    concat("ORGANISMS","ID"),
	                    $ext_organism,
	                    $opening_dt
	                )'
	            );
                if ($data_hash{'blobfile'} eq "") {
			        $pseudo_sql[1] = (
	                    'INSERT INTO organisms(
	                        db_organism,
	                        db_donor_type,
	                        db_breed,
	                        db_species,
	                        latitude,
	                        longitude,
	                        db_sire,
	                        db_dam,
	                        db_sex,
	                        birth_dt,
	                        birth_year,
	                        comment
	                    )
	                    VALUES (
	                        concat("ORGANISMS","ID",$ext_organism),
	                        $db_donor_type_loop,
	                        $db_breed_loop,
	                        $db_species_loop,
	                        $latitude,
	                        $longitude,
		                    concat("ORGANISMS","ID",$db_sire_loop),
	        	            concat("ORGANISMS","ID",$db_dam_loop),
	        	            $db_sex_loop,
		                    $birth_dt,
		                    $birth_year,
	                        $comment
	                    )'
	                );
                }
                else {
			        $pseudo_sql[1] = (
	                    'INSERT INTO organisms(
	                        db_organism,
	                        db_donor_type,
	                        db_breed,
	                        db_species,
	                        latitude,
	                        longitude,
	                        db_sire,
	                        db_dam,
	                        db_sex,
	                        birth_dt,
	                        birth_year,
	                        comment,
	                        db_photo
	                    )
	                    VALUES (
	                        concat("ORGANISMS","ID",$ext_organism),
	                        $db_donor_type_loop,
	                        $db_breed_loop,
	                        $db_species_loop,
	                        $latitude,
	                        $longitude,
		                    concat("ORGANISMS","ID",$db_sire_loop),
	        	            concat("ORGANISMS","ID",$db_dam_loop),
	        	            $db_sex_loop,
		                    $birth_dt,
		                    $birth_year,
	                        $comment,
	                        $image_id
	                    )'
	                );

                }
            }
			# END: prepare SQL to insert new organism
			
			# BEGIN: prepare SQL to update the existing organism
            elsif ($data_hash{'bt_submit'} eq "Update"){
                $apiis->log('info',"Updating Organisms ...");
                $apiis->log('debug', "hash_blobfile:". $data_hash{'blobfile'});
                if ($data_hash{'blobfile'} ne "") {
		  			# in case user have uploaded file
                    my $upload_dir = "/tmp";
                    my $filename = $query->param("blobfile");
                    my $blobid;
                    # check if there is filename in the field
                    # only then try to insert or update animal image
                    if ($filename) {
                        $apiis->log('debug', "Filename:". $filename);
                        my $file_info=$query->uploadInfo($filename);
                        my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
                        $apiis->log('debug', "Type:". $type);
                        $filename =~ s/.*[\/\\](.*)/$1/;
                        my $upload_filehandle = $query->upload("blobfile");
                        my $buffer='';
					unless (defined $type) {
			  			open UPLOADFILE, ">$upload_dir/$filename";
			  			binmode UPLOADFILE;
			  			while (<$upload_filehandle>){
			    			print UPLOADFILE;
			  			}
			  			close UPLOADFILE;
					}
		     		else {
                        my $data;
                        while (my $bytesread=read($upload_filehandle,$data,1024)) {
                            $buffer.=$data;
                        }
                    }
                    # if file content is empty raise error
                    if ($buffer eq '') {
                        $apiis->log('err', "Empty image");
                        $hash_ref->{err_status}=1;
                            push @errors,  Apiis::Errors->new(
                                type      => 'DATA',
                                severity  => 'ERR',
                                ext_fields=> [qw/ blobfile /],
                                from      => 'LO_organisms',
                                msg_short => "The file is empty, does not exist, or was not uploaded",
                            );
                            last EXIT;
                    }
                    # check if there is already existing image_id
                    # if image_id exists load file into the BLOBS table with this image_id (update)
                    # if image_id does not exist load file into the BLOBS table with new image_id (insert)
  		    		my $image_id;
		    		$image_id = $data_hash{image_id};
                    $apiis->log('debug', 'mimetype: ' . $type . $image_id);
                    if ($image_id) {
                        # update the existing image
                        if (defined $type) {
                            $blobid = load_blob($buffer,$type,$image_id);
                        }
                        else {
                            $blobid = load_blob('$upload_dir/$filename',,$image_id);
                        }
                        if($blobid eq $image_id){
                            $apiis->log('info', "The image was updated in the database");
                            $apiis->log("debug", "blobid=" . $blobid);
                        }
                        else {
                            $apiis->log('err', "Cannot update the image file in the database");
                            $hash_ref->{err_status}=1;
                            push @errors,  Apiis::Errors->new(
                                type      => 'DATA',
                                severity  => 'ERR',
                                ext_fields=> [qw/ blobfile /],
                                from      => 'LO_organisms',
                                msg_short => "Cannot update the image file in the database",
                            );
                            last EXIT;
                        }
                    }
                    else {
                        #load as new image
                        if (defined $type) {
                            $blobid = load_blob($buffer,$type);
                        }
                        else {
                            $blobid = load_blob('$upload_dir/$filename');
                        }
                        if($blobid){
                            $apiis->log('debug', "The image was inserted in the database");
                            $apiis->log("debug", "blobid=" . $blobid);
                        }
                        else {
                            $apiis->log('err', "Cannot insert the image file database");
                            $hash_ref->{err_status}=1;
                            push @errors,  Apiis::Errors->new(
                                type      => 'DATA',
                                severity  => 'ERR',
                                ext_fields=> [qw/ blobfile /],
                                from      => 'LO_organisms',
                                msg_short => "Cannot insert the image file in the database",
                            );
                            last EXIT;
                        }
                    }
                    # create error object and exit
                    $data_hash{image_id}=$blobid if ($blobid);
                } # end if defined filename
            } # end if ($data_hash{'blobfile'} ne "")
            $pseudo_sql[0] = (
                'UPDATE transfer SET
                    db_unit      = concat("ORGANISMS","ID"),
                    ext_organism = $ext_organism,
                    opening_dt = $opening_dt
                 WHERE guid = $transfer__guid'
            );
            if ($data_hash{'blobfile'} ne "") {
		      	# if there is image, update the image_id 
	      		# as the user may have not uploaded the image on the 
                $apiis->log('debug',"Updating with image ...");
                $pseudo_sql[1] = (
                    'UPDATE organisms SET
                    	db_donor_type = $db_donor_type_loop,
                        db_sire      = concat("ORGANISMS","ID",$db_sire_loop),
                        db_dam       = concat("ORGANISMS","ID",$db_dam_loop),
                        db_sex       = $db_sex_loop,
                        db_breed     = $db_breed_loop,
                        birth_dt     = $birth_dt,
                        birth_year   = $birth_year,
                        latitude     = $latitude,
                        longitude    = $longitude,
                        db_species   = $db_species_loop,
                        comment	     = $comment,
                        db_photo     = $image_id
                     WHERE guid = $my_guid'
                )
            }
            else {
            	$apiis->log('debug',"Updating without image ...");
                $pseudo_sql[1] = (
                    'UPDATE organisms SET
                        db_donor_type = $db_donor_type_loop,
                        db_sire      = concat("ORGANISMS","ID",$db_sire_loop),
                        db_dam       = concat("ORGANISMS","ID",$db_dam_loop),
                        db_sex       = $db_sex_loop,
                        db_breed     = $db_breed_loop,
                        birth_dt     = $birth_dt,
                        birth_year   = $birth_year,
                        latitude     = $latitude,
                        longitude    = $longitude,
                        db_species   = $db_species_loop,
                        comment	     = $comment
                     WHERE guid = $my_guid'
                 );

            }
        }
        # END: prepare SQL to update the existing organism
        # BEGIN: prepare SQL to delete the existing organism
        elsif ($data_hash{'bt_submit'} eq "Delete"){
        	# delete the record from organisms and related records in transfer and blobs table
        	# note1: the order is important first organisms, then the others
        	# (order of blobs and transfer does not matter)
        	# note2: only one record is deleted from transfer (no multiple channels)
            $apiis->log('info',"Deleting Organisms...");
            my @pseudo_select;
            $pseudo_select[0] = 'SELECT $db_photo FROM organisms WHERE guid = $my_guid';
            $hash_ref = meta_db( \@pseudo_select, \%data_hash );
            if ( $hash_ref->{err_status} ) {    # error handling
                push @errors, @{ $hash_ref->{err_ref} };
                last EXIT;
            }
            %data_hash = ( %data_hash, %{ $hash_ref->{data_ret} } );    # merging hashes
                $pseudo_sql[0] = (
                     'DELETE FROM organisms
                      WHERE guid = $my_guid
                     '
                );
                $pseudo_sql[1] = (
                     'DELETE FROM transfer
                      WHERE guid = $transfer__guid
                     '
                );
                if (exists $data_hash{'image_id'} and $data_hash{'image_id'}) {
                    $pseudo_sql[2] = (
                        'DELETE FROM blobs
                         WHERE blob_id = $image_id
                        '
                    );
                }
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);

            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_reuse_samples - Insert reused samples to a project 

This is the Load Object for inserting selected samples to a specified project

  Input:   1. Reference to the data hash with defined keys (@LO_keys)

  Output:  1. Error status (0 for OK, otherwise errors)
           2. An array of error objects

=cut
sub LO_reuse_samples {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info',"Loading LO_reuse_samples ...");

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys;
        @LO_keys = qw (
			destination_project_loop
			type_of_reusing_loop
        );
        # get a list of selected samples
        my @sample_arr = $query->param("destination_samples");
		my $sample_number = @sample_arr;
		my $k = 0;
		foreach (@sample_arr){
			$k++;
			push(@LO_keys, 'sample_id__'.$k);
		    my $sample_id = Molab::Ajax::get_uniqueID(
		    'samples','sample_id','ext_sample',$_);
		    $data_hash{'sample_id__'.$k} = $sample_id;	
			$apiis->log('info','sample ID $k :' . $data_hash{'sample_id__'.$k} );
		}
		$data_hash{'reused'} = 'R';
        
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
        my $seqval;
        my $sql_no = 0;
        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'type_of_reusing_loop'} eq 'SAMPLE'){
	            # insert $sample_number records to sample_collection
	            for (my $i=1; $i<=$sample_number; $i++){
			    	$pseudo_sql[$sql_no++] = (
		            	'INSERT INTO sample_collection(
		                  	project_id,
		                  	sample_id,
		                  	reused
		                )
		                VALUES (
		                  	$destination_project_loop,
		                  	$sample_id__' . $i .',
		                  	$reused
		               	)'
	           		);
	            }
            }
            elsif ($data_hash{'type_of_reusing_loop'} eq 'DNA'){
            	
            }
            else{
            	# do nothing            	
            }

            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_collect_samples3 - Insert a set of samples

This is the Load Object only for inserting a set of samples to proper stirage places

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects

=cut

sub LO_collect_samples3 {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    my $sample_id;
    my $storage_id;
    $apiis->log('info','Loading LO_collect_samples3...');
    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
			project_name
			protocol_name
			db_material_type
        );
        my $samples_number = $data_hash{'samples_number'};
        my $group_number = $data_hash{'group_counter'};
		for(my $i=1; $i<= $samples_number; $i++){
			push(@LO_keys, 'ext_organism__'.$i);
			push(@LO_keys, 'ext_sample__'.$i);
			push(@LO_keys, 'volume__'.$i);
			push(@LO_keys, 'db_user__'.$i);
			push(@LO_keys, 'sampling_dt__'.$i);
			push(@LO_keys, 'comment__'.$i);
		}
		for(my $k=1; $k<= $group_number; $k++){
			push(@LO_keys, 'storage_dt__'. $data_hash{'group_no__'.$k});
			push(@LO_keys, 'db_vessel_type__'. $data_hash{'group_no__'.$k});
			push(@LO_keys, 'sp'. $data_hash{'group_no__'.$k}.'_3');
			push(@LO_keys, 'sp'. $data_hash{'group_no__'.$k}.'_4');
		}
		
		$apiis->log('info','add all names to LO_keys');
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
		# the order of inserting: Samples, Storage, Storage_Samples, Collection
        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Insert"){
                $apiis->log('info', "Inserting...");
                for (my $j= 0; $j <$samples_number; $j++){
                	$sample_id = $apiis->DataBase->seq_next_val('seq_samples__sample_id');
                	$data_hash{'sample_id__'.($j+1)} = $sample_id;
                	$apiis->log('notice', 'sample_id:'.$sample_id);
                	$pseudo_sql[$j]= '
                		INSERT INTO samples(
                			sample_id,
                			ext_sample,
                			db_organism,
                			db_vessel_type,
                			db_material_type,
                			db_user,
                			protocol_id,
                			production_dt,
                			storage_dt,
                			comment
                		)
                		VALUES (
                			$sample_id__'. ($j+1) . ',
                			$ext_sample__'. ($j+1) . ',
                			concat("ORGANISMS","ID",$ext_organism__'. ($j+1) . '),
                			$db_vessel_type__' . $data_hash{'group_no__'. ($j+1)} . ',
                			$db_material_type,
                			$db_user__'. ($j+1) . ',
                			$protocol_name,
                			$sampling_dt__'. ($j+1) . ',
                			$storage_dt__' . $data_hash{'group_no__'.($j+1)} . ',
                			$comment__'. ($j+1) . '
                		)
                	'
                }
                # my $beginner = $samples_number;   
                for (my $i= 0; $i < $group_number; $i++){
                	$storage_id = $apiis->DataBase->seq_next_val('seq_storage__storage_id');
                	$apiis->log('notice', 'storage_id:'.$storage_id);
                	$data_hash{'storage_id__'.($i+1)} = $storage_id;
                	$pseudo_sql[$i+$samples_number] = '
                		INSERT INTO storage(
                			storage_id,
                			storage_name,
                			parent_id
                		)
                		VALUES(
                			$storage_id__' .($i+1).',
                			$sp'.($i+1).'_4,
                			$sp'.($i+1).'_3
                		)
                	'
                }

                for (my $k= 0; $k < $samples_number; $k++){
                	$apiis->log('notice', 'sample_ID:'. $data_hash{'sample_id__'.($k+1)});
                	$pseudo_sql[$k+$samples_number+$group_number] = '
                		INSERT INTO storage_samples(
                			storage_id,
                			sample_id,
                			no_units
                		)
                		VALUES(
                			$storage_id__' . $data_hash{'group_no__'. ($k+1)} .',
                			$sample_id__' . ($k+1) .',
                			$volume__' . ($k+1) .'
                		)
                	'
                }
				$data_hash{'reused'} = 'N';
                for (my $l= 0; $l < $samples_number; $l++){
                	$pseudo_sql[$l+$samples_number+$group_number+$samples_number] = '
                		INSERT INTO sample_collection(
                			sample_id,
                			project_id,
                			reused
                		)
                		VALUES(
                			$sample_id__' . ($l+1) .',
                			$project_name,
                			$reused
                		)
                	'
                }
            }

            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_samples - Process data for samples

This Load Object can be used to insert a new sample of a given material type.
It is also used to update or delete an existing sample in the current database.

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects

Effected tables for each type of action as follows:

=over 4

=item Action type: B<Insert>:

    - Insert one record to 'samples'
      data fields (sample_id, ext_sample, db_organism,db_vessel_type, db_material_type,
      db_user, protocol_id, production_dt, storage_dt,comment)
    - Insert one recored to 'storage'
      (storage_id, storage_name, parent_id)
    - Insert one recored to 'storage_samples'
      (storage_id, sample_id, no_units)
    - Insert one recored to 'sample_collection'
      (sample_id, project_id, reused)

=item Action type: B<Update>:
  
    - Get guid of the sample we want to update data
    - Update one record to 'samples' with the proper guid
      (ext_sample, db_organism,db_vessel_type,db_material_type,db_user,
      protocol_id, production_dt, storage_dt,comment)
    - Check if the storage place is exist
      If Not, insert one recored to 'storage' with new location
      (storage_id, storage_name, parent_id)
      Then, check if there is still any sample stored in old location
      If Not, delete the recore of the old location
    - Get the sample id from guid of the sample 
    - Update one recored to 'storage_samples' with the new location
      (storage_id, no_units)
   
=back


=cut

sub LO_samples {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    my $sample_id;
    my $storage_id;
    $apiis->log('info','Loading LO_samples...');
    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
        	sample_guid
			project_name_loop
			ext_organism_loop
			ext_sample
			db_material_type_loop
			production_dt
			protocol_name_loop
			storage_dt
			db_vessel_type_loop
			volume
			db_user_loop
			comment
			sp1_3
			sp1_4
        );
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
		# the order of inserting: Samples, Storage, Storage_Samples, Sample_Collection
        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Insert"){
                $apiis->log('info', "Inserting ...");
               	$sample_id = $apiis->DataBase->seq_next_val('seq_samples__sample_id');
               	$data_hash{'sample_id'} = $sample_id;
                $apiis->log('notice', 'sample_id:'.$sample_id);
               	$pseudo_sql[0]= '
               		INSERT INTO samples(
               			sample_id,
               			ext_sample,
               			db_organism,
               			db_vessel_type,
               			db_material_type,
               			db_user,
               			protocol_id,
               			production_dt,
               			storage_dt,
               			comment
               		)
               		VALUES (
               			$sample_id,
               			$ext_sample,
               			concat("ORGANISMS","ID",$ext_organism_loop),
               			$db_vessel_type_loop,
               			$db_material_type_loop,
               			$db_user_loop,
               			$protocol_name_loop,
               			$production_dt,
               			$storage_dt,
               			$comment
               		)';
                $storage_id = $apiis->DataBase->seq_next_val('seq_storage__storage_id');
                $data_hash{'storage_id'} = $storage_id;
                $pseudo_sql[1] = '
                	INSERT INTO storage(
                		storage_id,
                		storage_name,
                		parent_id
                	)
                	VALUES(
                		$storage_id,
                		$sp1_4,
                		$sp1_3
                	)';
               $pseudo_sql[2] = '
                	INSERT INTO storage_samples(
                		storage_id,
                		sample_id,
                		no_units
                	)
                	VALUES(
                		$storage_id,
                		$sample_id,
                		$volume
                	)';
				$data_hash{'reused'} = 'N';
               	$pseudo_sql[3] = (
               		'INSERT INTO sample_collection(
                		sample_id,
                		project_id,
                		reused
                	)
                	VALUES(
                		$sample_id,
                		$project_name_loop,
                		$reused
                	)'
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Update"){
				$pseudo_sql[0] = (
                    'UPDATE samples SET
               			ext_sample       = $ext_sample,
               			db_organism      = concat("ORGANISMS","ID",$ext_organism_loop),
               			db_vessel_type   = $db_vessel_type_loop,
               			db_material_type = $db_material_type_loop,
               			db_user          = $db_user_loop,
               			protocol_id      = $protocol_name_loop,
               			production_dt    = $production_dt,
               			storage_dt       = $storage_dt,
               			comment          = $comment
                     WHERE guid = $sample_guid'
                );
                my $storage_name = $data_hash{"sp1_4"};
	            my $parent_id  = $data_hash{"sp1_3"};
	            my $storage_id;
	            my $sql_no = 1;
        		# check if the cell name already exists
        		# storage_id= 0:not exist, 1:exist
		        ($storage_id,$err_status, $err_ref) = Molab::Main::check_stockname($storage_name,$parent_id);
				return ( $err_status, $err_ref ) if $err_status;
				# get sample_id from sample_guid
				my $sample_id = Molab::Ajax::get_uniqueID('samples','sample_id','guid',$data_hash{'sample_guid'});
				$apiis->log('notice','sample id:' . $sample_id);
				$data_hash{'sample_id'} = $sample_id;
				# get old location
				my $old_storage_id =  Molab::Ajax::get_uniqueID('storage_samples','storage_id','sample_id',$sample_id);
				$apiis->log('info','old location:' . $old_storage_id);
				if (!$storage_id) {  # not exist
			 		# insert new storage record
			  		$storage_id = $apiis->DataBase->seq_next_val('seq_storage__storage_id');
			  		$data_hash{'storage_id'} = $storage_id;
			  		$apiis->log('info','new location:' . $storage_id);
                	$pseudo_sql[$sql_no] = (
                		'INSERT INTO storage(
	                		storage_id,
	                		storage_name,
	                		parent_id
	                	)
	                	VALUES(
	                		$storage_id,
	                		$sp1_4,
	                		$sp1_3
	                	)'
	                );
	                $sql_no++;
					# update new location for the sample
					$pseudo_sql[$sql_no] = (
	                    'UPDATE storage_samples SET
	                		storage_id = $storage_id,
	                		no_units   = $volume
	                	 WHERE sample_id = $sample_id'
                	);
                	$sql_no++;
                	# check if there is still any sample stored in old location
                	# if Not (i.e. empty location), delete the old location
                	$data_hash{'old_storage_id'} = $old_storage_id;
                	if (Molab::Ajax::is_empty('storage_samples','storage_id',$old_storage_id) eq 1){
	                	$pseudo_sql[$sql_no] = (
	                    	'DELETE FROM storage
	                      	 WHERE storage_id = $old_storage_id'
	                	);
	                	$sql_no++;
                	}
				}
				else {
					$data_hash{'storage_id'} = $storage_id;
					$pseudo_sql[$sql_no] = (
	                    'UPDATE storage_samples SET
	                		storage_id = $storage_id,
	                		no_units   = $volume
	                	 WHERE sample_id = $sample_id'
                	);
				}
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
				# get sample_id from sample_guid
				my $sample_id = Molab::Ajax::get_uniqueID('samples','sample_id','guid',$data_hash{'sample_guid'});
				$data_hash{'sample_id'} = $sample_id;
				# get old location
				my $old_storage_id =  Molab::Ajax::get_uniqueID('storage_samples','storage_id','sample_id',$sample_id);
               	$data_hash{'old_storage_id'} = $old_storage_id;
              	$pseudo_sql[0] = (
	               	'DELETE FROM sample_collection
	                 WHERE sample_id = $sample_id'
	            );
              	$pseudo_sql[1] = (
	               	'DELETE FROM storage_samples
	                 WHERE sample_id = $sample_id'
	            );
              	$pseudo_sql[2] = (
	               	'DELETE FROM samples
	                 WHERE sample_id = $sample_id'
	            );
               	# check if there is still any sample stored in old location
               	# if Not (i.e. empty location), delete the old location

                if (Molab::Ajax::is_empty('storage_samples','storage_id',$old_storage_id) eq 1){
	               	$pseudo_sql[3] = (
	                   	'DELETE FROM storage
	                   	 WHERE storage_id = $old_storage_id'
	               	);
               	}
            }
            
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

=head2 LO_extract_dna4 - Insert a serial of DNA samples for a given project

This Load Object can be used to insert a number of DNA samples for a given
project.

	+ Input:  Reference to the data hash with defined keys (@LO_keys)
	+ Output: Error status and an array of error objects
	
For example, we want to extract 20 DNA samples from 20 separate tissues. Each
10 DNA samples are grouped in a gel for checking the validation of DNA. There 
are 2 invalid samples in the second group. The remanding samples should be 
sent to 3 storage places (6 samples in one group).

=over 4

=item Effected tables: 

    - Insert 18 records of valid samples to the table 'samples'
      where db_material_type = 'DNA'
      data fields (sample_id, ext_sample, db_organism,db_vessel_type,
      db_material_type, db_user, protocol_id, production_dt, storage_dt,comment)
    - Insert 2 records (gel images) to the table 'blobs'
      data fields (blob, db_mimetype)
    - For each blob_id (obtained from above step)
      Insert 10 records to the table 'gel_images' where gel_id = blob_id
      data fields (gel_id, sample_id, lane, valid)
    - Insert 18 records to the the table 'dna_extraction' where reused = 0
      data fields (dna_id, sample_id, project_id, validation_id, dna_concentration,
      dna_purity, reused)
    - Insert 18 recoreds to the table 'storage'
      (storage_id, storage_name, parent_id)
    - Insert 3 recoreds to the table 'storage_samples'
      (storage_id, sample_id, no_units)
   
=back


=cut

sub LO_extract_dna4 {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    my $dna_id;
    my $storage_id;
    $apiis->log('notice','Loading LO_extract_dna4...');
    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
			project_name
			protocol_name
        );
        $data_hash{'db_material_type'} = "DNA";
        $data_hash{'reused'} = "N";
        my $samples_number   = $data_hash{'samples_number'};
        my $gels_number      = $data_hash{'gels_number'};
        my $group_number     = $data_hash{'group_counter'}; 
        my $sample_id;
		for(my $i=1; $i<= $samples_number; $i++){
			push(@LO_keys, 'ext_sample__'.$i);
			push(@LO_keys, 'sample_id__'.$i);
			push(@LO_keys, 'dna_ext_sample__'.$i);
			push(@LO_keys, 'dna_volume__'.$i);
			push(@LO_keys, 'db_user__'.$i);
			push(@LO_keys, 'extraction_dt__'.$i);
			push(@LO_keys, 'dna_concentration__'.$i);
			push(@LO_keys, 'dna_purity__'.$i);
			push(@LO_keys, 'lane__'. $i);
			push(@LO_keys, 'is_valid__'. $i);
			push(@LO_keys, 'comment__'.$i);
			$data_hash{'sample_id__'.$i} = Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$data_hash{'ext_sample__'.$i});
			$apiis->log("notice", "SAMPLE ID:".$data_hash{'sample_id__'.$i} );
		}
		for(my $k=1; $k<= $group_number; $k++){
# 			push(@LO_keys, 'dna_checking_protocol__'. $data_hash{'gel_no__'.$k});
# 			push(@LO_keys, 'dna_checking_comment__'. $data_hash{'gel_no__'.$k});
			push(@LO_keys, 'storage_dt__'. $data_hash{'group_no__'.$k});
			push(@LO_keys, 'db_vessel_type__'. $data_hash{'group_no__'.$k});
			push(@LO_keys, 'sp'. $data_hash{'group_no__'.$k}.'_3');
			push(@LO_keys, 'sp'. $data_hash{'group_no__'.$k}.'_4');
		}
		
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
		# the order of inserting: Samples, Blobs, Gel_images, DNA_extraction, Validation, Storage, Storage_Samples
        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Insert"){
                $apiis->log('info', "Inserting...");
                my $sql_no = 0;
     			# insert all valid samples
                for (my $j= 0; $j <$samples_number; $j++){
                	if ( ($data_hash{'is_valid__'.($j+1)} eq 'Y') or
						 ($data_hash{'is_valid__'.($j+1)} eq 'U') ){
                   		# get new sample_id
	                	$dna_id = $apiis->DataBase->seq_next_val('seq_samples__sample_id');
	                	$data_hash{'dna_id__'.($j+1)} = $dna_id;
	                	$apiis->log('notice', 'dna_id:'.$dna_id);
	                	# insert a sample record with that sample_id
	                	$pseudo_sql[$sql_no++]= '
	                		INSERT INTO samples(
	                			sample_id,
	                			ext_sample,
	                			db_vessel_type,
	                			db_material_type,
	                			db_user,
	                			protocol_id,
	                			production_dt,
	                			storage_dt,
	                			comment
	                		)
	                		VALUES (
	                			$dna_id__'. ($j+1) . ',
	                			$dna_ext_sample__'. ($j+1) . ',
	                			$db_vessel_type__' . $data_hash{'group_no__'. ($j+1)} . ',
	                			$db_material_type,
	                			$db_user__'. ($j+1) . ',
	                			$protocol_name,
	                			$extraction_dt__'. ($j+1) . ',
	                			$storage_dt__' . $data_hash{'group_no__'.($j+1)} . ',
	                			$comment__'. ($j+1) . '
	                		)
	                	'
					}
                } # end for inserting valid samples
				# upload gel files to blobs
				$apiis->log('notice', "Start uploading gels");
                for (my $i= 0; $i < $gels_number; $i++){
		            # load the file into the BLOBS table and getting a new blobid
		    		if ($data_hash{'gel_file__'.($i+1)} ne "") {
						my $upload_dir = "/tmp";
						my $filename = $data_hash{'gel_file__'.($i+1)};
                        # get the single file name without the path
						$filename =~ s/.*[\/\\](.*)/$1/;
                        my $blobid;
			  			$blobid = load_blob("$upload_dir/$filename");
						if(defined $blobid){
				  			$apiis->log('notice', "The gel image was stored to database");
				  			$apiis->log('notice', "blobid=" . $blobid . "will be used to insert to gel_images");
		                	$data_hash{'gel_id__'.($i+1)} = $blobid;
						}
						else {
				  			$apiis->log('err', "Cannot load file to database");
				  			$hash_ref->{err_status}=1;
				  			push @errors, Apiis::Errors->new(
							    type      => 'DATA',
							    severity  => 'ERR',
							    ext_fields=> [qw/ blobfile /],
							    from      => 'LO_extract_dna4',
							    msg_short => "Cannot insert the gel image in the database",
							);
				  			last EXIT;
						}
	    			} 
		    		else {
						$apiis->log('err', "File name missing on insert");
						$hash_ref->{err_status}=1;
						push @errors,  Apiis::Errors->new(
							  type      => 'DATA',
							  severity  => 'ERR',
							  ext_fields=> [qw/ blobfile /],
							  from      => 'LO_extract_dna4',
							  msg_short => "Cannot insert gel image without uploaded file",
						);
				  		last EXIT;
		   			}
                } # end for uploading gel
                $apiis->log('notice', "End loading Gels");
				# insert definition of gel images for samples which were checked
				# all samples (good and bad) should be defined in gels
                $apiis->log('notice', "Number of samples:".$samples_number);
				for (my $i= 0; $i < $samples_number; $i++){
                	$apiis->log('notice', "sample " . $i+1);
					if ($data_hash{'gel_no__'.($i+1)} > 0){
						$pseudo_sql[$sql_no++] = '
							INSERT INTO gel_images(
								gel_id,
								sample_id,
								lane,
								valid
							)
							VALUES(
								$gel_id__' . $data_hash{'gel_no__'.($i+1)} . ',
								$sample_id__' . ($i+1) .',
								$lane__' . ($i+1) .',
								$is_valid__' . ($i+1) .'
							)
						'
                	}
				} # end for inserting gel images

				# insert validation info
				my $validation_id;
				$apiis->log('notice', "Number of Gels: " . $gels_number);
                for (my $i= 0; $i < $gels_number; $i++){
# 					push(@LO_keys, 'dna_checking_protocol__'. $data_hash{'gel_no__'.($i+1)});
#  					push(@LO_keys, 'dna_checking_comment__'. $data_hash{'gel_no__'.($i+1)});
					push(@LO_keys, 'dna_checking_protocol__'.($i+1));
 					push(@LO_keys, 'dna_checking_comment__'.($i+1));
					$validation_id = $apiis->DataBase->seq_next_val('seq_validation__validation_id');
	                $data_hash{'validation_id__'.($i+1)} = $validation_id;
                	$pseudo_sql[$sql_no++] = '
                		INSERT INTO validation(
                			validation_id,
                			gel_id,
                			protocol_id,
                			comment
                		)
                		VALUES(
                			$validation_id__' . ($i+1) . ',
                			$gel_id__' . ($i+1) . ',
                 			$dna_checking_protocol__'.($i+1) .',
                 			$dna_checking_comment__'. ($i+1) .'
						)
                	'
                } # end for inserting gel images
				
				# insert all details for dna extraction of all valid samples
				# keep track of what samples used in each isolation of dna
                for (my $i= 0; $i < $samples_number; $i++){
                	if ($data_hash{'is_valid__'.($i+1)} eq 'Y') {
	                	$pseudo_sql[$sql_no++] = '
	                		INSERT INTO dna_extraction(
	                			dna_id,
	                			sample_id,
	                			project_id,
	                			validation_id,
	                			dna_concentration,
	                			dna_purity,
	                			reused
	                		)
	                		VALUES(
	                			$dna_id__' . ($i+1) . ',
	                			$sample_id__' . ($i+1) . ',
	                			$project_name,
	                			$validation_id__'. $data_hash{'gel_no__'.($i+1)} .',
	                			$dna_concentration__' . ($i+1) . ',
	                			$dna_purity__' . ($i+1) . ',
	                			$reused
	                		)
	                	'
                	} elsif ($data_hash{'is_valid__'.($i+1)} eq 'U') {
	                	$pseudo_sql[$sql_no++] = '
	                		INSERT INTO dna_extraction(
	                			dna_id,
	                			sample_id,
	                			project_id,
	                			dna_concentration,
	                			dna_purity,
	                			reused
	                		)
	                		VALUES(
	                			$dna_id__' . ($i+1) . ',
	                			$sample_id__' . ($i+1) . ',
	                			$project_name,
	                			$dna_concentration__' . ($i+1) . ',
	                			$dna_purity__' . ($i+1) . ',
	                			$reused
	                		)
	                	'
					}
                }
                # insert new locations if any
                for (my $i= 0; $i < $group_number; $i++){
                	$storage_id = $apiis->DataBase->seq_next_val('seq_storage__storage_id');
                	$apiis->log('notice', 'storage_id:'.$storage_id);
                	$data_hash{'storage_id__'.($i+1)} = $storage_id;
                	$pseudo_sql[$sql_no++] = '
                		INSERT INTO storage(
                			storage_id,
                			storage_name,
                			parent_id
                		)
                		VALUES(
                			$storage_id__' .($i+1).',
                			$sp'.($i+1).'_4,
                			$sp'.($i+1).'_3
                		)
                	'
                }
				# insert details for the storage for valid DNA samples 
                for (my $k= 0; $k < $samples_number; $k++){
                	$apiis->log('notice', 'sample_ID:'. $data_hash{'dna_id__'.($k+1)});
                	if ( ($data_hash{'is_valid__'.($k+1)} eq 'Y') or
						 ($data_hash{'is_valid__'.($k+1)} eq 'U') ){
	                	$pseudo_sql[$sql_no++] = '
	                		INSERT INTO storage_samples(
	                			storage_id,
	                			sample_id,
	                			no_units
	                		)
	                		VALUES(
	                			$storage_id__' . $data_hash{'group_no__'. ($k+1)} .',
	                			$dna_id__' . ($k+1) .',
	                			$dna_volume__' . ($k+1) .'
	                		)
	                	'
                	}
                }           	
            } # end inserting

            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        } # end EXIT
    	return( $hash_ref->{err_status},\@errors);    
    } # end ERR_LABEL
}

sub LO_dna {
	my %data_hash = %{ shift() };
	my ($err_status,$err_ref,@errors,$hash_ref,$sample_id,$storage_id);
	$apiis->log('info','Loading LO_dna...');
	ERR_LABEL: {
		##################################################################
		# %data_hash must have the following keys from @LO_keys and the
		# appropriate values from the datastream:
		##################################################################
		my @LO_keys = qw (
			dna_guid
			de_guid
			va_guid
			gi_guid
			project_name_loop
			ext_sample
			ext_dna
			db_vessel_type_loop
			production_dt
			protocol_name_loop
			storage_dt
			dna_volume
			dna_concentration
			dna_purity
			db_user_loop
			comment
			dna_checking_protocol_loop
			lane
			gel_comment
			sp1_3
			sp1_4
		);
		# some basic checks:
		($err_status, $err_ref) = CheckLO(\%data_hash,\@LO_keys);
		return ( $err_status, $err_ref ) if $err_status;
		EXIT:{
			my @pseudo_sql;
			my $sql_no = 0;
			if ($data_hash{'bt_submit'} eq "Insert"){
				$apiis->log('info', "Inserting NOTHING...");
			}
			elsif ($data_hash{'bt_submit'} eq "Update"){
				# update DNA sample information
				$pseudo_sql[$sql_no++] = (
					'UPDATE samples SET
						ext_sample       = $ext_dna,
						db_vessel_type   = $db_vessel_type_loop,
               			db_user          = $db_user_loop,
						protocol_id      = $protocol_name_loop,
						production_dt    = $production_dt,
						storage_dt       = $storage_dt,
						comment          = $comment
					WHERE guid = $dna_guid'
				);
				# update the detail of DNA extraction
				$data_hash{'sample_id'} = Molab::Ajax::get_uniqueID(
				'samples','sample_id','ext_sample',$data_hash{'ext_sample'});
				$pseudo_sql[$sql_no++] = (
					'UPDATE dna_extraction SET
						sample_id         = $sample_id,
						project_id        = $project_name_loop,
						dna_concentration = $dna_concentration,
						dna_purity        = $dna_purity 
					WHERE guid = $de_guid'
				);
				# update DNA sample's storage
				my $storage_name = $data_hash{"sp1_4"};
				my $parent_id    = $data_hash{"sp1_3"};
				my $storage_id;
				# check if the cell name already exists
				# storage_id= 0: none, 1:exist
				($storage_id,$err_status,$err_ref) =
					Molab::Main::check_stockname($storage_name,$parent_id);
				return ($err_status,$err_ref) if $err_status;
				# get dna_id from dna_guid
				my $dna_id = Molab::Ajax::get_uniqueID(
					'samples','sample_id','guid',$data_hash{'dna_guid'});
				$apiis->log('notice','DNA id:' . $dna_id);
				$data_hash{'dna_id'} = $dna_id;
				# get old location
				my $old_storage_id =  Molab::Ajax::get_uniqueID(
					'storage_samples','storage_id','sample_id',$dna_id);
				$apiis->log('notice','old location:'.$old_storage_id);
				if (!$storage_id) {  # not exist
					# insert a new storage record
					$storage_id = next_id("storage","storage_id");
					$data_hash{'storage_id'} = $storage_id;
					$apiis->log('notice','new location:'.$storage_id);
					$pseudo_sql[$sql_no++] = (
						'INSERT INTO storage(
							storage_id,
							storage_name,
							parent_id
						)
						VALUES(
							$storage_id,
							$sp1_4,
							$sp1_3
						)'
					);
					# update new location for the sample
					$pseudo_sql[$sql_no++] = (
						'UPDATE storage_samples SET
							storage_id = $storage_id,
							no_units   = $dna_volume
						WHERE sample_id = $dna_id'
					);
					# check if there is still any sample stored in old location
					# if Not (i.e. empty location), delete the old location
					$data_hash{'old_storage_id'} = $old_storage_id;
					if (Molab::Ajax::is_empty(
						'storage_samples','storage_id',$old_storage_id) eq 1){
						$pseudo_sql[$sql_no++] = (
							'DELETE FROM storage
							WHERE storage_id = $old_storage_id'
						);
					}
				}
				else {  # inserting to the old location
					$data_hash{'storage_id'} = $storage_id;
					$pseudo_sql[$sql_no++] = (
						'UPDATE storage_samples SET
							storage_id = $storage_id,
							no_units   = $dna_volume
							WHERE sample_id = $dna_id'
					);
				}
				
				# update information for DNA validation
				if ($data_hash{'va_guid'} ne ""){
					$pseudo_sql[$sql_no++] = (
						'UPDATE validation SET
							protocol_id  = $dna_checking_protocol_loop,
							comment          = $gel_comment
						WHERE guid = $va_guid'
					);
					# update information for DNA validation
					$pseudo_sql[$sql_no++] = (
						'UPDATE gel_images SET
							lane  = $lane
						WHERE guid = $gi_guid'
					);
				}
				
			}
			elsif ($data_hash{'bt_submit'} eq "Delete"){
				# get sample_id from sample_guid
				my $dna_id = Molab::Ajax::get_uniqueID(
					'samples','sample_id','guid',$data_hash{'dna_guid'});
				$data_hash{'dna_id'} = $dna_id;
				# get old location
				my $old_storage_id =  Molab::Ajax::get_uniqueID(
					'storage_samples','storage_id','sample_id',$dna_id);
				$data_hash{'old_storage_id'} = $old_storage_id;
				$pseudo_sql[$sql_no++] = (
					'DELETE FROM storage_samples
					WHERE sample_id = $dna_id'
				);
				$pseudo_sql[$sql_no++] = (
					'DELETE FROM samples
					WHERE guid = $dna_guid'
				);
				$pseudo_sql[$sql_no++] = (
					'DELETE FROM dna_extraction
					WHERE guid = $de_guid'
				);
				$pseudo_sql[$sql_no++] = (
					'DELETE FROM gel_images
					WHERE guid = $gi_guid'
				);
				# check if there is still any sample stored in old location
				# if Not (i.e. empty location), delete the old location
				if (Molab::Ajax::is_empty(
					'storage_samples','storage_id',$old_storage_id) eq 1){
					$pseudo_sql[$sql_no++] = (
						'DELETE FROM storage
						WHERE storage_id = $old_storage_id'
					);
				}
			}
			$hash_ref = meta_db(\@pseudo_sql,\%data_hash);
			if($hash_ref->{err_status}){
				$apiis->DataBase->dbh->rollback;
				push @errors, @{ $hash_ref->{err_ref} };
			}
			else { $apiis->DataBase->dbh->commit; }
		}
		return( $hash_ref->{err_status},\@errors);
	}
}

sub LO_amplify_pcr2 {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_amplify_pcr2...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            pcr_amplification_id
            project_name_loop
            protocol_name_loop
            comment
        );
        # get a number of samples
        my $samples_number = 0;
    	if (defined $data_hash{'samples_number'}){
    		$samples_number = $data_hash{'samples_number'};
    		$apiis->log("notice", "A number of samples: ".$samples_number);
    	}
    	# convert ext_sample to dna_id and put to hash
        for(my $i=1; $i<= $samples_number; $i++){
			push(@LO_keys, 'dna_id__'.$i);
			$data_hash{'dna_id__'.$i} = Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$data_hash{"ext_sample__$i"});
        }
        # get a number of markers
        my @marker_arr = $query->param("markers_loop");
		my $markers_number = @marker_arr;
		my $k = 0;
		foreach (@marker_arr){
			$k++;
			push(@LO_keys, 'marker_id__'.$k);
			$data_hash{'marker_id__'.$k} = $_;
		}

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

#        my $today = $apiis->today;
#		$data_hash{opening_dt} = $today;
        EXIT:{
            my @pseudo_sql;
            my $sql_no = 0;
      		# inserts one record into the 'dna_extraction'
            $apiis->log('info', "Inserting pcr_amplification...");
            # get new amplification id
            my $amplification_id = $apiis->DataBase->seq_next_val('seq_pcr_amplification__amplification_id');
            $data_hash{'amplification_id'} = $amplification_id;
            $pseudo_sql[$sql_no++] = (
            	'INSERT INTO pcr_amplification(
                  	amplification_id,
                    ext_amplification,
                    project_id,
                    protocol_id,
                    comment
                )
                VALUES (
                  	$amplification_id,
                    $pcr_amplification_id,
                    $project_name_loop,
                    $protocol_name_loop,
                    $comment
                )'
            );
            
            # insert N records (selected samples) to amplifed_samples
            for (my $i=1; $i<=$samples_number; $i++){
				if (defined $data_hash{"selected__$i"}){
			    	$pseudo_sql[$sql_no++] = (
		            	'INSERT INTO amplified_samples(
		                  	amplification_id,
		                  	dna_id
		                )
		                VALUES (
		                  	$amplification_id,
		                  	$dna_id__' . $i .'
	                	)'
            		);
				}
            }
            # insert $markers_number records to pcr_markers
            for (my $i=1; $i<=$markers_number; $i++){
		    	$pseudo_sql[$sql_no++] = (
	            	'INSERT INTO pcr_markers(
	                  	amplification_id,
	                  	marker_id
	                )
	                VALUES (
	                  	$amplification_id,
	                  	$marker_id__' . $i .'
	               	)'
           		);
            }
                    
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_check_pcr_products2 {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_check_pcr_products2...');
    my $samples_number   = $data_hash{'samples_number'};
    my $gels_number      = $data_hash{'gels_number'};
    my $sample_id;
    
    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            pcr_amplification
            project_name
            protocol_name
            comment
        );
		for(my $i=1; $i<= $samples_number; $i++){
			push(@LO_keys, 'dna_id__'.$i);
			push(@LO_keys, 'lane__'. $i);
			push(@LO_keys, 'is_valid__'. $i);
			$data_hash{'dna_id__'.$i} = Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$data_hash{'ext_sample__'.$i});
			if (defined $data_hash{'valid__'.$i}){
				$data_hash{'is_valid__'.$i} ='Y';
			}
			else{
				$data_hash{'is_valid__'.$i} ='N';
			}
		}
		for(my $i=1; $i<= $gels_number; $i++){
			push(@LO_keys, 'checking_protocol__'.$i);
			push(@LO_keys, 'comment_protocol__'.$i);
		}

	    # Upload Gels
	    my $gels_number = $data_hash{'gels_number'};
	    EXIT:
    	{
	        my @pseudo_sql;
	        my $sql_no = 0;
	    	$apiis->log('notice', "Uploading " .$gels_number ." gels...");	            
	     	# upload gel files to server stored in /tmp
	        for (my $i= 0; $i < $gels_number; $i++){
		     	if ( $data_hash{'gel_file__'.($i+1)} ne "") {
					my $upload_dir = "/tmp";
					my $filename = $query->param("gel_file__".($i+1));
					$apiis->log('notice', "Prepare uploading the file $filename");
					my $file_info= $query->uploadInfo($filename);
					my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
					$filename =~ s/.*[\/\\](.*)/$1/;
	               	my $upload_filehandle = $query->upload("gel_file__".($i+1));
					my $buffer='';
					open UPLOADFILE, ">$upload_dir/$filename";
					binmode UPLOADFILE;
					while (<$upload_filehandle>){
		  				print UPLOADFILE;
					}
					close UPLOADFILE;
					$apiis->log('notice', $filename." was uploaded to ".$upload_dir);
	    		} 
		    	else {
					$apiis->log('err', "File name missing on insert");
					$hash_ref->{err_status}=1;
					push @errors,  Apiis::Errors->new(
						  type      => 'DATA',
						  severity  => 'ERR',
						  ext_fields=> [qw/ blobfile /],
						  from      => 'LO_check_pcr_products2',
						  msg_short => "Cannot insert gel image without uploaded file",
					);
					last EXIT;
		   		}
	        } # end for uploading gels to /tmp
			
			# upload gel files to blobs
            $apiis->log('notice', "Loading " .$gels_number ." gels to BLOB");
            for (my $i= 0; $i < $gels_number; $i++){
		        # load the file into the BLOBS table and getting a new blobid
		    	if ($data_hash{'gel_file__'.($i+1)} ne "") {
		    		$apiis->log('notice', $data_hash{'gel_file__'.($i+1)}." was found");
					my $upload_dir = "/tmp";
					my $filename = $data_hash{'gel_file__'.($i+1)};
                    # get the single file name without the path
                    $filename =~ s/.*[\/\\](.*)/$1/;
					my $blobid;
			 		$blobid = load_blob("$upload_dir/$filename");
					if(defined $blobid){
			  			$apiis->log('notice', "The gel image was stored to database");
			  			$apiis->log('notice', "blobid (".$blobid.") will be used to insert to gel_images");
		               	$data_hash{'gel_id__'.($i+1)} = $blobid;
					}
					else {
			  			$apiis->log('err', "Cannot load file to database");
			  			$hash_ref->{err_status}=1;
			  			push @errors, Apiis::Errors->new(
						    type      => 'DATA',
						    severity  => 'ERR',
						    ext_fields=> [qw/ blobfile /],
						    from      => 'LO_check_pcr_products2',
						    msg_short => "Cannot insert the gel image in the database",
						);
			  			last EXIT;
					}
	    		} 
		    	else {
					$apiis->log('err', "File name missing on insert");
					$hash_ref->{err_status}=1;
					push @errors,  Apiis::Errors->new(
						  type      => 'DATA',
						  severity  => 'ERR',
						  ext_fields=> [qw/ blobfile /],
						  from      => 'LO_check_pcr_products2',
						  msg_short => "Cannot insert gel image without uploaded file",
					);
			  		last EXIT;
		   		}
            } # end for uploading gels to blob
            
			# define gel images for DNA which were checked
			$apiis->log('notice', "Defining gel images (dna/lane)");
			for (my $i= 0; $i < $samples_number; $i++){
				if ($data_hash{'gel_no__'.($i+1)} > 0) {
					$pseudo_sql[$sql_no++] = '
						INSERT INTO gel_images(
							gel_id,
							sample_id,
							lane,
							valid
						)
						VALUES(
							$gel_id__' . $data_hash{'gel_no__'.($i+1)} . ',
							$dna_id__' . ($i+1) .',
							$lane__' . ($i+1) .',
							$is_valid__' . ($i+1) .'
						)
					';
				}
            } # end for defining gel images
			# insert validation info
			my $validation_id;
            for (my $i= 0; $i < $gels_number; $i++){
               	$validation_id = $apiis->DataBase->seq_next_val('seq_validation__validation_id');
               	$apiis->log('notice', "validation ID = ".$validation_id);
	            $data_hash{'validation_id__'.($i+1)} = $validation_id;
               	$pseudo_sql[$sql_no++] = '
               		INSERT INTO validation(
               			validation_id,
               			gel_id,
               			protocol_id,
               			comment
               		)
               		VALUES(
               			$validation_id__' . ($i+1) . ',
               			$gel_id__' . ($i+1) . ',
               			$checking_protocol__'. $data_hash{'gel_no__'.($i+1)} .',
                		$checking_comment__'. $data_hash{'gel_no__'.($i+1)} .'
                	)
                '
             } # end for inserting validation info
			 # update validation id for all samples in pcr_amplification
             for (my $i= 0; $i < $samples_number; $i++){
                $pseudo_sql[$sql_no++] = '
	            	UPDATE amplified_samples SET
	                	validation_id =	$validation_id__'. $data_hash{'gel_no__'.($i+1)} .'
	                WHERE 
	                	amplification_id = $amplification_id AND dna_id = $dna_id__'.($i+1);
             }
             
	        $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
	        if($hash_ref->{err_status}){
	            $apiis->DataBase->dbh->rollback;
	            push @errors, @{ $hash_ref->{err_ref} };
	        }
	        else {
	            $apiis->DataBase->dbh->commit;
	        }
    	} # end EXIT
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_pcr {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_pcr...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
        	pa_guid
        	va_guid
        	pcr_amplification_id
            project_name_loop
            pcr_amplification_protocol_loop
            pcr_validation_protocol_loop
            marker_loop
            pcr_comment
            gel_comment
            gel_file
        );
        # some basic checks:
        ($err_status,$err_ref) = CheckLO(\%data_hash,\@LO_keys);
        return ($err_status,$err_ref) if $err_status;
        
        EXIT:{
            my @pseudo_sql;
            my $sql_no = 0;
            if ($data_hash{'bt_submit'} eq "Insert"){
	        	# do nothing
            }
            elsif ($data_hash{'bt_submit'} eq "Update"){
            	my $gel_id = upload_file($query,'gel_file');
				if ($gel_id eq 0){
					$hash_ref->{err_status}=1;
					push @errors, Apiis::Errors->new(
			    		type      => 'DATA',
			    		severity  => 'ERR',
			    		ext_fields=> [qw/ blobfile /],
			    		from      => 'LO_pcr',
			    		msg_short => "Cannot insert the file into the database",
					);
				}
				else{
					$data_hash{'gel_id'} = $gel_id;
				}
            	$apiis->log('info',"Updating PCR ...");
	      		# updates one record in 'pcr_amplification'
	      		# ext_amplification, project_id, protocol_id, comment
                $pseudo_sql[$sql_no++] = (
                    'UPDATE pcr_amplification SET
                        ext_amplification = $pcr_amplification_id,
                        project_id        = $project_name_loop,
                        protocol_id       = $pcr_amplification_protocol_loop,
                        comment           = $pcr_comment
                     WHERE guid = $pa_guid
                    '
                );
				# update one record in 'pcr_markers'
				my $pa_id = Molab::Ajax::get_uniqueID(
					'pcr_amplification','amplification_id',
					'ext_amplification',$data_hash{'pcr_amplification_id'});
				$data_hash{'pa_id'} = $pa_id;
                $pseudo_sql[$sql_no++] = (
                    'UPDATE pcr_markers SET
                        marker_id       = $marker_loop
                     WHERE amplification_id = $pa_id
                    '
                );
                if ($data_hash{'va_guid'} ne ""){
					$pseudo_sql[$sql_no++] = (
						'UPDATE validation SET
							protocol_id = $pcr_validation_protocol_loop,
							gel_id		= $gel_id,
							comment     = $gel_comment
						WHERE guid = $va_guid
						'
					);
				}
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
                $apiis->log('info',"Deleting PCR ...");
				$data_hash{'amplification_id'} = Molab::Ajax::get_uniqueID(
					'pcr_amplification','amplification_id',
					'ext_amplification',$data_hash{'pcr_amplification_id'});
				              
                $pseudo_sql[$sql_no++] = (
                    'DELETE FROM amplified_samples
                     WHERE amplification_id = $amplification_id'
                );
		        $pseudo_sql[$sql_no++] = (
                    'DELETE FROM pcr_markers
                     WHERE amplification_id = $amplification_id'
                );
		        $pseudo_sql[$sql_no++] = (
                    'DELETE FROM pcr_amplification
                     WHERE amplification_id = $amplification_id'
                );
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }  # End EXIT
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_run_electrophoresis2 {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_run_electrophoresis2...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
            pcr_name_loop
            project_name_loop
            protocol_name_loop
            electrophoresis_method
            electrophoresis_purpose
        );
        # get a number of samples
        my $samples_number = 0;
    	if (defined $data_hash{'samples_number'}){
    		$samples_number = $data_hash{'samples_number'};
    		$apiis->log("notice", "A number of samples: ".$samples_number);
    	}
    	# convert ext_sample to dna_id and put to hash
        for(my $i=1; $i<= $samples_number; $i++){
        	if (defined $data_hash{"selected__$i"}){
				push(@LO_keys, 'dna_id__'.$i);
				$data_hash{'dna_id__'.$i} = Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$data_hash{"ext_sample__$i"});
				push(@LO_keys, 'pcrname__'.$i);
				$data_hash{'pcrname__'.$i} = $data_hash{"pcrname__$i"};
				$apiis->log('notice', 'pcrname__'.$i.'='.$data_hash{"pcrname__$i"});
        	}
        }
        # get a number of markers
#        my @marker_arr = $query->param("markers_loop");
#		my $markers_number = @marker_arr;
#		my $k = 0;
#		foreach (@marker_arr){
#			$k++;
#			push(@LO_keys, 'marker_id__'.$k);
#			$data_hash{'marker_id__'.$k} = $_;
#		}
#		$apiis->log("notice", "A number of markers: ".$markers_number);
		
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

#        my $today = $apiis->today;
#		$data_hash{opening_dt} = $today;
        EXIT:{
            my @pseudo_sql;
            my $sql_no = 0;
      		# inserts one record into the 'dna_extraction'
            $apiis->log('info', "Inserting electrophoresises...");
            # get new electrophoresis id
            my $electrophoresis_id = $apiis->DataBase->seq_next_val('seq_electrophoresises__electrophoresis_id');
            $data_hash{'electrophoresis_id'} = $electrophoresis_id;
            $apiis->log('info', "New electrophoresis ID:".$electrophoresis_id);
            $pseudo_sql[$sql_no++] = (
            	'INSERT INTO electrophoresises(
                  	electrophoresis_id,
                    ext_electrophoresis,
                    project_id,
                    protocol_id,
                    db_method,
                    purpose
                )
                VALUES (
                  	$electrophoresis_id,
                    $ext_electrophoresis,
                    $project_name_loop,
                    $protocol_name_loop,
                    $electrophoresis_method,
                    $electrophoresis_purpose
                )'
            );
            
            
            # 
            for (my $i=1; $i<=$samples_number; $i++){
              	if (defined $data_hash{"selected__$i"}){
              		# get a number of markers for that sample
              		$apiis->log('notice', 'Get pcrname__'.$i.'='.$data_hash{"pcrname__$i"});
              		my ($markers_array_ref, $markers_number) = Molab::Ajax::get_markers_to_array($data_hash{'pcrname__'.$i});
              		$apiis->log('notice', "Number of markers: " . $markers_number);
              		
		    		for (my $j=1; $j<=$markers_number; $j++){	
		    			$apiis->log('notice', "marker name: " . $markers_array_ref->[$j-1]);
		    			$data_hash{'marker_id__'.$j} = Molab::Ajax::get_uniqueID('markers','marker_id','marker_name',$markers_array_ref->[$j-1]);			
		    			$pseudo_sql[$sql_no++] = (
		            		'INSERT INTO analyzed_samples(
		                  		electrophoresis_id,
		                  		marker_id,
		                  		dna_id
		                	)
		                	VALUES (
		                  		$electrophoresis_id,
		                  		$marker_id__' . $j .',
		                  		$dna_id__' . $i .'
		               		)'
	           			);
              		}
              	}
            }
                    
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_load_electrophoresis2 {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_load_electrophoresis...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
			electrophoresis_id
			user_loop
			electrophoresis_dt
			software_loop
			comment
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        EXIT:{
            my @pseudo_sql;
	        # BEGIN - prepare SQL to INSERT data
	        if ($data_hash{'bt_submit'} eq "Update"){
	            # Uploading file and getting a new blobid
		    	if ($data_hash{'product_file'} ne "") {
					my $upload_dir = "/tmp";
					my $filename = $query->param("product_file");
					$apiis->log('notice', "Prepare uploading the file $filename");
					my $file_info= $query->uploadInfo($filename);
					my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
					$filename =~ s/.*[\/\\](.*)/$1/;
	                my $upload_filehandle = $query->upload("product_file");
					my $buffer='';
					unless (defined $type) {
			  			open UPLOADFILE, ">$upload_dir/$filename";
			  			binmode UPLOADFILE;
			  			while (<$upload_filehandle>){
			    			print UPLOADFILE;
			  			}
			  			close UPLOADFILE;
					} 
					else {
			  			my $data;
			  			$apiis->log('notice', "Type not defined");
			  			while (my $bytesread=read($upload_filehandle,$data,1024)) {
			    			$buffer.=$data;
			  			}
					}
					#if file content is empty raise error
					if ($buffer eq '') {
				  		$apiis->log('err', "Empty file");
				  		$hash_ref->{err_status}=1;
				  		push @errors,  
				  		Apiis::Errors->new(
							type      => 'DATA',
							severity  => 'ERR',
							ext_fields=> [qw/ blobfile /],
							from      => 'LO_load_electrophoresis2',
							msg_short => __("The file is empty, does not exist, or was not uploaded"),
						);
				  		last EXIT;
					}
					# load file into the BLOBS table
					$apiis->log('debug', 'mimetype: ' . $type);
					my $blobid;
					if (defined $type) {
				  		$blobid = load_blob($buffer,$type);
					} 
					else {
				  		$blobid = load_blob("$upload_dir/$filename");
					}
					if(defined $blobid){
				  		$apiis->log('info', "The product was stored to database");
				  		$apiis->log('debug', "blobid=". $blobid);
		                $data_hash{product_id} = $blobid;
					}
					else {
				  		$apiis->log('err', "Cannot load file to database");
				  		$hash_ref->{err_status}=1;
				  		push @errors, Apiis::Errors->new(
							    type      => 'DATA',
							    severity  => 'ERR',
							    ext_fields=> [qw/ blobfile /],
							    from      => 'LO_load_electrophoresis2',
							    msg_short => "Cannot insert the product file",
						);
				  		last EXIT;
					}
	    		} 
		    	else {
					$apiis->log('err', "Product file name missing on update");
					$hash_ref->{err_status}=1;
					push @errors,  Apiis::Errors->new(
							  type      => 'DATA',
							  severity  => 'ERR',
							  ext_fields=> [qw/ product_file /],
							  from      => 'LO_load_electrophoresis2',
							  msg_short => "Cannot insert without uploaded file",
					);
				  	last EXIT;
		   		}
                $pseudo_sql[0] = (
                    'UPDATE electrophoresises SET
                        db_user = $user_loop,
                        electrophoresis_dt = $electrophoresis_dt,
                        db_software = $software_loop,
                        product_id = $product_id,
                        comment = $comment
                     WHERE electrophoresis_id = $electrophoresis_id
                    '
                );
	        }
	        # END - prepare SQL to UPDATE data
            
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_electrophoresis {
	my %data_hash = %{ shift() };
	my $query = shift();
	my ( $err_status, $err_ref, @errors,$hash_ref );
	$apiis->log('info','Loading LO_electrophoresis...');
	ERR_LABEL:
	{
		##################################################################
		# %data_hash must have the following keys from @LO_keys and the
		# appropriate values from the datastream:
		##################################################################
		my @LO_keys = qw (
			guid
			project_name_loop
			ext_electrophoresis
			marker_loop
			product_file
			product_id
			protocol_loop
			software_loop
			electrophoresis_purpose_loop
			electrophoresis_method_loop
			user_loop
			electrophoresis_dt
			comment
		);
		# some basic checks:
		($err_status,$err_ref) = CheckLO(\%data_hash,\@LO_keys);
		return ($err_status,$err_ref) if $err_status;
		EXIT:{
			my @pseudo_sql;
			my $sql_no = 0;
			my $product_id = undef;
			if ($data_hash{'bt_submit'} eq "Insert"){
				# do nothing
			}
			elsif ($data_hash{'bt_submit'} eq "Update"){
				if ($data_hash{'product_file'} ne ''){
					$product_id = upload_file($query,'product_file');
					if ($product_id eq 0){
						$hash_ref->{err_status}=1;
						push @errors, Apiis::Errors->new(
							type      => 'DATA',
							severity  => 'ERR',
							ext_fields=> [qw/ blobfile /],
							from      => 'LO_electrophoresis',
							msg_short => "Cannot insert the file",
						);
					}
					else{
						$data_hash{'product_id'} = $product_id;
					}
				}
				$apiis->log('info',"Updating Electrophoresis ...");
				# updates one record in 'electrophoresises'
				if ((defined $product_id)and($product_id ne 0)){
					$pseudo_sql[$sql_no++] = (
						'UPDATE electrophoresises SET
							ext_electrophoresis = $ext_electrophoresis,
							db_user				= $user_loop,
							project_id			= $project_name_loop,
							db_software			= $software_loop,
							product_id			= $product_id,
							electrophoresis_dt	= $electrophoresis_dt,
							db_method			= $electrophoresis_method_loop,
							purpose				= $electrophoresis_purpose_loop,
							protocol_id			= $protocol_loop,
							comment				= $comment
						WHERE guid = $el_guid'
					);
				}
				else {
					$pseudo_sql[$sql_no++] = (
						'UPDATE electrophoresises SET
							ext_electrophoresis = $ext_electrophoresis,
							db_user				= $user_loop,
							project_id			= $project_name_loop,
							db_software			= $software_loop,
							electrophoresis_dt	= $electrophoresis_dt,
							db_method			= $electrophoresis_method_loop,
							purpose				= $electrophoresis_purpose_loop,
							protocol_id			= $protocol_loop,
							comment				= $comment
						WHERE guid = $el_guid'
					);
				}
				# update one record in analyzed_samples
				$pseudo_sql[$sql_no++] = (
					'UPDATE analyzed_samples SET
						marker_id = $marker_loop
					WHERE guid = $an_guid'
				);
			}
			elsif ($data_hash{'bt_submit'} eq "Delete"){
				$apiis->log('info',"Deleting Electrophoresis ...");
				$data_hash{'electrophoresis_id'} = Molab::Ajax::get_uniqueID(
					'electrophoresises','electrophoresis_id',
					'ext_electrophoresis',$data_hash{'ext_electrophoresis'});
				$pseudo_sql[$sql_no++] = (
					'DELETE FROM analyzed_samples
					WHERE electrophoresis_id = $electrophoresis_id'
				);
				$pseudo_sql[$sql_no++] = (
					'DELETE FROM electrophoresises
					WHERE electrophoresis_id = $electrophoresis_id'
				);
				if ($data_hash{'product_id'} ne ''){
					$pseudo_sql[$sql_no++] = (
						'DELETE FROM blobs
						WHERE blob_id = $product_id'
					);
				}
			}
			$hash_ref = meta_db(\@pseudo_sql,\%data_hash);
			if($hash_ref->{err_status}){
				$apiis->DataBase->dbh->rollback;
				push @errors, @{$hash_ref->{err_ref}};
			}
			else {
				$apiis->DataBase->dbh->commit;
			}
		}  # End EXIT
		return( $hash_ref->{err_status},\@errors);
	} # End ERR_LABEL
}

sub LO_analyze_genotyping2 {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    my $sample_id;
    my $storage_id;
    $apiis->log('info','Calling LO_analyze_genotyping2...');
    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
			project_id
			marker_id
        );
        my $samples_number = $data_hash{'samples_number'};
        $apiis->log('info',"Number of samples: $samples_number");
        for(my $i=1; $i<= $samples_number; $i++){
          if (defined $data_hash{'result__'.$i}){
        	for(my $j=1; $j<=$data_hash{'result__'.$i};$j++){	
        		$apiis->log('info',"Selected sample: $i");
				push(@LO_keys, 'dna_id__'.$i.'_'.$j);
				push(@LO_keys, 'allele1__'.$i.'_'.$j);
				push(@LO_keys, 'allele2__'.$i.'_'.$j);
				push(@LO_keys, 'comment__'.$i.'_'.$j);
				$data_hash{'dna_id__'.$i.'_'.$j} = 
				Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',
				$data_hash{"ext_dna__".$i."_".$j}); 
        	}
          }
		}
		
        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;
		# the order of inserting: Samples, Storage, Storage_Samples, Collection
        EXIT:{
          my @pseudo_sql;
          if ($data_hash{'bt_submit'} eq "Insert"){
          	my $sql_no = 0;
            $apiis->log('info', "Preparing SQL to insert data ...");
            for (my $i= 1; $i<=$samples_number; $i++){
           	  if (defined $data_hash{'result__'.$i}){
           	  	for(my $j=1; $j<=$data_hash{'result__'.$i};$j++){
	           	  my $microsatellite_id = 
	           	  $apiis->DataBase->seq_next_val('seq_microsatellites__microsatellite_id');
	           	  $data_hash{'microsatellite_id__'.$i.'_'.$j} = $microsatellite_id;
	           	  $apiis->log('notice', 'microsatellite_id:'.$microsatellite_id);
	           	  $pseudo_sql[$sql_no++]= '
	           		INSERT INTO microsatellites(
	           			microsatellite_id,
	           			dna_id,
	           			project_id,
	           			marker_id,
	           			allele1,
	           			allele2,
	           			comment
	           		)
	           		VALUES (
	           			$microsatellite_id__'. $i .'_'.$j. ',
	           			$dna_id__'. $i .'_'.$j. ',
	           			$project_id,
	           			$marker_id,
	           			$allele1__'. $i .'_'.$j. ',
	           			$allele2__'. $i .'_'.$j. ',
	           			$comment__'. $i .'_'.$j. '
	          		)
	          	  '
                }
           	  } 
            } # End for
          } 
          $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
          if($hash_ref->{err_status}){
              $apiis->DataBase->dbh->rollback;
              push @errors, @{ $hash_ref->{err_ref} };
          }
          else {
              $apiis->DataBase->dbh->commit;
          }
        } # End EXIT
        return( $hash_ref->{err_status},\@errors);
    } # End ERROR LABEL
}

sub LO_analyze_sequencing2 {
	my %data_hash = %{shift()};
	my $query = shift();
	my ($err_status,$err_ref,@errors,$hash_ref,$sample_id,$storage_id);
	$apiis->log('info','Calling LO_analyze_sequencing2...');
	ERR_LABEL: {
		##################################################################
		# %data_hash must have the following keys from @LO_keys and the
		# appropriate values from the datastream:
		##################################################################
		my @LO_keys = qw (
			project_id
			marker_id
		);
		my $samples_number = $data_hash{'samples_number'};
		$apiis->log('notice',"Number of samples: $samples_number");
		for(my $i=1; $i<= $samples_number; $i++){
			if (defined $data_hash{'result__'.$i}){
				for(my $j=1; $j<=$data_hash{'result__'.$i}; $j++){
					$apiis->log('notice',"Selected sample: $i");
					push(@LO_keys, 'dna_id__'.$i.'_'.$j);
					push(@LO_keys, 'comment__'.$i.'_'.$j);
					$data_hash{'dna_id__'.$i.'_'.$j} =
					Molab::Ajax::get_uniqueID('samples','sample_id',
					'ext_sample',$data_hash{"ext_dna__".$i.'_'.$j});
				}
			}
		}
		# some basic checks:
		($err_status, $err_ref ) = CheckLO(\%data_hash,\@LO_keys);
		return ( $err_status, $err_ref ) if $err_status;
		# inserting data to Sequences
		EXIT:{
			my @pseudo_sql;
			my $sq_data;
			if ($data_hash{'bt_submit'} eq "Insert"){
				my $sql_no = 0;
				$apiis->log('info', "Preparing SQL to insert data ...");
				for (my $i=1; $i<=$samples_number; $i++){
					if (defined $data_hash{'result__'.$i}){
						for(my $j=1; $j<=$data_hash{'result__'.$i};$j++){
							$apiis->log('info', "Uploading ".
								$data_hash{'input_file__'.$i.'_'.$j});
							$sq_data = Molab::Ajax::upload_sequence(
								$query,'input_file__'.$i.'_'.$j);
							$data_hash{'sq_data__'.$i.'_'.$j} = $sq_data;
							my $sq_id = next_id('sequences','sequence_id');
							$data_hash{'sq_id__'.$i.'_'.$j} = $sq_id;
							$apiis->log('notice', 'sequence_id:'.$sq_id);
							$pseudo_sql[$sql_no++]= '
								INSERT INTO sequences(
									sequence_id,
									dna_id,
									project_id,
									marker_id,
									sequence,
									comment
								)
								VALUES(
									$sq_id__'.$i.'_'.$j.',
									$dna_id__'.$i.'_'.$j.',
									$project_id,
									$marker_id,
									$sq_data__'.$i.'_'.$j.',
									$comment__'.$i.'_'.$j. '
								)'
						}
					} # End if
				} # End for
			}
			$hash_ref = meta_db(\@pseudo_sql,\%data_hash);
			if($hash_ref->{err_status}){
				$apiis->DataBase->dbh->rollback;
				push @errors, @{ $hash_ref->{err_ref} };
			}
			else { $apiis->DataBase->dbh->commit; }
		} # End EXIT 
		return( $hash_ref->{err_status},\@errors);
	} # End ERROR LABEL
}

sub LO_microsatellites {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_microsatellites...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
			guid
			allele1
			allele2
			comment      
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Update"){
	      		# update one record in the table 'microsatellites'  
                $apiis->log('info',"Updating microsatellites...");
                $pseudo_sql[0] = (
                    'UPDATE microsatellites SET
                        allele1 = $allele1,
                        allele2 = $allele2,   
                        comment = $comment
                     WHERE guid = $guid
                    '
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
	        	# delete one record in the table 'microsatellites'
                $apiis->log('info',"Deleting microsatellites...");
                $pseudo_sql[0] = (
                    'DELETE FROM microsatellites
                     WHERE guid = $guid
                    '
                );
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

sub LO_sequences {
    my %data_hash = %{ shift() };
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('info','Loading LO_sequences...');

    ERR_LABEL: {
        ##################################################################
        # %data_hash must have the following keys from @LO_keys and the
        # appropriate values from the datastream:
        ##################################################################
        my @LO_keys = qw (
			guid
			sequence
			comment      
        );

        # some basic checks:
        ( $err_status, $err_ref ) = CheckLO( \%data_hash, \@LO_keys );
        return ( $err_status, $err_ref ) if $err_status;

        EXIT:{
            my @pseudo_sql;
            if ($data_hash{'bt_submit'} eq "Update"){
	      		# update one record in the table 'microsatellites'  
                $apiis->log('info',"Updating sequences...");
                $pseudo_sql[0] = (
                    'UPDATE sequences SET
                        sequence = $sequence,
                        comment = $comment
                     WHERE guid = $guid
                    '
                );
            }
            elsif ($data_hash{'bt_submit'} eq "Delete"){
	        	# delete one record in the table 'microsatellites'
                $apiis->log('info',"Deleting sequences...");
                $pseudo_sql[0] = (
                    'DELETE FROM sequences
                     WHERE guid = $guid
                    '
                );
            }
            $hash_ref = meta_db(\@pseudo_sql,\%data_hash);
            if($hash_ref->{err_status}){
                $apiis->DataBase->dbh->rollback;
                push @errors, @{ $hash_ref->{err_ref} };
            }
            else {
                $apiis->DataBase->dbh->commit;
            }
        }
        return( $hash_ref->{err_status},\@errors);
    }
}

sub upload_file {
	my ($query,$field_name) = @_;
	my $upload_dir = "/tmp";
	my $filename = $query->param("$field_name");
	# sending the file to the server
	$apiis->log('notice',"Prepare uploading file $filename");
	my $file_info= $query->uploadInfo($filename);
	my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
	$filename =~ s/.*[\/\\](.*)/$1/;
	my $upload_filehandle = $query->upload($field_name);
	my $buffer='';
	open UPLOADFILE, ">$upload_dir/$filename";
	binmode UPLOADFILE;
	while (<$upload_filehandle>){ print UPLOADFILE; }
	close UPLOADFILE;
	$apiis->log('notice',$filename." was uploaded to ".$upload_dir);
	# storing the content of that file into the table 'blobs'
    $apiis->log('notice', "Loading $filename to the table BLOB");
	my $blobid = load_blob("$upload_dir/$filename");
	if(defined $blobid){
		$apiis->log('notice', "The gel image ($blobid) was stored to database");
		return $blobid;
	}
	else {
		$apiis->log('err', "Cannot load file to database");
		return 0;
	}
}

sub next_id{
	my ($table_name,$column_name) = @_;
	return $apiis->DataBase->seq_next_val("seq_".$table_name."__".$column_name);
}


1;

=head1 AUTHORS

Truong Van Chi Cong <cong.chi@fli.bund.de>

=cut
