package Molab::Main;

=head1 NAME

Molab::Main - Perl module for handling non-Ajax actions

=head1 DESCRIPTION

This module will check the form action (non-Ajax) to execute related tasks. 

=cut

=head1 SYNOPSIS

  use Main;
  my ($template,$data_session) =  Main::execute(\%$inputs,\%session,\$cgi);

=cut

##############################################################################
BEGIN {
    use Env qw( APIIS_HOME);
    die "\n\tAPIIS_HOME is not set!\n\n" unless $APIIS_HOME;
    push @INC, "$APIIS_HOME/lib";
}

use Apiis::Init;		# for activating Apiis components
use Data::Dumper;		# for debugging
use Labels;				# for translating labels in HTML templates
use Params;				# for generating data fields in web forms 
use LoadObjects;		# for processing database sessions
use Tooltip;			# for showing helps
use Spreadsheet::Read;	# for reading data from external files
use Spreadsheet::WriteExcel;
use open ':utf8';
use CGI qw/escapeHTML/;

push @INC, $apiis->project("molabis") . "/lib";
my $tmpl_dir    = "../../templates/";          	# path to HTML templates

################################################################################

=head2 execute

Generating standard web forms from HTML templates

=over 4

=item B<input>:

    1. String: Form name which is found in the HTML template
    2. Session ref: A reference to a data hash. This is a Session variable used
       to keep track of all data fields sent to the server before
    3. CGI ref: A reference to the CGI object

=item B<output>:

    1. HTML template: A HTML template 
    2. Session ref: A reference to the session (a hash)
    
=back

=cut

################################################################################
sub execute {
	my ($form_input,$data_session,$cgi) = @_;
	my ($content_tmpl,$tmpl_name,$layout);
    my %mysession;
    my $tmpl_dir = $apiis->project("molabis") . "/templates/";
    my $sid = $apiis->User->user_session_id;
    my $debug_path  = "/tmp/molab_debug_file";  # file for debugging
    my $action_status;
    my %form_desc;             					# hash for template data
    my %form_filter;           					# hash for filters on the form
    my %session_elems = (
    	'_session_id'=>1,'username'=>1,'password'=>1,'description_hash'=>1);

	# $debug=1 (debug), $debug=0 (non)
	my $debug = 0;
	my $html_data;
	if ($debug) { open DBG_FILE, ">>", $debug_path; }
	use POSIX qw(strftime);
	my ($begin_time, $end_time);
	use Time::HiRes::Value;
	$begin_time = Time::HiRes::Value->now();
	if ($debug) {
	    print DBG_FILE "\n"."--------------- start of request ---------------";
	    print DBG_FILE "\n".strftime "%b-%e-%Y %H:%M:%S", localtime;
	}
	print DBG_FILE "\n"."Debuging for Main::execute() ...." if $debug;
        my $layout = $form_input->{"layout"};
    # check if the user is working on the same module
    my $new_layout;
    if ($layout =~ /^collect_samples(.*)/ ){
    	$new_layout = 'collect_samples';
    } elsif ($layout =~ /^prepare_samples(.*)/ ){
    	$new_layout = 'prepare_samples';
    } elsif ($layout =~ /^extract_dna(.*)/ ) {
        $new_layout = 'extract_dna';
    } elsif ($layout =~ /^prepare_working_sol(.*)/ ) {
        $new_layout = 'prepare_working_sol';
    } elsif ($layout =~ /^[amplify_pcr](.*)/){
        $new_layout = 'amplify_pcr';
    } elsif ($layout =~ /^[check_pcr_products](.*)/){
        $new_layout = 'check_pcr_products';
    } elsif ($layout =~ /^[run_electrophoresis](.*)/){
        $new_layout = 'run_electrophoresis';
    } elsif ($layout =~ /^[load_electrophoresis](.*)/){
        $new_layout = 'load_electrophoresis';
    } elsif ($layout =~ /^[analyze_genotyping](.*)/){
        $new_layout = 'analyze_genotyping';
    } elsif ($layout =~ /^[analyze_sequencing](.*)/){
        $new_layout = 'analyze_sequencing';
    }
    else{
    	$new_layout = $layout;
    }
    my $old_layout = $$data_session{'layout'};
#    $apiis->log("notice",'old layout:'.$old_layout);
#    $apiis->log("notice",'new layout:'.$new_layout);
    if (not defined $$data_session{'layout'}){
    	$mysession{'layout'} = $new_layout;
    }
    else{
    	if ($old_layout eq $new_layout){
    		$mysession{'layout'} = $new_layout;
    		$apiis->log("notice",'<<<<< THE SAME MODULE >>>>>');
    	}
    	else{
    		$mysession{'layout'} = $new_layout;	
    		$apiis->log("notice",'<<<<< SELECT NEW MODULE >>>>>');
    		# clear the data session
    		foreach my $key (%$data_session){
    			if (not defined $session_elems{$key}){
    				delete $data_session->{$key};
    				$apiis->log("notice","delete $key");
    			}
    		}
    	}
    }
    my $status
      =  $apiis->Auth->check_system_tasks('form','access_rights_manager tool');
    my $layout = $form_input->{"layout"};
    if ($status) {
        # creating an empty template
        if (defined $layout and $layout ne "login" and $layout ne "logout"){
    		$tmpl_name = $layout. ".html";
    	} else {
			$tmpl_name = "homepage.html";
		}
        $content_tmpl = HTML::Template->new(
            filename          => $tmpl_dir . $tmpl_name,
            die_on_bad_params => 0
        );
        # getting keys for Unicode encoding
        $apiis->log('notice', "<FORM INPUTS>");
        foreach my $key (keys %$form_input) {
            $apiis->log('notice', "\t$key => ". %$form_input->{$key});
            print DBG_FILE "\n" . $key . "=>" . %$form_input->{$key} if $debug;
            if (defined %$form_input->{$key}){
                $form_desc{$key} = Encode::encode_utf8(%$form_input->{$key});
                if ($key =~ /^[F|f][i][l][t][e][r]_(.*)/){
                    $form_filter{$key} = %$form_input->{$key};
                }
            }
        }
        $apiis->log('notice', "</FORM INPUTS>");
        # CHECKING FORMS
        my @errors = { };
        my($err_status,$err_ref);
		my $sm = $form_input->{'bt_submit'};
		if ($layout ne 'login'){
			# e.g. uploading gel images
			if ($sm =~ /^[U|u][p][l][o][a][d](.*)/) {
				if ($form_input->{'layout'} eq "extract_dna4"){
					($err_status,$err_ref) = upload_gels(\%$form_input,$cgi);		
				}
				if ($form_input->{'layout'} eq "aligned_sequences"){
					$apiis->log('notice', "Uploading fasta file ...");
 					my $content_str = upload_fasta($cgi);
					$$data_session{'sequence_data'} = $content_str;
				}
			}
			if( ($sm =~ /^[U|u][p][d][a][t][e](.*)/) or
				($sm =~ /^[I|i][n][s][e][r][t](.*)/) or
				($sm =~ /^[D|d][e][l][e][t][e](.*)/)){
				$apiis->log('notice', "Prepare to call LO from Main.pm ...");
				# calling LO and getting errors if any
				{
					no strict "refs";
					my $function_name = "Molab::LoadObjects::LO_$layout";
					($err_status,$err_ref)= &$function_name(\%$form_input,$cgi);
				}
				# put all error messages to $mysession if any
				if (defined $err_status and $err_status) {
					foreach (@{$err_ref} ) {
						if ( defined $_->ext_fields ) {
			      		  push(@{$mysession{'errors_fields'} }, $_->ext_fields);
			    		} else {
			      		  push(@{$mysession{'errors_fields'} }, []);
			    		}                    
	                    push(@errors, $_->msg_short.'\n'.$_->msg_long.'\n\n');
	                    $apiis->log( 'notice', $_->msg_short.":".$_->msg_long);
		  			}
				}
			    # in case of successfull action put status message
	    		if (not defined $err_status or not $err_status) {
	        		$action_status = $sm . " action sucessfully executed";
	        		$apiis->log('notice', "Action in Main.pm $action_status");
	    		}
	    		else {
	    			$action_status = "";
	   	    		$apiis->log('notice', "Action status: Error"); 	
	    		}
        	}
			# reuse samples
        	elsif (($form_input->{'layout'} eq "reuse_samples") and
        	       ($sm =~ /^[S|s][u][b][m][i][t](.*)/)){
				$apiis->log( 'notice', "Prepare to call LO from Main.pm ...");
                # calling LO and getting errors if any
                {
                	no strict "refs";
                    my $function_name = "Molab::LoadObjects::LO_$layout";
                    ($err_status,$err_ref)= &$function_name(\%$form_input,$cgi);
                }
                # put all error messages to $mysession if any
                if (defined $err_status and $err_status) {
		  			foreach (@{$err_ref} ) {
	                   	if ( defined $_->ext_fields ) {
			      			push(@{$mysession{'errors_fields'}},$_->ext_fields);
			    		} else {
			      			push(@{$mysession{'errors_fields'} }, []);
			    		}                    
	                    push(@errors, $_->msg_short.' \n'.$_->msg_long.'\n\n');
	                    $apiis->log( 'notice', $_->msg_short .":".$_->msg_long);
		  			}
				}
			    # in case of successfull action put status message
	    		if (not defined $err_status or not $err_status) {
	        		$action_status = $sm . " action sucessfully executed";
	        		$apiis->log('notice', "Action in Main.pm $action_status");
	    		}
        	}        	
        	# collect samples
        	elsif ( ($form_input->{'layout'} eq "collect_samples1") or 
        			($form_input->{'layout'} eq "collect_samples2") or
        			($form_input->{'layout'} eq "collect_samples3") ) {
        		$apiis->log("notice","call Main::collect_samples ...");
				my $rs_ref;
				my $storage_places_ref;
				($storage_places_ref,$vessel_types_ref,
				$rs_ref,$mysession) = 
					collect_samples(\%$form_input,$data_session,$cgi);
				$content_tmpl->param(result_set => $rs_ref);
				$content_tmpl->param(storage_places => $storage_places_ref);
				$content_tmpl->param(vessel_types => $vessel_types_ref);
				$content_tmpl->param(loc_counter => $$mysession{'loc_counter'})
					if defined $$mysession{'loc_counter'};
        		$content_tmpl->param( vst_counter => $$mysession{'vst_counter'})
					if defined $$mysession{'vst_counter'};
        	}
        	# manage samples
        	elsif ($form_input->{'layout'} eq "samples"){
        		my @storage_places;
        		my $loc_counter;
        		($loc_counter,@storage_places) = get_storage_places();
        		$content_tmpl->param( storage_places => \@storage_places);
        		$content_tmpl->param( loc_counter => $loc_counter);
        	}
        	# prepare samples
        	elsif ($form_input->{'layout'} eq "prepare_samples2"){
        		$apiis->log("notice","call Main::prepare_samples ...");
				my ($rs_ref, $file_name)=
				 prepare_samples(\%$form_input,$data_session,$cgi);
				$content_tmpl->param( result_set=> $rs_ref);
				my $server_name = $cgi->server_name();
				my $save_file = "http://$server_name/cgi-bin/download_report.cgi?report_name=$file_name";
				$content_tmpl->param( save_file => $save_file);
        	}
			# extract DNA
        	elsif ( ($form_input->{'layout'} eq "extract_dna1") or 
        			($form_input->{'layout'} eq "extract_dna2") or
        			($form_input->{'layout'} eq "extract_dna3") or
        			($form_input->{'layout'} eq "extract_dna4")) {
        		$apiis->log("notice","call Main::extract_dna ...");
				my $valid_protocols_ref;
				my $gel_sets_ref;
				my $rs_ref;					# reference to a set of records
				my $storage_places_ref;
				my $rs_ref;
				($gel_sets_ref,$valid_protocols_ref,$storage_places_ref,
				$vessel_types_ref,$rs_ref,$mysession) = 
					extract_dna(\%$form_input,$data_session,$cgi);
				$content_tmpl->param( storage_places => $storage_places_ref)
					if ($form_input->{'layout'} eq "extract_dna4");
				$content_tmpl->param( vessel_types => $vessel_types_ref)
					if ($form_input->{'layout'} eq "extract_dna4");
				$content_tmpl->param(loc_counter => $$mysession{'loc_counter'})
					if ($form_input->{'layout'} eq "extract_dna4");
        		$content_tmpl->param(vst_counter => $$mysession{'vst_counter'})
					if ($form_input->{'layout'} eq "extract_dna4");
				$content_tmpl->param( gel_sets => $gel_sets_ref)
					if ($form_input->{'layout'} eq "extract_dna4");
					
				$content_tmpl->param(valid_protocols => $valid_protocols_ref)
					if ($form_input->{'layout'} eq "extract_dna3");
				$content_tmpl->param( protocol_counter => 
				    $$mysession{'protocol_counter'} )
					if ($form_input->{'layout'} eq "extract_dna3");
				$content_tmpl->param( result_set=> $rs_ref);
        	}
        	# manage dna
        	elsif ($form_input->{'layout'} eq "dna"){
        		my @storage_places;
        		my $loc_counter;
        		($loc_counter,@storage_places) = get_storage_places();
        		$content_tmpl->param( storage_places => \@storage_places);
        		$content_tmpl->param( loc_counter => $loc_counter);
        	}
        	# prepare working solutions
        	elsif ($form_input->{'layout'} eq "prepare_working_sol2"){
        		$apiis->log("notice","call Main::prepare_working_sols ...");
				my $rs_ref;
				$rs_ref = 
				  prepare_working_sols(\%$form_input,$data_session,$cgi);
				$content_tmpl->param( result_set=> $rs_ref);
        	}
			# amplify PCR
        	elsif ($form_input->{'layout'} eq "amplify_pcr2"){
				$apiis->log( 'notice', "Prepare to call LO from Main.pm ...");
                # calling LO and getting errors if any
                {
                	no strict "refs";
                    my $function_name = "Molab::LoadObjects::LO_$layout";
                    ($err_status,$err_ref)= &$function_name(\%$form_input,$cgi);
                }
                # put all error messages to $mysession if any
                if (defined $err_status and $err_status) {
		  			foreach (@{$err_ref} ) {
	                   	if ( defined $_->ext_fields ) {
			      			push(@{$mysession{'errors_fields'}},$_->ext_fields);
			    		} else {
			      			push(@{$mysession{'errors_fields'} }, []);
			    		}                    
	                    push(@errors, $_->msg_short.' \n'.$_->msg_long.'\n\n');
	                    $apiis->log( 'notice', $_->msg_short .":".$_->msg_long);
		  			}
				}
				
			    # in case of successfull action put status message
	    		if (not defined $err_status or not $err_status) {
	        		$action_status = $sm . " action sucessfully executed";
	        		$apiis->log('notice', "Action in Main.pm $action_status");
	        		$apiis->log("notice","call Main::amplify_pcr ...");
					my $rs_ref;
					$rs_ref = amplify_pcr(\%$form_input,$data_session,$cgi);
					$content_tmpl->param( result_set=> $rs_ref);
					my $project_name = Molab::Ajax::get_uniqueID(
					  'projects','project_name','project_id',
					  $form_input->{"project_name_loop"});
					my $protocol_name = Molab::Ajax::get_uniqueID(
					  'protocols','protocol_name','protocol_id',
					  $form_input->{"protocol_name_loop"});
					my @marker_arr = $cgi->param("markers_loop");
					my $markers = "";
					foreach (@marker_arr){
						$markers .= Molab::Ajax::get_uniqueID(
						'markers','marker_name','marker_id',$_) . ", ";
					}
					$content_tmpl->param(project_name => $project_name);
					$content_tmpl->param(protocol_name => $protocol_name);
					$content_tmpl->param(markers => $markers);
	    		}
	    		else {
	    			$action_status = "";
	   	    		$apiis->log('notice', "Action status: Error"); 	
	    		}
        	}
        	# check PCR products
        	elsif ($form_input->{'layout'} eq "check_pcr_products2"){
        		$apiis->log("notice","call Main::check_pcr_products ...");
				my $rs_ref;
				my $pcr_details_ref;
				my $samples_number;
				my $valid_protocols_ref;
				my $protocol_counter;
				($rs_ref,$pcr_details_ref,$samples_number,
				$valid_protocols_ref,$protocol_counter,$mysession) 
				  = check_pcr_products(\%$form_input,$data_session,$cgi);
				$content_tmpl->param( result_set=> $rs_ref);
				
				# load all values from hash to proper variables
				my $amplification_id  = $pcr_details_ref->{'amplification_id'};
				my $pcr_amplification = $pcr_details_ref->{'ext_amplification'};
				my $project_name      = $pcr_details_ref->{'project_name'};
				my $protocol_name     = $pcr_details_ref->{'protocol_name'};
				my $markers           = "markers";
				my $comment           = $pcr_details_ref->{'comment'};
				$content_tmpl->param(samples_number => $samples_number);
				$content_tmpl->param(project_name => $project_name);
				$content_tmpl->param(protocol_name => $protocol_name);
				$content_tmpl->param(pcr_amplification => $pcr_amplification);
				$content_tmpl->param(amplification_id => $amplification_id);
				$content_tmpl->param(markers => $markers);
				$content_tmpl->param(comment => $comment);
				$content_tmpl->param(valid_protocols => $valid_protocols_ref);
				$content_tmpl->param(protocol_counter => $protocol_counter);
        	}
			# run electrophoresis
        	elsif ($form_input->{'layout'} eq "run_electrophoresis2"){
				$apiis->log( 'notice', "Prepare to call LO from Main.pm ...");
                # calling LO and getting errors if any
                {
                	no strict "refs";
                    my $function_name = "Molab::LoadObjects::LO_$layout";
                    ($err_status,$err_ref)= &$function_name(\%$form_input,$cgi);
                }
                # put all error messages to $mysession if any
                if (defined $err_status and $err_status) {
		  			foreach (@{$err_ref} ) {
	                   	if ( defined $_->ext_fields ) {
			      			push(@{$mysession{'errors_fields'}},$_->ext_fields);
			    		} else {
			      			push(@{$mysession{'errors_fields'}},[]);
			    		}                    
	                    push(@errors, $_->msg_short.' \n'.$_->msg_long.'\n\n');
	                    $apiis->log( 'notice', $_->msg_short." : ".$_->msg_long);
		  			}
				}
			    # in case of successfull action put status message
	    		if (not defined $err_status or not $err_status) {
	        		$action_status = $sm . " action sucessfully executed";
	        		$apiis->log('notice', "Action status in Main.pm $action_status");
	        		$apiis->log("notice","call Main::run_electrophoresis ...");
					my $rs_ref;
					$rs_ref = run_electrophoresis(\%$form_input,$data_session,$cgi);
					$content_tmpl->param( result_set=> $rs_ref);
					my $project_name = Molab::Ajax::get_uniqueID('projects','project_name','project_id',$form_input->{"project_name_loop"});
					my $protocol_name = Molab::Ajax::get_uniqueID('protocols','protocol_name','protocol_id',$form_input->{"protocol_name_loop"});
					my @marker_arr = $cgi->param("markers_loop");
					my $markers = "";
					foreach (@marker_arr){
						$markers .= Molab::Ajax::get_uniqueID('markers','marker_name','marker_id',$_) . ", ";
					}
					$content_tmpl->param(project_name => $project_name);
					$content_tmpl->param(protocol_name => $protocol_name);
					$content_tmpl->param(markers => $markers);
	    		}
	    		else {
	    			$action_status = "";
	   	    		$apiis->log('notice', "Action status: Error"); 	
	    		}
        	}
        	# load electrophoresis
        	elsif ( ($form_input->{'layout'} eq "load_electrophoresis1") or 
        			($form_input->{'layout'} eq "load_electrophoresis2") ) {
        	    $apiis->log("notice","call Main::load_electrophoresis ...");
				my $rs_ref;
				my $storage_places_ref;
				my $project_name;
				my $electrophoresis_id;
				my $ext_electrophoresis;
				($mysession,$project_name,
				 $electrophoresis_id,$ext_electrophoresis) = 
					load_electrophoresis(\%$form_input,$data_session,$cgi);
				$content_tmpl->param(project_name => $project_name);
				$content_tmpl->param(electrophoresis_id => $electrophoresis_id);
				$content_tmpl->param(ext_electrophoresis => $ext_electrophoresis);
				$content_tmpl->param(result_set => $rs_ref);
   			}
        	# genotyping
        	elsif ($form_input->{'layout'} eq "analyze_genotyping2") {
        		$apiis->log("notice","call Main::analyze_genotyping ...");
				my $rs_ref;
				my $storage_places_ref;
				my $project_id;
				my $marker_id;
				($rs_ref,$mysession,$project_id,$marker_id) = 
					analyze_genotyping(\%$form_input,$data_session,$cgi);
				$content_tmpl->param( project_id => $project_id);
				$content_tmpl->param( marker_id  => $marker_id);
				$content_tmpl->param( result_set => $rs_ref);
        	}
        	# sequencing
        	elsif ( ($form_input->{'layout'} eq "analyze_sequencing1") or 
        			($form_input->{'layout'} eq "analyze_sequencing2") ) {
        		$apiis->log("notice","call Main::analyze_sequencing ...");
				my $rs_ref;
				my $storage_places_ref;
				my $project_id;
				my $marker_id;
				($rs_ref,$mysession,$project_id,$marker_id) = 
					analyze_sequencing(\%$form_input,$data_session,$cgi);
				$content_tmpl->param(project_id => $project_id);
				$content_tmpl->param(marker_id  => $marker_id);
				$content_tmpl->param(result_set => $rs_ref);
        	}
		}
		# RELOADING DATA
		$msg = "START: Reloading data for $layout";
		print DBG_FILE "\n$msg" if $debug;
		$apiis->log("notice", $msg);
		# store errors to the session if any
		if ( defined $mysession{'errors_fields'} ){
			foreach my $error_field ( @{$mysession{'errors_fields'}} ){
				if (defined @{$error_field}[0]){
					$apiis->log('err',"ERROR field: ".@{$error_field}[0]);
					$form_desc{'class_'.@{$error_field}[0]}="form_input_error";
					$mysession{'errors'} .= " ".@{$error_field}[0].":".pop(@errors);
				} else {
					$mysession{'errors'} .= " ".pop(@errors);
				}
			}
			$mysession{'errors_fields'} = undef;
		}
		else { $apiis->log('debug', "No error was found!"); }
		my $div;
		# read all labels
		%form_labels = %{ Molab::Labels::get_labels(uc($layout)) };
		# Start: reload for each elements in the template
		foreach ($content_tmpl->param) {
			# translate for labels
			if (/^[l|L]\_(.*)/) {
				if ($form_labels{$_}){
					$content_tmpl->param( $_ => $form_labels{$_} );
				}
			}
			# high-light error fields in the red color
			elsif (/^[c|C][l|L][a|A][s|S][s|S]\_(.*)/) {
				if ($layout eq 'storages'){
					if ( defined $form_desc{$_} ) {
						$content_tmpl->param($_ =>"form_input_storage_error");
					} else {
						$content_tmpl->param($_ => "form_input_storage");
					}
				}
				else {
					if ($_ eq "class_sequence"){
						if (defined $form_desc{$_}){
							$content_tmpl->param($_ =>"form_input_sequences_error");
						} else {
							$content_tmpl->param($_ => "form_input_sequences");
						}
					}
					else {
						if (defined $form_desc{$_}){
							$content_tmpl->param($_ => "form_input_error");
						}
						else {
							$content_tmpl->param($_ => "form_input");
						}
					}
				}
			}
			# load data for filters on the search area
			elsif (/^[f|F][i|I][l|L][t|T][e|E][r|R]\_(.*)/){
				if (/(.*)\_[l|L][o|O][o|O][p|P]$/){
					if(($layout eq 'dna')
						or($layout eq 'organisms') or($layout eq 'samples'))
					{
						$content_tmpl->param($_ => Molab::Params::make_filter(
									$layout,$_,%$form_input->{$_},'yes'));
					}
					elsif (
						(($layout eq 'microsatellites')&&($_ eq 'filter_4_loop'))
					  or (($layout eq 'contacts')&&($_ eq 'filter_3_loop'))
 					  or (($layout eq 'projects')&&($_ eq 'filter_1_loop'))
					  or (($layout eq 'sequences')&&($_ eq 'filter_4_loop'))
					  or (($layout eq 'electrophoresis')&&($_ eq 'filter_1_loop'))
					  or (($layout eq 'electrophoresis')&&($_ eq 'filter_2_loop'))
					  or (($layout eq 'electrophoresis')&&($_ eq 'filter_3_loop'))
					  or (($layout eq 'pcr')&&($_ eq 'filter_1_loop'))
					  or (($layout eq 'pcr')&&($_ eq 'filter_2_loop'))
					  or (($layout eq 'pcr')&&($_ eq 'filter_3_loop')) )
					{
						$content_tmpl->param($_ => Molab::Params::make_filter(
						$layout,$_,%$form_input->{$_},'yes'));
					}
					elsif (($layout eq 'codes') && ($_ eq 'filter_1_loop')){
						$content_tmpl->param( $_ => Molab::Params::make_filter(
							$layout,$_,%$form_input->{'classname'},'no'));
					}
					else {
						$content_tmpl->param( $_ => Molab::Params::make_filter(
							$layout,$_,%$form_input->{$_},'no'));
					}
				}
				else { # do nothing
				}
			}
			# load data for drop-down lists on the form
			elsif (/(.*)\_[l|L][o|O][o|O][p|P]$/) {
				my $sql = Molab::Params::get_list_sql($layout,$_);
				my $current_option = "";
				if (defined $form_desc{$_}){
					$current_option = $form_desc{$_};
				}
				elsif(defined $$data_session{$_}){
					$current_option = $$data_session{$_};
				}
				print DBG_FILE "\n\t".$_."= ".$current_option if $debug;
				$apiis->log("debug", "\tL: ".$_."=>".$current_option);
				if (($_ eq 'markers_loop') and ($layout eq 'amplify_pcr1')){
					$content_tmpl->param( $_ =>	Molab::Params::make_multilist(
						$sql,$current_option,0,0,'yes'));
				}
				elsif (($_ eq 'marker_loop') and ($layout eq 'ms_toolkit')){
					$content_tmpl->param( $_ =>	Molab::Params::make_multilist(
						$sql,$current_option,0,0,'no'));
				}
				elsif (($_ eq 'target_project_loop')
					and ($layout eq 'reuse_samples')){
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,'yes'));
				}
				elsif (($_ eq 'db_species_loop')
					and ($layout eq 'reuse_samples')){
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,'yes'));
				}
				elsif (($_ eq 'db_breed_loop')
					and ($layout eq 'reuse_samples')){
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,'yes'));
				}
				elsif ($_ eq 'db_species_loop'){
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,'no'));
				}
				elsif ($_ eq 'pcr_name_loop'){
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,'yes'));
				}
				elsif (($_ eq 'db_breed_loop')
					and ($layout eq 'insert_organisms')){
					if ($current_option eq ""){
	                    $current_option = "unknown_breed";
					}
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,''));
				}
				elsif (($_ eq 'db_breed_loop')
					and ($layout eq 'organisms')){
					if ($current_option eq ""){
	                    $current_option = "unknown_breed";
					}
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,''));
				}
				elsif ($_ eq 'db_sex_loop'){
					if ($current_option eq ""){
	                    $current_option = "u";
					}
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,''));
				}
				elsif ($_ eq 'db_sire_loop'){
					if ($current_option eq ""){
	                    $current_option = "unknown_sire";
					}
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,''));
				}
				elsif ($_ eq 'db_dam_loop'){
					if ($current_option eq ""){
	                    $current_option = "unknown_dam";
					}
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,''));
				}
				else{
					$content_tmpl->param( $_ =>	Molab::Params::make_list(
						$sql,$current_option,0,0,'no'));
				}
			}
			# reload labels of buttons
			elsif (/^[b][t]\_(.*)/){
				if ($_ eq "bt_submit"){
					if ($form_desc{$_} eq ""){
						$content_tmpl->param($_ => "No Action");
					} else {
						$content_tmpl->param($_ => $form_desc{$_});
					}
				}
			}
			# make helps
			elsif (/^[b|B][u|U][b|B][b|B][l|L][e|E]\_(.*)/) {
				$apiis->log('debug', "\tH: ".$_.", ".$1);
				$content_tmpl->param($_ => Molab::Tooltip::build(uc($layout),$1));
   				$div =  '<div id="'. $1 . '"  class="hideClass">'
					. __("No description").'</div>'."\n";
			}
			# reload data for input fields
			else {
				if ($layout ne 'storages'){
					if ( $form_desc{$_} ) {
					  $html_data = escapeHTML($form_desc{$_});
						$content_tmpl->param($_ => $html_data);
					}
					elsif(defined $$data_session{$_}) {
					  $html_data = escapeHTML($$data_session{$_});
						$content_tmpl->param( $_ => $html_data);
					}
 					$apiis->log("info", "\tT: ". $_ ."=>");
				}
				$apiis->log("info", "\tT: ".$_ ."=>".$html_data);
			}
		}
		# End: reload for each elements in the template
		# set hidden values (action and session) in the content template	
		$content_tmpl->param('action' => $action);
		$content_tmpl->param('sid' => $sid);
		$content_tmpl->param('DIV', $div );
		$msg = "END: Reloading data for $layout";
		$apiis->log("notice", $msg);
		print DBG_FILE "\n$msg" if $debug;
	}
	else {
    }
	# stop debugging
	if ($debug) {
		$end_time = Time::HiRes::Value->now();
		print DBG_FILE  "\nChecking errors from LO: ". $mysession{'errors'};
		print DBG_FILE "\n"."time interval: ".$end_time->sub($begin_time). " s";
		print DBG_FILE "\n" . strftime "%b-%e-%Y %H:%M:%S", localtime;
		print DBG_FILE "\n"."--------------- end of request ---------------\n";
	}
	return ($content_tmpl, \%mysession, $mysession, $action_status);
}

sub upload_file {
	my $cgi = shift;
	my $filename;
	my $file_info;
	my $upload_dir = "/tmp";
	if ($cgi->param("blobfile") ne "") {
		$filename = $cgi->param("blobfile");
		$file_info=$cgi->uploadInfo($filename);
		$filename =~ s/.*[\/\\](.*)/$1/;
		my $upload_filehandle = $cgi->upload("blobfile");
		my $buffer='';
		open UPLOADFILE, ">$upload_dir/$filename";
		binmode UPLOADFILE;
		while (<$upload_filehandle>){
			print UPLOADFILE;
		}
		close UPLOADFILE;
	}
	return 1;
}

sub upload_fasta {
	my $cgi = shift;
	my $filename;
	my $file_info;
	my $upload_dir = "/tmp";
	my $content_str ="";
	if ($cgi->param("fasta_file") ne "") {
		$filename = $cgi->param("fasta_file");
		$file_info=$cgi->uploadInfo($filename);
		$filename =~ s/.*[\/\\](.*)/$1/;
		my $upload_filehandle = $cgi->upload("fasta_file");
		open UPLOADFILE, ">$upload_dir/$filename";
		binmode UPLOADFILE;
		while (<$upload_filehandle>){
			print UPLOADFILE;
		}
		close UPLOADFILE;
		open FILE,  "<$upload_dir/$filename" or die $!;
		while (my $line = <FILE>) {

			$content_str .= $line;
		}
		close FILE;
	}
	return $content_str;
}

sub collect_samples{
	my ($form_input,$data_session_ref,$cgi) = @_; 
	# uploade the file to server if have
	my $uploaded;
	my %my_session;
    my $ref;       # the variable used to point to the input file
	if ($cgi->param("action") eq "upload_file"){
  		$apiis->log("notice", "start uploading file ...");
        $uploaded = upload_file($cgi);
    }
	# declare nessary variables    
    my $project_name_loop;
    my $protocol_name_loop;
    my $samples_number;
	my @storage_places = ( );
	my @record_sets = ( );
	my @vessel_types = ( );
	
	# read a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    }
    elsif (defined $$data_session_ref{"samples_number"})  {
    	$samples_number = $$data_session_ref{"samples_number"};
    }
    else {
    	$samples_number = 0;
    }

    # From the current form we need to collect the step number (variable 'step')
    # and what action (variable 'action') has been done. These values will help
    # us to know what procedures should be used to change the status and to 
    # reload the data in the next form.
    my $step   = $form_input->{'step'};
    my $action = $form_input->{'action'};
    my $layout = $form_input->{'layout'};
    # Send to log
    $apiis->log("notice", "How many samples? --> ".$samples_number); 
    $apiis->log("notice", "Which step?       --> ".$step);
    $apiis->log("notice", "What action?      --> ".$action);
    
    # if the user pushed on the button 'Next' at Step '1'   
	if (($step eq "1") and ($action eq "next")){
		$apiis->log('debug','[Form 1]  ---next--> [Form 2]');
		# store current data in form 1 
		$my_session{"project_name_loop"}     = $form_input->{'project_name_loop'};
		$my_session{"project_name"}          = $form_input->{'project_name_loop'};
    	$my_session{"protocol_name_loop"}    = $form_input->{'protocol_name_loop'};
    	$my_session{"protocol_name"}         = $form_input->{'protocol_name_loop'};
    	$my_session{"samples_number"}        = $form_input->{'samples_number'};
    	$my_session{"db_donor_type_loop"}    = $form_input->{'db_donor_type_loop'};
    	$my_session{"db_donor_type"}         = $form_input->{'db_donor_type_loop'};
    	$my_session{"db_material_type_loop"} = $form_input->{'db_material_type_loop'};
    	$my_session{"db_material_type"}      = $form_input->{'db_material_type_loop'};
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    $hash_ref{'ext_organism'} = "ext_organism__".$row;
		    $hash_ref{'ext_sample'}   = "ext_sample__".$row;
		    $hash_ref{'volume'}       = "volume__".$row;
		    $hash_ref{'db_user'}      = "db_user__".$row;
	        $hash_ref{'sampling_dt'}  = "sampling_dt__".$row;
			$hash_ref{'comment'}      = "comment__".$row;
			my $sql = Molab::Params::get_list_sql($layout,'db_user_loop');
			$hash_ref{'ext_organism_val'} = $$data_session_ref{"ext_organism__".$row}
				if (defined $$data_session_ref{"ext_organism__". $row});
		    $hash_ref{'ext_sample_val'}   = $$data_session_ref{"ext_sample__".$row}
		    	if (defined $$data_session_ref{"ext_sample__". $row});
		    $hash_ref{'volume_val'}      = $$data_session_ref{"volume__".$row}
		    	if (defined $$data_session_ref{"volume__". $row});
		    $hash_ref{'sampling_dt_val'} = $$data_session_ref{"sampling_dt__".$row}
		    	if (defined $$data_session_ref{"sampling_dt__".$row});
		    if (defined $$data_session_ref{"db_user__". $row}){	
		    	$hash_ref{'db_user_loop'}= Molab::Params::make_list($sql,$$data_session_ref{"db_user__".$row},1,0,'no')
		    }
		    else {
		    	$hash_ref{'db_user_loop'}= Molab::Params::make_list($sql,'',1,0,'no');
		    }
			$hash_ref{'comment_val'}     = $$data_session_ref{"comment__". $row}
				if (defined $$data_session_ref{"comment__". $row});		
			push(@record_sets, \%hash_ref);
		}
	}
	# if the user pushed on the button 'Upload' at Step '2'
	elsif ( ($step eq "2") and ($action eq "upload_file") ){
		$apiis->log('debug','[Form 2]  ---reload--> [Form 2]');
	    # picking up data fields from the uploaded data file
	    my $upload_dir = "/tmp";
	    my $filename = $cgi->param("blobfile");
	    $apiis->log("notice", "Reading from: " . "$upload_dir/$filename");
		$ref = ReadData ("$upload_dir/$filename", sep => ",");
		$apiis->log("notice", "This file contains ", $ref->[1]{maxcol}, " columns (fields) and ", $ref->[1]{maxrow}, " rows (lines).\n");
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    $hash_ref{'ext_organism'} = "ext_organism__".$row;
		    $hash_ref{'ext_sample'}   = "ext_sample__".$row;
		    $hash_ref{'volume'}       = "volume__".$row;
		    $hash_ref{'db_user'}      = "db_user__".$row;
	        $hash_ref{'sampling_dt'}  = "sampling_dt__".$row;
			$hash_ref{'comment'}      = "comment__".$row;
			my $sql = Molab::Params::get_list_sql($layout,'db_user_loop');    	    
			my $user = Molab::Params::make_list($sql,$ref->[1]{cell}[5][$row+1],1,0,'no');
			$hash_ref{'ext_organism_val'} = $ref->[1]{cell}[1][$row+1];     
		    $hash_ref{'ext_sample_val'}   = $ref->[1]{cell}[2][$row+1];
		    $hash_ref{'volume_val'}      = $ref->[1]{cell}[3][$row+1];
		    $hash_ref{'sampling_dt_val'} = $ref->[1]{cell}[4][$row+1];
		    $hash_ref{'db_user_loop'}    = $user;
			$hash_ref{'comment_val'}     = $ref->[1]{cell}[6][$row+1];	
			push(@record_sets, \%hash_ref);
		}
	}
	# the back button in the second form was pushed
	elsif ( ($step eq "2") and ($action eq "back") ){
		$apiis->log('notice','[Form 1]  <--back--- [Form 2]');
		# store data in the second form
		for(my $row=1; $row<=$samples_number; $row++){
			$my_session{"ext_organism__".$row} = $form_input->{"ext_organism__".$row};
    		$my_session{"ext_sample__".$row} = $form_input->{"ext_sample__".$row};
    		$my_session{"volume__".$row} = $form_input->{"volume__".$row};
    		$my_session{"sampling_dt__".$row} = $form_input->{"sampling_dt__".$row};
    		$my_session{"db_user__".$row} = $form_input->{"db_user__".$row};
    		$my_session{"comment__".$row} = $form_input->{"comment__".$row};
    	}
	}
	# the next button in the second form was pushed
	elsif ( ($step eq "2") and ($action eq "next") ){
		$apiis->log('notice','[Form 2]  ---next--> [Form 3]');
		# collect data in the second form and send to the session variable
		for(my $row=1; $row<=$samples_number; $row++){
			$my_session{"ext_organism__".$row} = $form_input->{"ext_organism__".$row};
			$apiis->log('notice', "ext_organism__".$row . "=" . $my_session{"ext_organism__".$row});
    		$my_session{"ext_sample__".$row} = $form_input->{"ext_sample__".$row};
    		$my_session{"volume__".$row} = $form_input->{"volume__".$row};
    		$my_session{"sampling_dt__".$row} = $form_input->{"sampling_dt__".$row};
    		$my_session{"db_user__".$row} = $form_input->{"db_user__".$row};
    		$my_session{"comment__".$row} = $form_input->{"comment__".$row};
    	}
    	($my_session{"loc_counter"},@storage_places) = get_storage_places();
    	# build a data hash for types of vessels
    	$vst_counter = 0;
    	my $sql = "select ext_code as value, short_name as name from codes where class='VESSEL_TYPE'";
    	my $id = 0;
    	my $sql_ref = $apiis->DataBase->user_sql($sql);
    	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
			$id++; $vst_counter++;
			$hash_ref->{'id'} = $id;
			push(@vessel_types, $hash_ref);
    	}
    	$my_session{'vst_counter'} = $vst_counter;
    	
    	# build a data hash for results
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'row_id'} = "row__".$row;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    
#		    $hash_ref{'ext_organism'}  = "ext_organism__". $row;
			$hash_ref{'ext_organism'} = "ext_organism__". $row; 

		    $hash_ref{'ext_sample'}   = "ext_sample__". $row;
		    $hash_ref{'sampling_dt'}  = "sampling_dt__". $row;
		    $hash_ref{'volume'}       = "volume__". $row;
		    $hash_ref{'db_user'}      = "db_user__". $row;
		    $hash_ref{'comment'}      = "comment__". $row;
		    $hash_ref{'sample_group'} = "sample_group__". $row;
		    
#		    $hash_ref{'ext_organism_val'} = $my_session{"ext_organism__". $row};
			$hash_ref{'ext_organism_val'} = $my_session{"ext_organism__". $row};
			
		    $hash_ref{'ext_sample_val'}  = $my_session{"ext_sample__". $row};
		    $hash_ref{'sampling_dt_val'} = $my_session{"sampling_dt__". $row};
		    $hash_ref{'volume_val'}      = $my_session{"volume__". $row};
		    $hash_ref{'db_user_val'}     = $my_session{"db_user__". $row};
		    $hash_ref{'comment_val'}     = $my_session{"comment__". $row};
		    
			$hash_ref{'group_no'} = "group_no__".$row;
			push(@record_sets, \%hash_ref);
		}
		
	}
	# the back button in the third form was pushed
	elsif ( ($step eq "3") and ($action eq "back") ){
		$apiis->log('notice','[Form 3]  <--back--- [Form 2]');
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    $hash_ref{'ext_organism'}  = "ext_organism__". $row;
		    $hash_ref{'ext_sample'}    = "ext_sample__". $row;
		    $hash_ref{'volume'}       = "volume__". $row;
		    $hash_ref{'db_user'}      = "db_user__". $row;
	        $hash_ref{'sampling_dt'}  = "sampling_dt__". $row;
			$hash_ref{'comment'}      = "comment__". $row;
			my $sql = Molab::Params::get_list_sql($layout,'db_user_loop');    	    
	    	my $user_ref = Molab::Params::make_list($sql,$$data_session_ref{"db_user__". $row},1,0,'no');
			$hash_ref{'ext_organism_val'} = $$data_session_ref{"ext_organism__". $row};
		    $hash_ref{'ext_sample_val'}   = $$data_session_ref{"ext_sample__". $row};
		    $hash_ref{'volume_val'}      = $$data_session_ref{"volume__". $row};
		    $hash_ref{'sampling_dt_val'} = $$data_session_ref{"sampling_dt__". $row};
		    $hash_ref{'db_user_loop'}    = $user_ref;    
			$hash_ref{'comment_val'}     = $$data_session_ref{"comment__". $row};	
			push(@record_sets, \%hash_ref);
		}	
	}
	else {
		$apiis->log('notice','Loading data for the first form');
	}
	return (\@storage_places,\@vessel_types,\@record_sets,\%my_session);
}

# build a data hash for the storage places
sub get_storage_places{
	# get all locations at level 1
	my $sql_level1 = "select storage_id as node_value, storage_name as node_name from storage where parent_id = 0 order by storage_id";
	my $sql_ref_level1 = $apiis->DataBase->user_sql($sql_level1);
	my $node_id_level1 = 0;
   	my $loc_counter = 0;
   	while ( my $hash_ref_level1 = $sql_ref_level1->handle->fetchrow_hashref) {
   		$loc_counter++;
   		$apiis->log('notice', 'loc_counter='.$loc_counter);
       	$node_id_level1++;
       	$hash_ref_level1->{'node_id'} = $node_id_level1;
		push(@storage_places, $hash_ref_level1);
		# get all locations at level 2
		my $sql_level2 = "select storage_id as node_value, storage_name as node_name from storage where parent_id = " . $hash_ref_level1->{'node_value'} . " order by storage_id";
		my $sql_ref_level2 = $apiis->DataBase->user_sql($sql_level2);
		my $node_id_level2 = 0;
		while ( my $hash_ref_level2 = $sql_ref_level2->handle->fetchrow_hashref) {
			$node_id_level2++;
			$hash_ref_level2->{'node_id'} = $node_id_level1."_".$node_id_level2;
			push(@storage_places, $hash_ref_level2);
			# get all locations at level 3
			my $sql_level3 = "select storage_id as node_value, storage_name as node_name from storage where parent_id = " . $hash_ref_level2->{'node_value'} . " order by storage_id";
			my $sql_ref_level3 = $apiis->DataBase->user_sql($sql_level3);
			my $node_id_level3 = 0;
			while ( my $hash_ref_level3 = $sql_ref_level3->handle->fetchrow_hashref) {
				$node_id_level3++;
				$hash_ref_level3->{'node_id'} = $node_id_level1."_".$node_id_level2."_".$node_id_level3;
				push(@storage_places, $hash_ref_level3);
				# get all locations at level 4
				my $sql_level4 = "select storage_id as node_value, storage_name as node_name from storage where parent_id = " . $hash_ref_level3->{'node_value'} . " order by storage_id";
				my $sql_ref_level4 = $apiis->DataBase->user_sql($sql_level4);
				my $node_id_level4 = 0;
				while ( my $hash_ref_level4 = $sql_ref_level4->handle->fetchrow_hashref) {
					$node_id_level4++;
					$hash_ref_level4->{'node_id'} = $node_id_level1."_".$node_id_level2."_".$node_id_level3."_".$node_id_level4;
					push(@storage_places, $hash_ref_level4);
				}
			}
		}
   	}
   	return ($loc_counter,@storage_places);
}

=head2 check_stockname - Check if a stock name is existing in a given level

This function checks if there is a duplicate of stock name in a given in the
stock. Return value is true if there are any storage which keep the same name
in checked level. Otherwise we obtain a false value.

=over 4

=item B<input>:

    1. Stock name
    2. Stock ID in checked level

=item B<output>:

    True or False, error_status, errors

=back

=cut

sub check_stockname {
  my ($stock_name, $parent_id) = @_;
  $apiis->log('debug','Query::check_stockname');
  my $result = 0;  # default is False 
  my $sql = "SELECT storage_name,storage_id FROM storage WHERE parent_id = '" . $parent_id ."'";
  my $sql_ref = $apiis->DataBase->user_sql($sql);
  if ($apiis->status or $sql_ref->status ) {
    my @errors;
    push @errors, $apiis->errors;
    push @errors, $sql_ref->errors;
    return $result, 1, \@errors;
  }

  while ( my $hash_ref = $sql_ref->handle->fetchrow_hashref) {
    if ($hash_ref->{'storage_name'} eq $stock_name) {
      $result = $hash_ref->{'storage_id'};
    }
  }
  return $result;
}

sub prepare_samples{
	my ($form_input,$data_session_ref,$cgi) = @_;
	my $file_name;
# 	$apiis->log("notice", "DUMP2:" . Dumper($form_input));
	# get a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    	$apiis->log("notice", "Get samples_number from Form: ".$samples_number);
    }
	# build a data hash for results
	for (my $row=1; $row<=$samples_number; $row++){
		if (defined $form_input->{"selected__$row"}){
			my %hash_ref;
			$hash_ref{'row_id'} = "row__".$row;
			$hash_ref{'no'} = $row;
			# set styles for rows in the result table
			if ($row % 2){
			   	$hash_ref{'row_style'} = "table_row_even";
			}
			else {
			  	$hash_ref{'row_style'} = "table_row_odd";
			}
			# get sample_id from ext_sample
			my $sample_id =  Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$form_input->{"ext_sample__$row"});
			# get storage_id level 5 from sample_id
			my $storage_id_l5 =  Molab::Ajax::get_uniqueID('storage_samples','storage_id','sample_id',$sample_id);
			# get storage_name level 5 from storage_id level 5
			my $storage_name_l5 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l5);
			# get storage_id level 4 (parent_id level 5) from storage_id level 5
			my $storage_id_l4 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l5);
			# get storage_name level 4 from storage_id level 4
			my $storage_name_l4 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l4);
			# get storage_id level 3 (parent_id level 4) from storage_id level 4
			my $storage_id_l3 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l4);
			# get storage_name level 3 from storage_id level 3
			my $storage_name_l3 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l3);
			# get storage_id level 2 (parent_id level 3) from storage_id level 3 
			my $storage_id_l2 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l3);
			# get storage_name level 2 from storage_id level 2
			my $storage_name_l2 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l2);
			# get storage_id level 1 (parent_id level 2) from storage_id level 2
			my $storage_id_l1 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l2);
			# get storage_name level 1 from storage_id level 1
			my $storage_name_l1 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l1);
			
	    	$hash_ref{'result_col1'} = $form_input->{"ext_sample__$row"};
	    	$hash_ref{'result_col2'} = $storage_name_l1;
	    	$hash_ref{'result_col3'} = $storage_name_l2;
	    	$hash_ref{'result_col4'} = $storage_name_l3;
	    	$hash_ref{'result_col5'} = $storage_name_l4;
	    	$hash_ref{'result_col6'} = $storage_name_l5;
	    	$hash_ref{'result_col7'} = "";
	    	$hash_ref{'result_col8'} = ""; 
			push(@record_sets, \%hash_ref);
		}
	}
	# make a data file
    my $output_dir  = "/tmp/";
	$file_name = create_filename("xls");
	# create a new Excel workbook
	my $workbook = Spreadsheet::WriteExcel->new("$output_dir/$file_name");
	my $format1 = $workbook->add_format();
    $format1->set_align('center');
	$format1->set_bold();
	my $format2 = $workbook->add_format();
    $format2->set_align('left');
	my $worksheet = $workbook->add_worksheet();
 	# write a header to data file
	$worksheet->write(0,0,"Sample ID",$format1);
	$worksheet->write(0,1,"DNA ID",$format1);
	$worksheet->write(0,2,"Volume",$format1);
	$worksheet->write(0,3,"Extraction Date",$format1);
	$worksheet->write(0,4,"Concentration",$format1);
	$worksheet->write(0,5,"DNA Purity",$format1);
	$worksheet->write(0,6,"Extractor",$format1);
	$worksheet->write(0,7,"Comments",$format1);
	
	my $row = 1;
	for (my $row=1; $row<=$samples_number; $row++){
		my $sample_id = Encode::encode_utf8($form_input->{"ext_sample__$row"});
		$worksheet->write($row,0,$sample_id);
		$worksheet->write($row,1,$sample_id."_DNA");
	}
	return (\@record_sets,$file_name);
}

sub create_filename{
	my $ext = shift;
	my($sec,$min,$hrs,$day,$month,$year) = (localtime)[0,1,2,3,4,5];
	my $filename = ($year+1900).($month+1).$day.$hrs.$min.$sec;
	$filename .= ".$ext";
	return $filename;
}

sub prepare_working_sols{
	my ($form_input,$data_session_ref,$cgi) = @_;
	# get a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    	$apiis->log("notice", "Get samples_number from Form: ".$samples_number);
    }
	# build a data hash for results
	for (my $row=1; $row<=$samples_number; $row++){
		if (defined $form_input->{"selected__$row"}){
			my %hash_ref;
			$hash_ref{'row_id'} = "row__".$row;
			$hash_ref{'no'} = $row;
			# set styles for rows in the result table
			if ($row % 2){
			   	$hash_ref{'row_style'} = "table_row_even";
			}
			else {
			  	$hash_ref{'row_style'} = "table_row_odd";
			}
			# get sample_id from ext_sample
			my $sample_id =  Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$form_input->{"ext_sample__$row"});
			# get storage_id level 5 from sample_id
			my $storage_id_l5 =  Molab::Ajax::get_uniqueID('storage_samples','storage_id','sample_id',$sample_id);
			# get storage_name level 5 from storage_id level 5
			my $storage_name_l5 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l5);
			# get storage_id level 4 (parent_id level 5) from storage_id level 5
			my $storage_id_l4 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l5);
			# get storage_name level 4 from storage_id level 4
			my $storage_name_l4 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l4);
			# get storage_id level 3 (parent_id level 4) from storage_id level 4
			my $storage_id_l3 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l4);
			# get storage_name level 3 from storage_id level 3
			my $storage_name_l3 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l3);
			# get storage_id level 2 (parent_id level 3) from storage_id level 3 
			my $storage_id_l2 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l3);
			# get storage_name level 2 from storage_id level 2
			my $storage_name_l2 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l2);
			# get storage_id level 1 (parent_id level 2) from storage_id level 2
			my $storage_id_l1 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l2);
			# get storage_name level 1 from storage_id level 1
			my $storage_name_l1 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l1);
			
	    	$hash_ref{'result_col1'}  = $form_input->{"ext_sample__$row"};
	    	$hash_ref{'result_col2'}  = $storage_name_l1;
	    	$hash_ref{'result_col3'}  = $storage_name_l2;
	    	$hash_ref{'result_col4'}  = $storage_name_l3;
	    	$hash_ref{'result_col5'}  = $storage_name_l4;
	    	$hash_ref{'result_col6'}  = $storage_name_l5;
	    	$hash_ref{'result_col7'}  = $form_input->{"ss_concentration__$row"};
	    	$hash_ref{'result_col8'}  = $form_input->{"ws_concentration__$row"}; 
	    	$hash_ref{'result_col9'}  = $form_input->{"ss_volume__$row"};
	    	$hash_ref{'result_col10'} = $form_input->{"ws_volume__$row"}; 
	    	$hash_ref{'result_col11'} = $form_input->{"water_volume__$row"};
			push(@record_sets, \%hash_ref);
		}
	}
	return (\@record_sets);
}

sub extract_dna{
	my ($form_input,$data_session_ref,$cgi) = @_; 
	# uploade the file to server if have
	my $uploaded;
	my %my_session;
    my $ref;       # the variable used to point to the input file
	if ($cgi->param("action") eq "upload_file"){
  		$apiis->log("notice", "start uploading file ...");
        $uploaded = upload_file($cgi);
    }
	# declare nessary variables    
    my $project_name_loop;
    my $protocol_name_loop;
    my $samples_number;
	my @storage_places = ( );
	my @record_sets = ( );
	my @vessel_types = ( );
	my @gels_set = ( );
	my @dna_valid_protocols = ( );
	
	# read a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    }
    elsif (defined $$data_session_ref{"samples_number"})  {
    	$samples_number = $$data_session_ref{"samples_number"};
    }
    else {
    	$samples_number = 0;
    }

    # From the current form we need to collect the step number (variable 'step')
    # and what action (variable 'action') has been done. These values will help
    # us to know what procedures should be used to change the status and to 
    # reload the data in the next form.
    my $step   = $form_input->{'step'};
    my $action = $form_input->{'action'};
    my $layout = $form_input->{'layout'};
    # Send to log
    $apiis->log("notice", "How many samples? --> ".$samples_number); 
    $apiis->log("notice", "Which step?       --> ".$step);
    $apiis->log("notice", "What action?      --> ".$action);
    
    # if the user clicks on the button 'Next' at Step '1'   
	if (($step eq "1") and ($action eq "next")){
		$apiis->log('notice','[Form 1]  ---next--> [Form 2]');
		# send data at Step 1 to Session
		$my_session{"project_name_loop"}     = $form_input->{'project_name_loop'};
		$my_session{"project_name"}          = $form_input->{'project_name_loop'};
    	$my_session{"protocol_name_loop"}    = $form_input->{'protocol_name_loop'};
    	$my_session{"protocol_name"}         = $form_input->{'protocol_name_loop'};
    	$my_session{"samples_number"}        = $form_input->{'samples_number'};
		# prepare form for Step 2
		for (my $row=1; $row<=$samples_number; $row++){
			my %hash_ref;
			$hash_ref{'row_id'} = "row__".$row;
			$hash_ref{'no'} = $row;
			# set styles for rows in the result table
			if ($row % 2){
			   	$hash_ref{'row_style'} = "table_row_even";
			}
			else {
			  	$hash_ref{'row_style'} = "table_row_odd";
			}
			# assign names to data fields 
	    	$hash_ref{'ext_sample'}            = "ext_sample__".$row;
	    	$hash_ref{'dna_ext_sample'}        = "dna_ext_sample__".$row;
	    	$hash_ref{'dna_volume'}            = "dna_volume__".$row;
	    	$hash_ref{'extraction_dt'}         = "extraction_dt__".$row;
	    	$hash_ref{'dna_concentration'}     = "dna_concentration__".$row;
	    	$hash_ref{'dna_purity'}            = "dna_purity__".$row;
	    	$hash_ref{'db_user'}               = "db_user__".$row;
	    	$hash_ref{'comment'}               = "comment__".$row;
			# reload values from Session if possible
			my $sql = Molab::Params::get_list_sql($layout,'db_user_loop');
		    $hash_ref{'ext_sample_val'}   = $$data_session_ref{"ext_sample__".$row}
		    	if (defined $$data_session_ref{"ext_sample__". $row});
		    $hash_ref{'dna_ext_sample_val'} = $$data_session_ref{"dna_ext_sample__".$row}
		    	if (defined $$data_session_ref{"dna_ext_sample__". $row});
		    $hash_ref{'dna_volume_val'} = $$data_session_ref{"dna_volume__".$row}
		    	if (defined $$data_session_ref{"dna_volume__". $row});
		    $hash_ref{'extraction_dt_val'} = $$data_session_ref{"extraction_dt__".$row}
		    	if (defined $$data_session_ref{"extraction_dt__".$row});
		    $hash_ref{'dna_concentration_val'} = $$data_session_ref{"dna_concentration__".$row}
		    	if (defined $$data_session_ref{"dna_concentration__". $row});
		    $hash_ref{'dna_purity_val'} = $$data_session_ref{"dna_purity__".$row}
		    	if (defined $$data_session_ref{"dna_purity__". $row});
		    if (defined $$data_session_ref{"db_user__". $row}){	
		    	$hash_ref{'db_user_loop'}= Molab::Params::make_list($sql,$$data_session_ref{"db_user__".$row},0,0,'no')
		    }
		    else {
		    	$hash_ref{'db_user_loop'}= Molab::Params::make_list($sql,'',1,0,'no');
		    }
			$hash_ref{'comment_val'}     = $$data_session_ref{"comment__". $row}
				if (defined $$data_session_ref{"comment__". $row});		
	    	push(@record_sets, \%hash_ref);
		}
	}
	# if the user clicks on the button 'Upload' at Step '2'
	elsif ( ($step eq "2") and ($action eq "upload_file") ){
		$apiis->log('notice','[Form 2]  ---reload--> [Form 2]');
	    # picking up data fields from the uploaded data file
	    my $upload_dir = "/tmp";
	    my $filename = $cgi->param("blobfile");
	    $apiis->log("notice", "Reading from: " . "$upload_dir/$filename");
		$ref = ReadData ("$upload_dir/$filename", sep => ",");
		$apiis->log("notice", "This file contains ", $ref->[1]{maxcol}, " columns (fields) and ", $ref->[1]{maxrow}, " rows (lines).\n");
		# rebuild the form at Step 2
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    # assign names to data fields 
	    	$hash_ref{'ext_sample'}            = "ext_sample__".$row;
	    	$hash_ref{'dna_ext_sample'}        = "dna_ext_sample__".$row;
	    	$hash_ref{'dna_volume'}            = "dna_volume__".$row;
	    	$hash_ref{'extraction_dt'}         = "extraction_dt__".$row;
	    	$hash_ref{'dna_concentration'}     = "dna_concentration__".$row;
	    	$hash_ref{'dna_purity'}            = "dna_purity__".$row;
	    	$hash_ref{'db_user'}               = "db_user__".$row;
	    	$hash_ref{'comment'}               = "comment__".$row;
	    	$hash_ref{'valid_dna'}             = "valid_dna__".$row;
			# load values from data file
			$hash_ref{'ext_sample_val'}        = $ref->[1]{cell}[1][$row+1];
			$hash_ref{'dna_ext_sample_val'}    = $ref->[1]{cell}[2][$row+1];
		    $hash_ref{'dna_volume_val'}        = $ref->[1]{cell}[3][$row+1];
		    $hash_ref{'extraction_dt_val'}     = $ref->[1]{cell}[4][$row+1];
		    $hash_ref{'dna_concentration_val'} = $ref->[1]{cell}[5][$row+1];
		    $hash_ref{'dna_purity_val'}        = $ref->[1]{cell}[6][$row+1];
			my $sql = Molab::Params::get_list_sql($layout,'db_user_loop');	    
			my $user = Molab::Params::make_list($sql,$ref->[1]{cell}[7][$row+1],1,0,'no');
		    $hash_ref{'db_user_loop'}    = $user;
			$hash_ref{'comment_val'}     = $ref->[1]{cell}[8][$row+1];	
			push(@record_sets, \%hash_ref);
		}
	}
 	# if the user clicks on the button 'Back' at Step '2'
	elsif ( ($step eq "2") and ($action eq "back") ){
		$apiis->log('notice','[Form 1]  <--back--- [Form 2]');
		# send data at Step 2 to Session
		for(my $row=1; $row<=$samples_number; $row++){
    		$my_session{"ext_sample__".$row}        = $form_input->{"ext_sample__".$row};
    		$my_session{"dna_ext_sample__".$row}    = $form_input->{"dna_ext_sample__".$row};
    		$my_session{"dna_volume__".$row}        = $form_input->{"dna_volume__".$row};
    		$my_session{"extraction_dt__".$row}     = $form_input->{"extraction_dt__".$row};
    		$my_session{"dna_concentration__".$row} = $form_input->{"dna_concentration__".$row};
    		$my_session{"dna_purity__".$row}        = $form_input->{"dna_purity__".$row};
    		$my_session{"db_user__".$row}           = $form_input->{"db_user__".$row};
    		$my_session{"comment__".$row}           = $form_input->{"comment__".$row};
    	}
	}
	# if the user clicks on the button 'Next' at Step '2'
	elsif ( ($step eq "2") and ($action eq "next") ){
		$apiis->log('notice','[Form 2]  ---next--> [Form 3]');
		# send data at Step 2 to Session
		for(my $row=1; $row<=$samples_number; $row++){
    		$my_session{"ext_sample__".$row}        = $form_input->{"ext_sample__".$row};
    		$my_session{"dna_ext_sample__".$row}    = $form_input->{"dna_ext_sample__".$row};
    		$my_session{"dna_volume__".$row}        = $form_input->{"dna_volume__".$row};
    		$my_session{"extraction_dt__".$row}     = $form_input->{"extraction_dt__".$row};
    		$my_session{"dna_concentration__".$row} = $form_input->{"dna_concentration__".$row};
    		$my_session{"dna_purity__".$row}        = $form_input->{"dna_purity__".$row};
    		$my_session{"db_user__".$row}           = $form_input->{"db_user__".$row};
    		$my_session{"comment__".$row}           = $form_input->{"comment__".$row};
    	}
    	# assign the whole set of storage places
    	($my_session{"loc_counter"},@storage_places) = get_storage_places();
    	# assign a set of vessel types
    	$vst_counter = 0;
    	my $sql = "select ext_code as value, short_name as name from codes where class='VESSEL_TYPE'";
    	my $id = 0;
    	my $sql_ref = $apiis->DataBase->user_sql($sql);
    	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
			$id++; $vst_counter++;
			$hash_ref->{'id'} = $id;
			push(@vessel_types, $hash_ref);
    	}
    	$my_session{'vst_counter'} = $vst_counter;
    	# assign a set of dna validation protocols
    	my $protocol_counter = 0;
    	my $sql = "select protocol_id as value, protocol_name as name from protocols pt ";
		$sql   .= "LEFT JOIN codes cd on cd.db_code=pt.db_protocol_type ";
		$sql   .= "WHERE cd.ext_code='DNA validation'";
    	my $id = 0;
    	my $sql_ref = $apiis->DataBase->user_sql($sql);
    	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
			$id++; $protocol_counter++;
			$hash_ref->{'id'} = $id;
			push(@dna_valid_protocols, $hash_ref);
    	}
    	$my_session{'protocol_counter'} = $protocol_counter;
    	# prepare the form at Step 3
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'row_id'} = "row__".$row;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    # assign names to data fields 
	    	$hash_ref{'ext_sample'}         = "ext_sample__".$row;
	    	$hash_ref{'dna_ext_sample'}     = "dna_ext_sample__".$row;
	    	$hash_ref{'lane'}               = "lane__".$row;
	    	$hash_ref{'valid'}              = "valid__".$row;
	    	$hash_ref{'extraction_dt'}      = "extraction_dt__".$row;
	    	$hash_ref{'dna_concentration'}  = "dna_concentration__".$row;
	    	$hash_ref{'dna_purity'}         = "dna_purity__".$row;
	    	$hash_ref{'db_user'}            = "db_user__".$row;
	    	$hash_ref{'comment'}            = "comment__".$row;
		    $hash_ref{'sample_group'}       = "sample_group__". $row;
			$hash_ref{'gel_no'}             = "gel_no__".$row;
		    $hash_ref{'checking'}          	= "checking__". $row;
			$hash_ref{'dna_ext_sample_val'} = $my_session{"dna_ext_sample__".$row}
		    	if (defined $my_session{"dna_ext_sample__". $row});
			$hash_ref{'ext_sample_val'}        = $my_session{"ext_sample__". $row}
				if (defined $$data_session_ref{"ext_sample__". $row});
		    $hash_ref{'extraction_dt_val'}     = $my_session{"extraction_dt__". $row}
		    	if (defined $my_session{"extraction_dt__". $row});
		    $hash_ref{'lane_val'}              = $$data_session_ref{"lane__". $row}
		    	if (defined $$data_session_ref{"lane__". $row});
		    $hash_ref{'valid_val'}             = $$data_session_ref{"valid__". $row}
		    	if (defined $$data_session_ref{"valid__". $row});
		    $hash_ref{'dna_concentration_val'} = $my_session{"dna_concentration__". $row}
		    	if (defined $my_session{"dna_concentration__". $row});
		    $hash_ref{'dna_purity_val'}        = $my_session{"dna_purity__". $row}
		    	if (defined $my_session{"dna_purity__". $row});
		    $hash_ref{'db_user_val'}           = $my_session{"db_user__". $row}
		    	if (defined $my_session{"db_user__". $row});
		    $hash_ref{'comment_val'}           = $my_session{"comment__". $row}
		    	if (defined $my_session{"comment__". $row});
			push(@record_sets, \%hash_ref);
		}
	}
	# if the user clicks on the button 'Back' at Step '3'
	elsif ( ($step eq "3") and ($action eq "back") ){
		$apiis->log('notice','[Form 2]  <--back--- [Form 3]');
		# send data at Step 3 to Session
		for(my $row=1; $row<=$samples_number; $row++){
    		$my_session{"dna_ext_sample__".$row}    = $form_input->{"dna_ext_sample__".$row};
    		$my_session{"lane__".$row}              = $form_input->{"lane__".$row};
    		$my_session{"valid__".$row}             = $form_input->{"valid__".$row};
    	}
		# prepare form for Step 2
		for (my $row=1; $row<=$samples_number; $row++){
			my %hash_ref;
			$hash_ref{'row_id'} = "row__".$row;
			$hash_ref{'no'} = $row;
			# set styles for rows in the result table
			if ($row % 2){
			   	$hash_ref{'row_style'} = "table_row_even";
			}
			else {
			  	$hash_ref{'row_style'} = "table_row_odd";
			}
			# assign names to data fields 
	    	$hash_ref{'ext_sample'}           = "ext_sample__".$row;
	    	$hash_ref{'dna_ext_sample'}       = "dna_ext_sample__".$row;
	    	$hash_ref{'dna_volume'}            = "dna_volume__".$row;
	    	$hash_ref{'extraction_dt'}         = "extraction_dt__".$row;
	    	$hash_ref{'dna_concentration'}     = "dna_concentration__".$row;
	    	$hash_ref{'dna_purity'}            = "dna_purity__".$row;
	    	$hash_ref{'db_user'}               = "db_user__".$row;
	    	$hash_ref{'comment'}               = "comment__".$row;
	    	$hash_ref{'valid_dna'}             = "valid_dna__".$row;
			# reload values from Session if possible
			my $sql = Molab::Params::get_list_sql($layout,'db_user_loop');
		    $hash_ref{'ext_sample_val'}   = $$data_session_ref{"ext_sample__".$row}
		    	if (defined $$data_session_ref{"ext_sample__". $row});
		    $hash_ref{'dna_ext_sample_val'} = $$data_session_ref{"dna_ext_sample__".$row}
		    	if (defined $$data_session_ref{"dna_ext_sample__". $row});
		    $hash_ref{'dna_volume_val'} = $$data_session_ref{"dna_volume__".$row}
		    	if (defined $$data_session_ref{"dna_volume__". $row});
		    $hash_ref{'extraction_dt_val'} = $$data_session_ref{"extraction_dt__".$row}
		    	if (defined $$data_session_ref{"extraction_dt__".$row});
		    $hash_ref{'dna_concentration_val'} = $$data_session_ref{"dna_concentration__".$row}
		    	if (defined $$data_session_ref{"dna_concentration__". $row});
		    $hash_ref{'dna_purity_val'} = $$data_session_ref{"dna_purity__".$row}
		    	if (defined $$data_session_ref{"dna_purity__". $row});
		    if (defined $$data_session_ref{"db_user__". $row}){	
		    	$hash_ref{'db_user_loop'}= Molab::Params::make_list($sql,$$data_session_ref{"db_user__".$row},1,0,'no')
		    }
		    else {
		    	$hash_ref{'db_user_loop'}= Molab::Params::make_list($sql,'',1,0,'no');
		    }
			$hash_ref{'comment_val'}     = $$data_session_ref{"comment__". $row}
				if (defined $$data_session_ref{"comment__". $row});		
	    	push(@record_sets, \%hash_ref);
		}	
	}
	# if the user clicks on the button 'Next' at Step '3'
	elsif ( ($step eq "3") and ($action eq "next") ){
		$apiis->log('notice','[Form 3]  ---next--> [Form 4]');
		# send data at Step 3 to Session
		for(my $row=1; $row<=$samples_number; $row++){
    		$my_session{"dna_ext_sample__".$row} = $form_input->{"dna_ext_sample__".$row};
    		$my_session{"lane__".$row}           = $form_input->{"lane__".$row};
    		$my_session{"valid__".$row}          = $form_input->{"valid__".$row};
    		$my_session{"gel_no__".$row}         = $form_input->{"gel_no__".$row};
			$my_session{"checking__".$row}       = $form_input->{"checking__".$row};
    	}
    	$my_session{"gels_number"} = $form_input->{"gels_number"};
		$apiis->log('notice', "Number of gels: ".$my_session{"gels_number"});
		for(my $row=1; $row<=$my_session{"gels_number"} ; $row++){
    		$my_session{"gel_file__".$row}  = $form_input->{"gel_file__".$row};
    		$my_session{"dna_checking_protocol__".$row} = $form_input->{"checking_protocol__".$row};
    		$my_session{"dna_checking_comment__".$row}  = $form_input->{"checking_comment__".$row};
			$apiis->log('notice', "dna_checking_comment__".$row .$my_session{"dna_checking_comment__".$row});
    	}

    	# prepare form for Step 4
    	($my_session{"loc_counter"},@storage_places) = get_storage_places();
    	# build a data hash for types of vessels
    	$vst_counter = 0;
    	my $sql = "select ext_code as value, short_name as name from codes where class='VESSEL_TYPE'";
    	my $id = 0;
    	my $sql_ref = $apiis->DataBase->user_sql($sql);
    	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
			$id++; $vst_counter++;
			$hash_ref->{'id'} = $id;
			push(@vessel_types, $hash_ref);
    	}
    	$my_session{'vst_counter'} = $vst_counter;
    	# build a data hash for results
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'row_id'} = "row__".$row;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    $hash_ref{'ext_sample'}        = "ext_sample__".$row;
	    	$hash_ref{'dna_ext_sample'}    = "dna_ext_sample__".$row;
		    $hash_ref{'sampling_dt'}       = "sampling_dt__". $row;
		    $hash_ref{'dna_volume'}        = "dna_volume__". $row;
		    $hash_ref{'dna_concentration'} = "dna_concentration__". $row;
		    $hash_ref{'dna_purity'}        = "dna_purity__". $row;
		    $hash_ref{'extraction_dt'}     = "extraction_dt__". $row;
		    $hash_ref{'valid'}             = "valid__". $row;
		    $hash_ref{'is_valid'}          = "is_valid__". $row;
		    $hash_ref{'lane'}              = "lane__". $row;
		    $hash_ref{'db_user'}           = "db_user__". $row;
		    $hash_ref{'comment'}           = "comment__". $row;
		    $hash_ref{'sample_group'}      = "sample_group__". $row;
		    $hash_ref{'gel_no'}            = "gel_no__". $row;
		    $hash_ref{'ext_sample_val'}     = $$data_session_ref{"ext_sample__".$row}
		    	if (defined $$data_session_ref{"ext_sample__". $row});
		    $hash_ref{'dna_ext_sample_val'} = $$data_session_ref{"dna_ext_sample__".$row}
		    	if (defined $$data_session_ref{"dna_ext_sample__". $row});
            if ($my_session{"checking__".$row} eq "yes"){
				if (defined $my_session{"valid__". $row}){
					$hash_ref{'status_val'}   = "Valid";
					$hash_ref{'is_valid_val'} = "Y";
					$hash_ref{'status_class'} = "extract_dna__valid_status";
				}
				else{
					$hash_ref{'status_val'}   = "Invalid";
					$hash_ref{'is_valid_val'} = "N";
					$hash_ref{'status_class'} = "extract_dna__invalid_status";
				}
			}
			elsif ($my_session{"checking__".$row} eq "no"){
				$hash_ref{'status_val'}   = "Unknown";
				$hash_ref{'is_valid_val'} = "U";
			}
		    
		    $hash_ref{'extraction_dt_val'}     = $$data_session_ref{"extraction_dt__".$row}
		    	if (defined $$data_session_ref{"extraction_dt__". $row});
		    $hash_ref{'dna_volume_val'}        = $$data_session_ref{"dna_volume__".$row}
		    	if (defined $$data_session_ref{"dna_volume__". $row});
		    $hash_ref{'dna_concentration_val'} = $$data_session_ref{"dna_concentration__".$row}
		    	if (defined $$data_session_ref{"dna_concentration__". $row});
		    $hash_ref{'dna_purity_val'}        = $$data_session_ref{"dna_purity__".$row}
		    	if (defined $$data_session_ref{"dna_purity__". $row});
		    $hash_ref{'lane_val'}              = $my_session{"lane__".$row}
		    	if (defined $my_session{"lane__". $row});
		    $hash_ref{'db_user_val'}           = $$data_session_ref{"db_user__".$row}
		    	if (defined $$data_session_ref{"db_user__". $row});
		    $hash_ref{'comment_val'}           = $$data_session_ref{"comment__".$row}
		    	if (defined $$data_session_ref{"comment__". $row});
		    $hash_ref{'gel_no_val'}            = $my_session{"gel_no__".$row}
		    	if (defined $my_session{"gel_no__". $row});
			$hash_ref{'group_no'}              = "group_no__".$row;
			push(@record_sets, \%hash_ref);
		}
    	# build a data hash for gels
		for (my $row=1; $row<=$my_session{"gels_number"}; $row++){
		   	my %hash_ref;
		    $hash_ref{'gel_file'}              = "gel_file__".$row;
		    $hash_ref{'dna_checking_protocol'} = "dna_checking_protocol__".$row;
		    $hash_ref{'dna_checking_comment'}  = "dna_checking_comment__".$row;
		    $hash_ref{'gel_file_val'}          = $my_session{"gel_file__".$row}
		    	if (defined $my_session{"gel_file__". $row});
		    $hash_ref{'dna_checking_protocol_val'}= $my_session{"dna_checking_protocol__".$row}
		    	if (defined $my_session{"dna_checking_protocol__". $row});
		    $hash_ref{'dna_checking_comment_val'} = $my_session{"dna_checking_comment__".$row}
		    	if (defined $my_session{"dna_checking_comment__". $row});
		    push(@gel_sets, \%hash_ref);
		}
	}
	# if the user clicks on the button 'Next' at Step '3'
	elsif ( ($step eq "4") and ($action eq "back") ){
		$apiis->log('notice','[Form 3]  <---back-- [Form 4]');
		# send data at Step 4 to Session
    	# assign the whole set of storage places
    	($my_session{"loc_counter"},@storage_places) = get_storage_places();
    	# assign a set of vessel types
    	$vst_counter = 0;
    	my $sql = "select ext_code as value, short_name as name from codes where class='VESSEL_TYPE'";
    	my $id = 0;
    	my $sql_ref = $apiis->DataBase->user_sql($sql);
    	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
			$id++; $vst_counter++;
			$hash_ref->{'id'} = $id;
			push(@vessel_types, $hash_ref);
    	}
    	$my_session{'vst_counter'} = $vst_counter;
    	# assign a set of dna validation protocols
    	my $protocol_counter = 0;
    	my $sql = "select protocol_id as value, protocol_name as name from protocols pt ";
		$sql   .= "LEFT JOIN codes cd on cd.db_code=pt.db_protocol_type ";
		$sql   .= "WHERE cd.ext_code='DNA validation'";
		my $id = 0;
    	my $sql_ref = $apiis->DataBase->user_sql($sql);
    	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
			$id++; $protocol_counter++;
			$hash_ref->{'id'} = $id;
			push(@dna_valid_protocols, $hash_ref);
    	}
    	$my_session{'protocol_counter'} = $protocol_counter;
    	# prepare the form at Step 3
		for (my $row=1; $row<=$samples_number; $row++){
		   	my %hash_ref;
		   	$hash_ref{'row_id'} = "row__".$row;
		   	$hash_ref{'no'} = $row;
		    # set styles for rows in the result table
		    if ($row % 2){
		       	$hash_ref{'row_style'} = "table_row_even";
		    }
		    else {
		      	$hash_ref{'row_style'} = "table_row_odd";
		    }
		    # assign names to data fields 
	    	$hash_ref{'ext_sample'}            = "ext_sample__".$row;
	    	$hash_ref{'dna_ext_sample'}        = "dna_ext_sample__".$row;
	    	$hash_ref{'lane'}                  = "lane__".$row;
	    	$hash_ref{'valid'}                 = "valid__".$row;
	    	$hash_ref{'extraction_dt'}         = "extraction_dt__".$row;
	    	$hash_ref{'dna_concentration'}     = "dna_concentration__".$row;
	    	$hash_ref{'dna_purity'}            = "dna_purity__".$row;
	    	$hash_ref{'db_user'}               = "db_user__".$row;
	    	$hash_ref{'comment'}               = "comment__".$row;
		    $hash_ref{'sample_group'}          = "sample_group__". $row;	    
			$hash_ref{'gel_no'}                = "gel_no__".$row;
			
			$hash_ref{'dna_ext_sample_val'}    = $$data_session_ref{"dna_ext_sample__".$row}
		    	if (defined $$data_session_ref{"dna_ext_sample__". $row});
			$hash_ref{'ext_sample_val'}        = $$data_session_ref{"ext_sample__". $row}
				if (defined $$data_session_ref{"ext_sample__". $row});
		    $hash_ref{'extraction_dt_val'}     = $$data_session_ref{"extraction_dt__". $row}
		    	if (defined $$data_session_ref{"extraction_dt__". $row});
		    $hash_ref{'lane_val'}              = $$data_session_ref{"lane__". $row}
		    	if (defined $$data_session_ref{"lane__". $row});
		    $hash_ref{'valid_val'}             = $$data_session_ref{"valid__". $row}
		    	if (defined $$data_session_ref{"valid__". $row});
		    $hash_ref{'dna_concentration_val'} = $$data_session_ref{"dna_concentration__". $row}
		    	if (defined $$data_session_ref{"dna_concentration__". $row});
		    $hash_ref{'dna_purity_val'}        = $$data_session_ref{"dna_purity__". $row}
		    	if (defined $$data_session_ref{"dna_purity__". $row});
		    $hash_ref{'db_user_val'}           = $$data_session_ref{"db_user__". $row}
		    	if (defined $$data_session_ref{"db_user__". $row});
		    $hash_ref{'comment_val'}           = $$data_session_ref{"comment__". $row}
		    	if (defined $$data_session_ref{"comment__". $row});
			push(@record_sets, \%hash_ref);
		}
	}
	else {
		$apiis->log('notice','Loading data for the first form');
	}
	return (\@gel_sets,\@dna_valid_protocols,\@storage_places,\@vessel_types,\@record_sets,\%my_session);
}

sub upload_gels {
    my %data_hash = %{ shift() };
    my $query = shift();
    my ( $err_status, $err_ref, @errors,$hash_ref );
    $apiis->log('notice','Uploading gels...');
    my $gels_number = $data_hash{'gels_number'};
    EXIT:
    {
        my @pseudo_sql;
        if ($data_hash{'bt_submit'} eq "Upload"){
            $apiis->log('info', "Uploading...");
            my $sql_no = 0;
            
	    	# upload gel files to server stored in /tmp
            for (my $i= 0; $i < $gels_number; $i++){
		      	if ( $data_hash{'gel_file__'.($i+1)} ne "") {
					my $upload_dir = "/tmp";
					my $filename = $query->param("gel_file__".($i+1));
					$apiis->log('notice', "Prepare uploading the file $filename");
					my $file_info= $query->uploadInfo($filename);
					my $type=$file_info->{'Content-Type'} if (ref $file_info eq 'HASH');
					$filename =~ s/.*[\/\\](.*)/$1/;
	               	my $upload_filehandle = $query->upload("gel_file__".($i+1));
					my $buffer='';
#					unless (defined $type) {
						open UPLOADFILE, ">$upload_dir/$filename";
						binmode UPLOADFILE;
						while (<$upload_filehandle>){
			  				print UPLOADFILE;
						}
						close UPLOADFILE;
#					} 
#					else {
#			  			my $data;
#			  			
#			  			while (my $bytesread=read($upload_filehandle,$data,1024)) {
#			    			$buffer.=$data;
#			    			$apiis->log('notice', "--");
#			  			}
#					}
#					# if file content is empty raise error
#					if ($buffer eq '') {
#				  		$apiis->log('err', "Empty file");
#				  		$hash_ref->{err_status}=1;
#				  		push @errors,  
#				  			Apiis::Errors->new(
#							type      => 'DATA',
#							severity  => 'ERR',
#							ext_fields=> [qw/ blobfile /],
#							from      => 'upload_gels',
#							msg_short => __("The file is empty, does not exist, or was not uploaded"),
#						);
#				  		last EXIT;
#					}
	    		} 
		    	else {
					$apiis->log('err', "File name missing on insert");
					$hash_ref->{err_status}=1;
					push @errors,  Apiis::Errors->new(
						  type      => 'DATA',
						  severity  => 'ERR',
						  ext_fields=> [qw/ blobfile /],
						  from      => 'upload_gels',
						  msg_short => "Cannot insert gel image without uploaded file",
					);
					last EXIT;
		   		}
            } # end for uploading gel
        }
    } # end EXIT
    return( $hash_ref->{err_status},\@errors);
}

sub amplify_pcr{
	my ($form_input,$data_session_ref,$cgi) = @_;
	
	# get a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    	$apiis->log("notice", "Get samples_number from Form: ".$samples_number);
    }
	# build a data hash for results
	my $no = 0;
	for (my $row=1; $row<=$samples_number; $row++){
		if (defined $form_input->{"selected__$row"}){
			my %hash_ref;
			$no++;
			$hash_ref{'row_id'} = "row__".$no;
			$hash_ref{'no'} = $no;
			# set styles for rows in the result table
			if ($no % 2){
			   	$hash_ref{'row_style'} = "table_row_even";
			}
			else {
			  	$hash_ref{'row_style'} = "table_row_odd";
			}
			# get sample_id from ext_sample
			my $sample_id =  Molab::Ajax::get_uniqueID('samples','sample_id','ext_sample',$form_input->{"ext_sample__$row"});
			# get storage_id level 5 from sample_id
			my $storage_id_l5 =  Molab::Ajax::get_uniqueID('storage_samples','storage_id','sample_id',$sample_id);
			# get storage_name level 5 from storage_id level 5
			my $storage_name_l5 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l5);
			# get storage_id level 4 (parent_id level 5) from storage_id level 5
			my $storage_id_l4 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l5);
			# get storage_name level 4 from storage_id level 4
			my $storage_name_l4 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l4);
			# get storage_id level 3 (parent_id level 4) from storage_id level 4
			my $storage_id_l3 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l4);
			# get storage_name level 3 from storage_id level 3
			my $storage_name_l3 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l3);
			# get storage_id level 2 (parent_id level 3) from storage_id level 3 
			my $storage_id_l2 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l3);
			# get storage_name level 2 from storage_id level 2
			my $storage_name_l2 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l2);
			# get storage_id level 1 (parent_id level 2) from storage_id level 2
			my $storage_id_l1 =  Molab::Ajax::get_uniqueID('storage','parent_id','storage_id',$storage_id_l2);
			# get storage_name level 1 from storage_id level 1
			my $storage_name_l1 =  Molab::Ajax::get_uniqueID('storage','storage_name','storage_id',$storage_id_l1);
			
	    	$hash_ref{'result_col1'} = $form_input->{"ext_sample__$row"};
	    	$hash_ref{'result_col2'} = $storage_name_l1;
	    	$hash_ref{'result_col3'} = $storage_name_l2;
	    	$hash_ref{'result_col4'} = $storage_name_l3;
	    	$hash_ref{'result_col5'} = $storage_name_l4;
	    	$hash_ref{'result_col6'} = $storage_name_l5;
	    	$hash_ref{'result_col7'} = "";
	    	$hash_ref{'result_col8'} = ""; 
	    	$hash_ref{'result_col9'} = "";
			push(@record_sets, \%hash_ref);
		}
	}
	return (\@record_sets);
}

sub check_pcr_products{
	my ($form_input,$data_session_ref,$cgi) = @_;
	my %my_session;		# for storing local data sessions
	my $sql;			# for storing a valid sql statement 
	# retrieve related data from amplification ID
	my $amplification_id = $form_input->{"pcr_amplification_loop"};
	
	# make a hash of PCR details such as project name, protocol name, ..
	$sql  = "SELECT pa.amplification_id, pa.ext_amplification, ";
	$sql .= "       pj.project_name, ptc.protocol_name, pa.comment ";
	$sql .= "FROM pcr_amplification pa, protocols ptc, projects  pj ";
	$sql .= "WHERE pa.project_id = pj.project_id ";
	$sql .= "AND pa.protocol_id = ptc.protocol_id ";
	$sql .= "AND pa.amplification_id = '$amplification_id'";
	my $pcr_details_ref = Molab::Ajax::read_data($sql);
	$apiis->check_status;
	
	# create an array of DNA samples
	$sql  = "SELECT s.ext_sample as dna_id ";
	$sql .= "FROM amplified_samples a, samples s ";
	$sql .= "WHERE a.dna_id=s.sample_id ";
	$sql .= "AND amplification_id='$amplification_id'";
	$apiis->log('notice',"SQL: " . $sql);
	my $dna_list_ref = $apiis->DataBase->user_sql($sql);
    $apiis->check_status;
    
    my @dna_lists = ( );
    my $row_no = 0;		# count the number of records
	# process and put data records to the '@dna_lists' array 
    while ( my $hash_ref = $dna_list_ref->handle->fetchrow_hashref) {
		$row_no++;
        $hash_ref->{'no'} = $row_no;
        $hash_ref->{'row_id'} = "row__".$row_no;
		# set styles for rows in the result table
		if ($row_no % 2){
		   	$hash_ref->{'row_style'} = "table_row_even";
		}
		else {
		  	$hash_ref->{'row_style'} = "table_row_odd";
		}
	    # assign names to data fields
    	$hash_ref->{'ext_sample'}     = "ext_sample__".$row_no;
    	$hash_ref->{'lane'}           = "lane__".$row_no;
    	$hash_ref->{'valid'}          = "valid__".$row_no;
	    $hash_ref->{'sample_group'}   = "sample_group__". $row_no;	    
		$hash_ref->{'gel_no'}         = "gel_no__".$row_no;
		$hash_ref->{'ext_sample_val'} = $hash_ref->{'dna_id'};
	    $hash_ref->{'lane_val'}       = "";
		$hash_ref->{'checking'}      	= "checking__". $row_no;
		push(@dna_lists, \%$hash_ref);
    }
    # prepare a set of protocols for PCR validation
    my @valid_protocols = ( );
   	my $protocol_counter = 0;
   	my $sql;
   	
   	$sql = "SELECT protocol_id as value,protocol_name as name
            FROM protocols pt
              LEFT JOIN codes cd on cd.db_code=pt.db_protocol_type
            WHERE cd.ext_code='PCR validation'";
   	my $id = 0;
   	my $sql_ref = $apiis->DataBase->user_sql($sql);
   	while (my $hash_ref = $sql_ref->handle->fetchrow_hashref){
		$id++; 
		$protocol_counter++;
		$hash_ref->{'id'} = $id;
		push(@valid_protocols, $hash_ref);
    }
    $my_session{'protocol_counter'} = $protocol_counter;
	return (\@dna_lists,$pcr_details_ref,$row_no,
			\@valid_protocols,$protocol_counter,\%my_session);
}

sub run_electrophoresis{
	my ($form_input,$data_session_ref,$cgi) = @_;
	# get a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    	$apiis->log("notice", "Get samples_number from Form: ".$samples_number);
    }  
	# build a data hash for results
	my $no = 0;
	for (my $row=1; $row<=$samples_number; $row++){
		if (defined $form_input->{"selected__$row"}){
			my %hash_ref;
			$no++;
			$hash_ref{'row_id'} = "row__".$no;
			$hash_ref{'no'} = $no;
			# set styles for rows in the result table
			if ($no % 2){
			   	$hash_ref{'row_style'} = "table_row_even";
			}
			else {
			  	$hash_ref{'row_style'} = "table_row_odd";
			}
	    	$hash_ref{'result_col1'} = $form_input->{"pcrname__$row"};
	    	$hash_ref{'result_col2'} = $form_input->{"ext_sample__$row"};
	    	$hash_ref{'result_col3'} = $form_input->{"markers__$row"};
			push(@record_sets, \%hash_ref);
		}
	}
	return (\@record_sets);
}

sub load_electrophoresis{
	my ($form_input,$data_session_ref,$cgi) = @_;
	my %my_session;		# for storing local data sessions
	my $sql;			# for storing a valid sql statement 
	my $project_name;
	my $electrophoresis_id;
	my $ext_electrophoresis;
	
	# read a step number and what action must be executed
    my $step   = $form_input->{'step'};
    my $action = $form_input->{'action'};
    # send these data to log
    $apiis->log("info", "Which step?       --> ".$step);
    $apiis->log("info", "What action?      --> ".$action);
	
    # if the user pushed on the button 'Next' at Step '1'   
	if (($step eq "1") and ($action eq "next")){
		$apiis->log('notice','[Form 1]  ---next--> [Form 2]');
		# send data at Step 1 to Session 
		$my_session{"project_name_loop"} = $form_input->{'project_name_loop'};
    	$my_session{"ext_electrophoresis_loop"}= 
    	  $form_input->{'ext_electrophoresis_loop'};
    	$project_name = 
    	  Molab::Ajax::get_uniqueID(
    	    'projects','project_name','project_id',
    	    $form_input->{'project_name_loop'}
    	  );
    	$electrophoresis_id  = $form_input->{'ext_electrophoresis_loop'};
    	$ext_electrophoresis = 
    	  Molab::Ajax::get_uniqueID(
    	    'electrophoresises','ext_electrophoresis',
    	    'electrophoresis_id',$electrophoresis_id
    	  );
	}
	return (\%my_session,$project_name,$electrophoresis_id,$ext_electrophoresis);
}

sub analyze_genotyping{
	my ($form_input,$data_session_ref,$cgi) = @_; 
	# uploade the file to server if possible
	my $uploaded;
	my %my_session;
    my $ref;       # the variable used to point to the input file
	if ($cgi->param("action") eq "upload_file"){
  		$apiis->log("notice", "start uploading file ...");
        $uploaded = upload_file($cgi);
    }
	# declare nessary variables    
    my $project_name_loop;
    my $protocol_name_loop;
    my $samples_number;
	my @record_sets = ( );
	my $project_id;
	my $marker_id;
		
	# read a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    }
    elsif (defined $$data_session_ref{"samples_number"})  {
    	$samples_number = $$data_session_ref{"samples_number"};
    }
    else {
    	$samples_number = 0;
    }
    # read a step number and what action must be executed
    my $step   = $form_input->{'step'};
    my $action = $form_input->{'action'};
    # send these data to log
    $apiis->log("info", "How many samples? --> ".$samples_number); 
    $apiis->log("info", "Which step?       --> ".$step);
    $apiis->log("info", "What action?      --> ".$action);
    
    # if the user pushed on the button 'Next' at Step '1'   
	if (($step eq "1") and ($action eq "next")){
		$apiis->log('debug','[Form 1]  ---next--> [Form 2]');
		# send data at Step 1 to Session 
		$my_session{"project_name_loop"} = $form_input->{'project_name_loop'};
    	$my_session{"markers_loop"}      = $form_input->{'markers_loop'};
    	$my_session{"samples_number"}    = $form_input->{'samples_number'};
    	$my_session{"selected_samples"}  = $form_input->{'selected_samples'};
    	$my_session{"project_id"}        = $form_input->{'project_name_loop'};
    	$my_session{"marker_id"}         = $form_input->{'markers_loop'};
    	$project_id = $form_input->{'project_name_loop'};
    	$marker_id  = $form_input->{'markers_loop'};
    	my $no = 0;
    	my $block = 0;
    	my $seq = 0;
		for (my $row=1; $row<=$samples_number; $row++){
		  $apiis->log('info',"$row:".$form_input->{"selected__$row"});
		  $block++;
		  if (defined $form_input->{"selected__$row"}){
		  	$no = 0;
			for(my $result=1;$result<=$form_input->{"result__$row"};$result++){	
			   	my %hash_ref; $no++;$seq++;
			   	$hash_ref{'no'} = $seq;
			    # set styles for rows in the result table
			    if ($block % 2){
			       	$hash_ref{'row_style'} = "table_row_even";
			    }
			    else {
			      	$hash_ref{'row_style'} = "table_row_odd";
			    }
			    $hash_ref{'ext_dna'}  = "ext_dna__".$row."_".$no;
			    $hash_ref{'allele1'} = "allele1__".$row."_".$no;
			    $hash_ref{'allele2'} = "allele2__".$row."_".$no;
				$hash_ref{'comment'} = "comment__".$row."_".$no;
				if ($result eq '1'){
					$hash_ref{'result'}  = "result__".$row;
				}
				else{
					$hash_ref{'result'} = "dummy";
				}
											
				$hash_ref{'ext_dna_val'} = $form_input->{"ext_sample__".$row}
					if (defined $form_input->{"ext_sample__". $row});
				$hash_ref{'result_val'} = $form_input->{"result__".$row}
					if (defined $form_input->{"result__". $row});
				$apiis->log('notice',"ext_dna:".$hash_ref{'ext_dna_val'});	
				push(@record_sets, \%hash_ref);
			}
		  }
		}
	}
	# if the user pushed on the button 'Upload' at Step '2'
	elsif ( ($step eq "2") and ($action eq "upload_file") ){
		$apiis->log('debug','[Form 2]  ---reload--> [Form 2]');
	    # picking up data fields from the uploaded data file
	    my $upload_dir = "/tmp";
	    my $filename = $cgi->param("blobfile");
	    $apiis->log("notice", "Reading from: " . "$upload_dir/$filename");
		$ref = ReadData ("$upload_dir/$filename", sep => ",");
		$apiis->log("notice", "This file contains ", 
		  $ref->[1]{maxcol}, " columns (fields) and ", 
		  $ref->[1]{maxrow}, " rows (lines).\n");
		my $no = 0;
		for (my $row=1; $row<=$samples_number; $row++){
			$no = 0;
			for(my $result=1;$result<=$form_input->{"result__$row"};$result++){	
				$no++;
				$hash_ref{'no'} = $no;
				$apiis->log("notice", "ext_dna__".$row."_".$no."=". $form_input->{"ext_dna__".$row."_".$no});
				if (defined $form_input->{"ext_dna__".$row."_".$no}){
					my %hash_ref;
					# set styles for rows in the result table
					if ($no % 2){
						$hash_ref{'row_style'} = "table_row_even";
					}
					else {
						$hash_ref{'row_style'} = "table_row_odd";
					}
					if ($result eq '1'){
						$hash_ref{'result'}  = "result__".$row;
					}
					else{
						$hash_ref{'result'} = "dummy";
					}
					$hash_ref{'ext_dna'}  = "ext_dna__".$row."_".$no;
					$hash_ref{'allele1'} = "allele1__".$row."_".$no;
					$hash_ref{'allele2'} = "allele2__".$row."_".$no;
					$hash_ref{'comment'} = "comment__".$row."_".$no;			
					$hash_ref{'ext_dna_val'} = $ref->[1]{cell}[1][$row+1];
					$hash_ref{'allele1_val'} = $ref->[1]{cell}[2][$row+1];
					$hash_ref{'allele2_val'} = $ref->[1]{cell}[3][$row+1];
					$hash_ref{'comment_val'} = $ref->[1]{cell}[4][$row+1];
					
					$hash_ref{'result_val'} = $form_input->{"result__".$row}
						if (defined $form_input->{"result__". $row});
					push(@record_sets, \%hash_ref);
				}
			}
		}
	}
	else {
		$apiis->log('notice','Loading data for the first form');
	}
	return (\@record_sets,\%my_session,$project_id,$marker_id);
}

sub analyze_sequencing{
	my ($form_input,$data_session_ref,$cgi) = @_; 
	# uploade the file to server if have
	my $uploaded;
	my %my_session;
    my $ref;       # the variable used to point to the input file
	if ($cgi->param("action") eq "upload_file"){
  		$apiis->log("notice", "start uploading file ...");
        $uploaded = upload_file($cgi);
    }
	# declare nessary variables    
    my $samples_number;
	my @record_sets = ( );
		
	# read a number of samples
    if (defined $form_input->{'samples_number'}){
    	$samples_number = $form_input->{'samples_number'};
    }
    elsif (defined $$data_session_ref{"samples_number"})  {
    	$samples_number = $$data_session_ref{"samples_number"};
    }
    else {
    	$samples_number = 0;
    }
    # reload the data in the next form.
    my $step   = $form_input->{'step'};
    my $action = $form_input->{'action'};
    # send to log
    $apiis->log("notice", "How many samples? --> ".$samples_number); 
    $apiis->log("notice", "Which step?       --> ".$step);
    $apiis->log("notice", "What action?      --> ".$action);
    
    # if the user pushed on the button 'Next' at Step '1'   
	if (($step eq "1") and ($action eq "next")){
		$apiis->log('debug','[Form 1]  ---next--> [Form 2]');
		# send data at Step 1 to Session 
		$my_session{"project_name_loop"} = $form_input->{'project_name_loop'};
    	$my_session{"markers_loop"}      = $form_input->{'markers_loop'};
    	$my_session{"samples_number"}    = $form_input->{'samples_number'};
    	$my_session{"selected_samples"}  = $form_input->{'selected_samples'};
    	$my_session{"project_id"}        = $form_input->{'project_name_loop'};
    	$my_session{"marker_id"}         = $form_input->{'markers_loop'};
    	$project_id = $form_input->{'project_name_loop'};
    	$marker_id  = $form_input->{'markers_loop'};
    	my $no = 0;
    	my $block = 0;
    	my $seq = 0;
		for (my $row=1; $row<=$samples_number; $row++){
	      $apiis->log('notice',"$row:".$form_input->{"selected__$row"});
	      $block++;
		  if (defined $form_input->{"selected__$row"}){
			$no = 0;
			for(my $result=1;$result<=$form_input->{"result__$row"};$result++){				
			   	my %hash_ref; $no++;$seq++;
			   	$hash_ref{'no'} = $seq;
			    # set styles for rows in the result table
			    if ($block % 2){
			       	$hash_ref{'row_style'} = "table_row_even";
			    }
			    else {
			      	$hash_ref{'row_style'} = "table_row_odd";
			    }
			    $hash_ref{'ext_dna'}     = "ext_dna__".$row."_".$no;
			    $hash_ref{'input_file'}  = "input_file__".$row."_".$no;
			    $hash_ref{'comment'}     = "comment__".$row."_".$no;		
				if ($result eq '1'){
					$hash_ref{'result'}  = "result__".$row;	
				}
				else{
					$hash_ref{'result'} = "dummy";
				}
				$hash_ref{'ext_dna_val'} = $form_input->{"ext_sample__".$row}
					if (defined $form_input->{"ext_sample__". $row});
				$apiis->log('notice',"ext_dna:".$hash_ref{'ext_dna_val'});	
				$hash_ref{'result_val'} = $form_input->{"result__".$row}
					if (defined $form_input->{"result__". $row});
				push(@record_sets, \%hash_ref);
			}
		  }
		}
	}
	# if the user pushed on the button 'Upload' at Step '2'
	elsif ( ($step eq "2") and ($action eq "upload_file") ){
		$apiis->log('debug','[Form 2]  ---reload--> [Form 2]');
	    # picking up data fields from the uploaded data file
	    my $upload_dir = "/tmp";
	    my $filename = $cgi->param("blobfile");
	    $apiis->log("notice", "Reading from: " . "$upload_dir/$filename");
		$ref = ReadData ("$upload_dir/$filename", sep => ",");
		$apiis->log("notice", "This file contains ", $ref->[1]{maxcol}, " columns (fields) and ", $ref->[1]{maxrow}, " rows (lines).\n");
		my $no = 0;
		for (my $row=1; $row<=$samples_number; $row++){
			if (defined $form_input->{"ext_dna__". $row}){
			   	my %hash_ref;
			   	$no++;
			   	$hash_ref{'no'} = $no;
			    # set styles for rows in the result table
			    if ($no % 2){
			       	$hash_ref{'row_style'} = "table_row_even";
			    }
			    else {
			      	$hash_ref{'row_style'} = "table_row_odd";
			    }
			    $hash_ref{'ext_dna'}  = "ext_dna__".$row;
			    $hash_ref{'allele1'} = "allele1__".$row;
			    $hash_ref{'allele2'} = "allele2__".$row;
				$hash_ref{'comment'} = "comment__".$row;			
				$hash_ref{'ext_dna_val'} = $ref->[1]{cell}[1][$row];
				$hash_ref{'allele1_val'} = $ref->[1]{cell}[2][$row];
				$hash_ref{'allele2_val'} = $ref->[1]{cell}[3][$row];
				push(@record_sets, \%hash_ref);
			}
		}
	}
	# the back button in the second form was pushed
	elsif ( ($step eq "2") and ($action eq "back") ){
		$apiis->log('notice','[Form 1]  <--back--- [Form 2]');
		# store data in the second form
		for(my $row=1; $row<=$samples_number; $row++){
			$my_session{"ext_organism__".$row} = $form_input->{"ext_organism__".$row};
    		$my_session{"ext_sample__".$row} = $form_input->{"ext_sample__".$row};
    		$my_session{"volume__".$row} = $form_input->{"volume__".$row};
    		$my_session{"sampling_dt__".$row} = $form_input->{"sampling_dt__".$row};
    		$my_session{"db_user__".$row} = $form_input->{"db_user__".$row};
    		$my_session{"comment__".$row} = $form_input->{"comment__".$row};
    	}
	}
	else {
		$apiis->log('notice','Loading data for the first form');
	}
	return (\@record_sets,\%my_session,$project_id,$marker_id);
}

##############################################################################

=head1 AUTHORS

  Truong Van Chi Cong <cong.chi@fli.bund.de>
  
=cut

1;
